// Copyright (c) 2008-2011 Mighty Heave Software LLC.
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
#include "amgchk.h"
#include "grammar.h"
#include <stdlib.h>
#include <math.h>
#include <float.h>
#include <time.h>
#include "ints.h"
#ifdef _WIN32
#include <windows.h>
#endif

class pnode;
class parsetree;
typedef hashmap<parsetree*,int,4> amgmap;

static uint64_t GetSystemTicks()
{
#ifdef _WIN32
	LARGE_INTEGER PerformanceCount;
	QueryPerformanceCounter(&PerformanceCount);
	return (uint64_t)PerformanceCount.QuadPart;
#else
  return 0;
#endif
}

static double TicksToSeconds(uint64_t ticks)
{
#ifdef _WIN32
	static uint64_t TicksPerSecond = 0;
	if( ! TicksPerSecond )
	{
		LARGE_INTEGER PerformanceFrequency;
		QueryPerformanceFrequency(&PerformanceFrequency);
		TicksPerSecond = (uint64_t)PerformanceFrequency.QuadPart;
	}
	return (double)(ticks / TicksPerSecond) + (double)(ticks % TicksPerSecond) / (double)TicksPerSecond;
#else
  return (double)ticks;
#endif
}

class amgpath
{
public:
    amgmap m_mappings;
    vector<parsetree*> m_amgnodes;
};

class nlink
{
public:
    nlink() : m_prev(0), m_t(0), m_reject(false) {}
    ~nlink();
    pnode *m_prev;
    parsetree *m_t;
    bool m_reject;
};

class pnode
{
public:
    int m_ref;
    int m_state;
    bool m_reject;
    vector<nlink> m_links;
    bool m_processed;

    pnode() : m_ref(1), m_state(-1), m_reject(false), m_processed(false)
    {
    }

    pnode(int s) : m_ref(1), m_state(s), m_reject(false), m_processed(false)
    {
    }

    void AddRef()
    {
        ++m_ref;
    }

    void DecRef()
    {
        if( --m_ref == 0 )
            delete this;
    }
};

enum e_treetype
{
    TreeUndefined,
    TreeTerminal,
    TreeProduction,
    TreeAmbiguousNonterminal,
    TreeSkip
};

class parsetree
{
    int m_ref;

public:
    e_treetype m_ttype;
    int m_sym;
    vector<parsetree*> m_subtrees;
    int m_rank;
    int m_reduceid;

    parsetree(e_treetype ttype, int sym)
        : m_ttype(ttype), m_sym(sym), m_ref(1), m_rank(101), m_reduceid(0)
    {
    }

    ~parsetree()
    {
        for( vector<parsetree*>::iterator curtree = m_subtrees.begin(); curtree != m_subtrees.end(); ++curtree )
            (*curtree)->DecRef();
        m_subtrees.clear();
    }

    void AddRef()
    {
        ++m_ref;
    }

    void DecRef()
    {
        if( --m_ref == 0 )
            delete this;
    }

    bool HasAmbiguity()
    {
        if( m_ttype == TreeAmbiguousNonterminal )
            return true;
        for( vector<parsetree*>::iterator curtree = m_subtrees.begin(), endtree = m_subtrees.end(); curtree != endtree; ++curtree )
            if( (*curtree)->HasAmbiguity() )
                return true;
        return false;
    }

    bool IsIdentical(parsetree *rhs)
    {
        if( m_ttype == rhs->m_ttype && m_sym == rhs->m_sym && m_reduceid == rhs->m_reduceid && m_subtrees == rhs->m_subtrees )
            return true;
        return false;
    }

    void collect_ambiguous_nodes_shallow(vector<parsetree*> &nodes)
    {
        if( m_ttype == TreeAmbiguousNonterminal )
        {
            nodes.push_back(this);
            return; // go no deeper
        }
        for( vector<parsetree*>::iterator curtree = m_subtrees.begin(), endtree = m_subtrees.end(); curtree != endtree; ++curtree )
            (*curtree)->collect_ambiguous_nodes_shallow(nodes);
    }

    static int check_sl_pref(Grammar *g, parsetree *prev, parsetree *next)
    {
        if( prev->m_ttype != TreeProduction || next->m_ttype != TreeProduction || prev->m_sym  != next->m_sym )
            return 0; // bad input
        Production *p = g->m_productions[prev->m_sym];
        for( size_t i = 0, n = p->m_items.size(); i < n; ++i )
        {
            sl_decl cursl = p->m_sldecls[i];
            if( cursl == sl_neutral )
                continue;
            int prevcnt = prev->m_subtrees[i]->symbol_count();
            int nextcnt = next->m_subtrees[i]->symbol_count();
            if( prevcnt == nextcnt )
                continue;
            if( cursl == sl_short )
                // prefer shorter, so if prevcnt is 2 and nextcnt is 3, prefer prev (return positive value)
                return nextcnt - prevcnt;
            // prefer longer, so if prevcnt is 3 and nextcnt is 2, prefer prev (return positive value)
            return prevcnt - nextcnt;
        }
        return 0;
    }

    int symbol_count()
    {
        switch( m_ttype )
        {
        case TreeTerminal:
            return 1;
        case TreeProduction:
            {
            int count = 0;
            for( vector<parsetree*>::iterator curtree = m_subtrees.begin(), endtree = m_subtrees.end(); curtree != endtree; ++curtree )
                count += (*curtree)->symbol_count();
            return count;
            }
        case TreeAmbiguousNonterminal:
            if( m_subtrees.size() == 0 )
                return 0;
            return m_subtrees[0]->symbol_count();
        default:
            break;
        }
        return 0;
    }

    static void Indent(outbuf *stream, int indentdepth)
    {
        for( int i = 0, n = indentdepth/10; i < n; ++i )
            ob_puts(stream, "                                        ");
        for( int i = 0, n = indentdepth%10; i < n; ++i )
            ob_puts(stream, "    ");
    }

    void print_symbols(Grammar *g, outbuf *stream, amgmap *mappings, int indentdepth = 0)
    {
        switch( m_ttype )
        {
        case TreeTerminal:
            Indent(stream, indentdepth);
            g->m_symbols[m_sym]->Print(stream);
            ob_putc(stream, '\n');
            break;
        case TreeProduction:
            {
            Indent(stream, indentdepth);
            Production *p = g->m_productions[m_sym];
            p->Print(stream);
            if( m_reduceid != 0 )
                ob_printf(stream, " [Test %d]", m_reduceid);
            ob_puts(stream, " = (\n");
            for( vector<parsetree*>::iterator curtree = m_subtrees.begin(), endtree = m_subtrees.end(); curtree != endtree; ++curtree )
                (*curtree)->print_symbols(g, stream, mappings, indentdepth+1);
            Indent(stream, indentdepth);
            ob_puts(stream, ")\n");
            }
            break;
        case TreeAmbiguousNonterminal:
            if( mappings )
            {
                Indent(stream, indentdepth);
                int choice = mappings->find(this)->second;
                m_subtrees[choice]->print_symbols(g, stream, mappings, indentdepth);
            }
            else
            {
                Indent(stream, indentdepth);
                ob_puts(stream, "Ambiguous Reduction = (\n");
                for( vector<parsetree*>::iterator curtree = m_subtrees.begin(), endtree = m_subtrees.end(); curtree != endtree; ++curtree )
                    (*curtree)->print_symbols(g, stream, mappings, indentdepth+1);
                Indent(stream, indentdepth);
                ob_puts(stream, ")\n");
            }
            break;
        case TreeSkip:
            Indent(stream, indentdepth);
            ob_puts(stream, "<Optional, not used in this reduction>\n");
            break;
        default:
            Indent(stream, indentdepth);
            ob_puts(stream, "<Unknown>\n");
            break;
        }
    }

    void collect_reduction_ids(amgmap &mappings, vector<int> &reduction_ids)
    {
        // locate all ambiguous nodes.
        switch( m_ttype )
        {
        case TreeTerminal:
            break;
        case TreeProduction:
            if( m_reduceid != 0 )
                reduction_ids.push_back(m_reduceid);
            for( vector<parsetree*>::iterator curtree = m_subtrees.begin(), endtree = m_subtrees.end(); curtree != endtree; ++curtree )
                (*curtree)->collect_reduction_ids(mappings, reduction_ids);
            break;
        case TreeAmbiguousNonterminal:
            {
            int choice = mappings.find(this)->second;
            m_subtrees[choice]->collect_reduction_ids(mappings, reduction_ids);
            }
            break;
        default:
            break;
        }
    }

    void Copy(parsetree *next)
    {
        m_ttype = next->m_ttype;
        m_sym = next->m_sym;
        for( vector<parsetree*>::iterator cursubtree = m_subtrees.begin(); cursubtree != m_subtrees.end(); ++cursubtree )
            (*cursubtree)->DecRef();
        m_subtrees = next->m_subtrees;
        for( vector<parsetree*>::iterator cursubtree = m_subtrees.begin(); cursubtree != m_subtrees.end(); ++cursubtree )
            (*cursubtree)->AddRef();
        m_rank = next->m_rank;
        m_reduceid = next->m_reduceid;
    }
};

nlink::~nlink()
{
    if( m_prev )
        m_prev->DecRef();
    if( m_t )
        m_t->DecRef();
}

class lrshift
{
public:
    pnode *m_prevstate;
    int m_shiftto;
    int m_token;
};

class queried_actions
{
protected:
    bool m_bEnd;
    LRAction *m_cur;
    LRAction *m_end;
    GType m_gtype;
    AType m_atype;
    int m_symbol;

    bool test()
    {
        if( m_cur->m_gtype != m_gtype )
            return false;
        if( (m_cur->m_atype & m_atype) == ATypeNone )
            return false;
        if( m_cur->m_gtype == GTypeSymbol && (m_symbol == -1 || m_symbol == m_cur->m_s->GetIndex()) )
            return true;
        if( m_cur->m_gtype == GTypeProduction && (m_symbol == -1 || m_symbol == m_cur->m_p->GetIndex()) )
            return true;
        return false;
    }

public:
    queried_actions(const queried_actions &rhs)
        :   m_bEnd(rhs.m_bEnd), m_cur(rhs.m_cur), m_end(rhs.m_end),
            m_gtype(GTypeNone), m_atype(ATypeNone), m_symbol(-1)
    {
    }

    queried_actions()
        :   m_bEnd(true), m_cur(0), m_end(0),
            m_gtype(GTypeNone), m_atype(ATypeNone), m_symbol(-1)
    {
    }

    queried_actions(Grammar *g, int state, GType gtype, int symbol, AType action)
        :   m_bEnd(false), m_cur(0), m_end(0),
            m_gtype(gtype), m_atype(action), m_symbol(symbol)
    {
        if( state < 0 || state >= (int)g->m_lrstates.size() )
        {
            m_bEnd = true;
            return;
        }
        LRStateSet *ss = g->m_lrstates[state];
        m_cur = ss->m_actions.begin();
        m_end = ss->m_actions.end();
        if( test() )
            return;
        while( ! m_bEnd )
        {
            ++m_cur;
            if( m_cur == m_end )
                m_bEnd = true;
            else if( test() )
                break;
        }
    }

    queried_actions operator++()
    {
        while( ! m_bEnd )
        {
            ++m_cur;
            if( m_cur == m_end )
                m_bEnd = true;
            else if( test() )
                break;
        }
        return *this;
    }

    bool operator==(const queried_actions &rhs) const
    {
        if( m_bEnd && rhs.m_bEnd )
            return true;
        if( m_cur == rhs.m_cur )
            return true;
        return false;
    }

    bool operator!=(const queried_actions &rhs) const
    {
        return ! operator==(rhs);
    }

    LRAction *operator->()
    {
        return m_cur;
    }

    LRAction &operator*()
    {
        return *m_cur;
    }
};

template<>
class simplehash < vector<int> >
{
public:
    uint32_t operator()(const vector<int> &v) const
    {
        return hash_uint32s((const uint32_t*)v.begin(), v.size(), 101);
    }
};

class symbolgenerator
{
public:
    virtual void next_symbols(hashset<int> &symbols) = 0;
};

class searchinfo
{
public:
    AmbiguitiesOptions *m_poptions;
    symbolgenerator *m_sgen;
    vector<int> m_acceptstring;
    uint64_t m_pathsprocessed;
    vector<int> m_fanouts;
    vector<int> m_pos;
    uint64_t m_starttime;
    uint64_t m_endtime;
    uint64_t m_lastupdate;
    int m_endmarkerid;

    searchinfo(AmbiguitiesOptions *poptions, symbolgenerator *sgen)
        :   m_poptions(poptions), m_sgen(sgen), m_pathsprocessed(0),
            m_lastupdate(0), m_endmarkerid(-1), m_starttime(0), m_endtime(0)
    {
    }
};

class glr_parser
{
public:
    glr_parser(Grammar *g);
    ~glr_parser();

    void next_symbols(hashset<int> &symbols);
    void update_states(int nexttok);
    void update_states(const vector<int> &nexttoks);
    void find_and_report_ambiguities(AmbiguitiesOptions *poptions);
    void find_and_report_ambiguities_in_string(AmbiguitiesOptions *poptions, const vector<int> &symbols);

    Grammar *m_g;
    vector<pnode*>          m_activestacks;
    parsetree              *m_accepttree;
    parsetree              *m_skipper;

private:
    vector<pnode*>          m_for_actor;
    vector<pnode*>          m_for_actor_delayed;
    vector<parsetree*>      m_reducepath;
    vector<lrshift>         m_for_shifter;
    int                     m_nexttok;
    hashmap<int,parsetree*> m_tokens;
    bool                    m_cull_nonambig;
    int                     m_nextreduceid;

private:
    pnode *FindState(vector<pnode*> &states, int nextstate);
    pnode *FindOrAddState(vector<pnode*> &states, int nextstate, bool *pbIsNew);
    nlink *FindLink(pnode *node, pnode *prev);
    nlink *FindOrAddLink(pnode *node, pnode *prev, bool bReject, bool *pbIsNew);
    bool drop_rejected_paths(pnode *state);
    void do_reductions(pnode *state, int count, Production *p, AType action, int reduceid);
    void reducer(bool bReject, pnode *state, int nextstate, parsetree *t);
    void do_new_reductions(nlink *plinknew, bool bFound, pnode *state, int count, Production *p, AType action, int reduceid);
    int make_reduce_id(pnode *state, LRAction &action, Production *p);

private:
    void update_start();
    void update_step(int nexttok);
    void update_shifters();
    void update_end();
    parsetree *terminal_node(int sym);
    parsetree *reduce_nodes(Production *p, int reduceid);
    parsetree *merge_nodes(parsetree *prev, parsetree *next);
    void pnode_updatelink(pnode *node, nlink *plink, bool reject);
    bool do_ambiguities_search(searchinfo &info, int depth);

private:
    void run_reduce_tests(vector<amgmap> &ambiguous_reductions);
};

class symbolgenerator_glrparser : public symbolgenerator
{
public:
    glr_parser *m_parser;
    symbolgenerator_glrparser(glr_parser *parser) : m_parser(parser)
    {
    }

    virtual void next_symbols(hashset<int> &symbols)
    {
        m_parser->next_symbols(symbols);
    }
};

class symbolgenerator_string : public symbolgenerator
{
public:
    vector<int> m_string;
    size_t m_pos;

    symbolgenerator_string(const vector<int> &string) : m_string(string), m_pos(0)
    {
    }

    virtual void next_symbols(hashset<int> &symbols)
    {
        if( m_pos < m_string.size() )
            symbols.insert( m_string[m_pos++] );
    }
};

glr_parser::glr_parser(Grammar *g) : m_g(g), m_nexttok(-1), m_accepttree(0), m_skipper(0), m_nextreduceid(0), m_cull_nonambig(false)
{
    FindOrAddState(m_activestacks, 0, 0);
    m_skipper = new parsetree(TreeSkip, 0);
}

glr_parser::~glr_parser()
{
    for( hashmap<int,parsetree*>::iterator cur = m_tokens.begin(), endtok = m_tokens.end(); cur != endtok; ++cur )
        cur->second->DecRef();
    if( m_accepttree )
        m_accepttree->DecRef();
    for( vector<pnode*>::iterator curnode = m_activestacks.begin(), endnode = m_activestacks.end(); curnode != endnode; ++curnode )
        (*curnode)->DecRef();
    if( m_skipper )
        m_skipper->DecRef();
}

parsetree *glr_parser::terminal_node(int sym)
{
    parsetree *t = 0;
    hashmap<int,parsetree*>::iterator iter = m_tokens.find(sym);
    if( iter == m_tokens.end() )
    {
        t = new parsetree(TreeTerminal, sym);
        m_tokens[sym] = t;
    }
    else
        t = iter->second;
    t->AddRef();
    return t;
}

parsetree *glr_parser::reduce_nodes(Production *p, int reduceid)
{
    int pindex = -1;
    if( p->m_copyof )
        pindex = p->m_copyof->GetIndex();
    else
        pindex = p->GetIndex();
    parsetree *t = new parsetree(TreeProduction, pindex);
    t->m_rank= p->m_rank;
    t->m_subtrees.insert(t->m_subtrees.end(), m_reducepath.begin()+(m_reducepath.size()-p->m_items.size()), m_reducepath.end());
    t->m_subtrees.reverse();
    for( size_t i = 0, n = t->m_subtrees.size(); i < n; ++i )
    {
        parsetree *subt = t->m_subtrees[i];
        subt->AddRef();
    }
    t->m_reduceid = reduceid;
    // All is well, but introduce the skippoints.
    if( p->m_copyof )
    {
        for( vector<SkipRange>::iterator curskip = p->m_skips.begin(), endskip = p->m_skips.end(); curskip != endskip; ++curskip )
        {
            for( int i = curskip->m_start, n = curskip->m_stop; i < n; ++i )
            {
                m_skipper->AddRef();
                t->m_subtrees.insert(t->m_subtrees.begin()+i, m_skipper);
            }
        }
    }
    return t;
}

// prev may be NULL.
// next may have an ambiguity, but is never TreeAmbiguous
parsetree *glr_parser::merge_nodes(parsetree *prev, parsetree *next)
{
    if( ! prev || prev == next )
        return next;
    if( prev->m_rank < next->m_rank )
    {
        next->DecRef();
        return prev;
    }
    if( next->m_rank < prev->m_rank )
    {
        // overwrite next with prev.
        prev->Copy(next);
        next->DecRef();
        return prev;
    }

    // already ambiguous.
    if( prev->m_ttype == TreeAmbiguousNonterminal )
    {
        // Watch for re-reduction.  Nonambiguous parse is gone, do we still need this?
        for( vector<parsetree*>::iterator curtree = prev->m_subtrees.begin(), endsubtree = prev->m_subtrees.end(); curtree != endsubtree; ++curtree )
        {
            if( (*curtree)->IsIdentical(next) )
            {
                next->DecRef();
                return prev;
            }
        }
        // Eliminate long/short versions.
        // There will be only one match.  If not, it means the previous comparison held both options.
        for( vector<parsetree*>::iterator curtree = prev->m_subtrees.begin(), endsubtree = prev->m_subtrees.end(); curtree != endsubtree; ++curtree )
        {
            if( (*curtree)->m_sym != next->m_sym )
                continue;
            int result = parsetree::check_sl_pref(m_g, *curtree, next);
            if( result == 0 )
                break; // keep both
            if( result > 0 )
                next->DecRef();
            else if( result < 0 )
            {
                (*curtree)->DecRef();
                *curtree = next;
            }
            return prev;
        }
        // add to ambiguous forest
        prev->m_subtrees.push_back(next);
        return prev;
    }

    // Watch for re-reduction.  Nonambiguous parse is gone, do we still need this?
    if( prev->IsIdentical(next) )
    {
        next->DecRef();
        return prev;
    }

    // Eliminate long/short versions of the same production
    if( prev->m_sym == next->m_sym )
    {
        int result = parsetree::check_sl_pref(m_g, prev, next);
        if( result > 0 )
        {
            next->DecRef();
            return prev;
        }
        else if( result < 0 )
        {
            // Overwrite next with prev
            prev->Copy(next);
            next->DecRef();
            return prev;
        }
    }

    // We have to merge to create an ambiguous node.
    // prev may already have parents.
    // rewrite prev as an ambiguous node.
    parsetree *subprev = new parsetree(prev->m_ttype, prev->m_sym);
    subprev->m_ttype = prev->m_ttype;
    subprev->m_sym = prev->m_sym;
    subprev->m_subtrees = prev->m_subtrees;
    subprev->m_rank = prev->m_rank;
    subprev->m_reduceid = prev->m_reduceid;

    prev->m_rank = subprev->m_rank; // Rank is already established to be equal.
    prev->m_ttype = TreeAmbiguousNonterminal;
    prev->m_sym = m_g->m_productions[next->m_sym]->m_nt->GetIndex();
    prev->m_subtrees.clear();
    prev->m_subtrees.insert(prev->m_subtrees.end(), subprev);
    prev->m_subtrees.insert(prev->m_subtrees.end(), next);
    prev->m_reduceid = 0;

    return prev;
}

void glr_parser::pnode_updatelink(pnode *node, nlink *plink, bool reject)
{
    plink->m_reject |= reject;
    // Any state that has all rejected links is considered rejected.
    // Shortcut: Previously rejected states that do not add an unrejected link remain rejected.
    // Shortcut: Previously accepted states that do not cause a link to be rejected remain unrejected.
    if( node->m_reject == plink->m_reject )
        return;
    // The long(er) way.  See if any link is unrejected.
    bool rflag = true;
    for( int i = 0, n = node->m_links.size(); i < n; ++i )
    {
        if(node->m_links[i].m_reject)
            continue;
        rflag = false;
        break;
    }
    node->m_reject = rflag;
}

pnode *glr_parser::FindState(vector<pnode*> &states, int nextstate)
{
    pnode *state = 0;
    vector<pnode*>::iterator fnd = states.begin();
    vector<pnode*>::iterator endfnd = states.end();
    while( fnd != endfnd )
    {
        if( (*fnd)->m_state == nextstate )
            break;
        ++fnd;
    }
    if( fnd == endfnd )
        return 0;
    state = *fnd;
    return state;
}

pnode *glr_parser::FindOrAddState(vector<pnode*> &states, int nextstate, bool *pbIsNew)
{
    pnode *state = 0;
    vector<pnode*>::iterator fnd = states.begin();
    vector<pnode*>::iterator endfnd = states.end();
    while( fnd != endfnd )
    {
        if( (*fnd)->m_state == nextstate )
            break;
        ++fnd;
    }
    if( fnd == endfnd )
    {
        if( pbIsNew )
            *pbIsNew = true;
        state = new pnode(nextstate);
        states.push_back(state);
    }
    else
    {
        if( pbIsNew )
            *pbIsNew = false;
        state = *fnd;
    }
    return state;
}

nlink *glr_parser::FindLink(pnode *node, pnode *prev)
{
    nlink *plink = 0;
    for( size_t i = 0, n = node->m_links.size(); i < n; ++i )
    {
        if( node->m_links[i].m_prev != prev )
            continue;
        plink = &node->m_links[i];
        break;
    }
    return plink;
}

nlink *glr_parser::FindOrAddLink(pnode *node, pnode *prev, bool bReject, bool *pbIsNew)
{
    nlink *plink = 0;
    for( size_t i = 0, n = node->m_links.size(); i < n; ++i )
    {
        if( node->m_links[i].m_prev != prev )
            continue;
        plink = &node->m_links[i];
        if( pbIsNew )
            *pbIsNew = false;
        break;
    }
    if( plink == 0 )
    {
        if( pbIsNew )
            *pbIsNew = true;
        plink = node->m_links.newitem();
        plink->m_reject = bReject;
        prev->AddRef();
        plink->m_prev = prev;
        // Shortcut: Adding an unreject nlink causes the node to become unrejected.
        if( ! bReject )
        {
            node->m_reject = false;
            return plink;
        }
    }
    pnode_updatelink(node, plink, bReject);
    return plink;
}

bool glr_parser::drop_rejected_paths(pnode *state)
{
    if( state->m_state == 0 )
        return true;
    if( state->m_reject )
        return false;
    for( size_t i = 0, n = state->m_links.size(); i < n; ++i )        
    {
        nlink *plink = &state->m_links[i];
        pnode *prevnode = plink->m_prev;
        if( plink->m_reject || ! drop_rejected_paths(prevnode) )
        {
            prevnode->DecRef();
            state->m_links.removeat(i, 1);
            --i;
            continue;
        }
    }
    if( state->m_links.size() == 0 )
        return false;
    return true;
}

void glr_parser::do_new_reductions(nlink *plinknew, bool bFound, pnode *state, int count, Production *p, AType action, int reduceid)
{
    if( count > 0 )
    {
        for( size_t i = 0; i < state->m_links.size(); ++i )
        {
            nlink *plink = &state->m_links[i];
            bool bIsNewLink = (plink == plinknew) ?  true : false;
            pnode *prevnode = plink->m_prev;
            m_reducepath.push_back(plink->m_t);
            do_new_reductions(plinknew, bFound || bIsNewLink, prevnode, count-1, p, action, reduceid);
            m_reducepath.pop_back();
        }
        return;
    }
    if( ! bFound )
        return; // skipping, not a new path.
    bool bReject = (action & ATypeReject) ? true : false;
    queried_actions curaction(m_g, state->m_state, GTypeProduction, p->GetIndex(), ATypeShift);
    queried_actions endaction;
    if( curaction == endaction )
        return;
    parsetree *t = reduce_nodes(p, reduceid);
    reducer(bReject, state, curaction->m_shiftto->m_number, t);
    return;
}

int glr_parser::make_reduce_id(pnode *state, LRAction &action, Production *p)
{
    int reduceid = 0;
    int pindex = p->GetIndex();
    if( (action.m_atype == ATypeReduce || action.m_atype == ATypeAccept) && p->IsRuntimeResolvable() )
        reduceid = ++m_nextreduceid;
    return reduceid;
}

void glr_parser::reducer(bool bReject, pnode *state, int nextstate, parsetree *t)
{
    bool bIsNewNode = false;
    pnode *pnextstate = FindOrAddState(m_activestacks, nextstate, &bIsNewNode);
    bool bIsNewLink = false;
    nlink *plink = FindOrAddLink(pnextstate, state, bReject, &bIsNewLink);
    if( bIsNewNode )
    {
        plink->m_t = t;
        if( pnextstate->m_reject )
            m_for_actor_delayed.push_back(pnextstate);
        else
            m_for_actor.push_back(pnextstate);
        return;
    }

    if( bIsNewLink )
    {
        plink->m_t = t;
        vector<pnode*> tmp_activestacks = m_activestacks;
        for( vector<pnode*>::iterator curstateiter = tmp_activestacks.begin(), endstateiter = tmp_activestacks.end(); curstateiter != endstateiter; ++curstateiter )
        {
            pnode *curstate = *curstateiter;
            if( curstate->m_reject )
                continue;
            if( ! curstate->m_processed )
                continue; // state not processed yet, skipping additional reductions
            queried_actions curaction(m_g, curstate->m_state, GTypeSymbol, m_nexttok, (AType)(ATypeReduce|ATypeReject));
            queried_actions endaction;
            if( curaction == endaction )
                continue;
            for( ; curaction != endaction; ++curaction )
            {
                Production * p = curaction->m_reduceby;
                int reduceid = make_reduce_id(curstate, *curaction, p);
                do_new_reductions(plink, false, curstate, p->m_items.size(), p, curaction->m_atype, reduceid);
            }
        }
        return;
    }

    plink->m_t = merge_nodes(plink->m_t, t);
}

void glr_parser::do_reductions(pnode *state, int count, Production *p, AType action, int reduceid)
{
    if( count > 0 )
    {
        for( size_t i = 0; i < state->m_links.size(); ++i )
        {
            nlink *plink = &state->m_links[i];
            pnode *prevnode = plink->m_prev;
            m_reducepath.push_back(plink->m_t);
            do_reductions(prevnode, count-1, p, action, reduceid);
            m_reducepath.pop_back();
        }
        return;
    }

    bool bAccept = (action & ATypeAccept) ? true : false;
    bool bReject = (action & ATypeReject) ? true : false;
    if( bAccept )
    {
        if( bReject )
            return;
        parsetree *t = reduce_nodes(p, reduceid);
        if( m_cull_nonambig )
        {
            if( t->HasAmbiguity() )
                m_accepttree = merge_nodes(m_accepttree, t);
            else
                t->DecRef();
        }
        else
            m_accepttree = merge_nodes(m_accepttree, t);
        return;
    }
    queried_actions gotoaction(m_g, state->m_state, GTypeProduction, p->GetIndex(), ATypeShift);
    queried_actions endaction;
    if( gotoaction == endaction )
        return;
    parsetree *t = reduce_nodes(p, reduceid);
    // This is where the reject-if would go.
    reducer(bReject, state, gotoaction->m_shiftto->m_number, t);
}

void glr_parser::update_start()
{
    m_for_shifter.clear();
}

void glr_parser::update_step(int nexttok)
{
    m_for_actor = m_activestacks;
    m_for_actor_delayed.clear();
    m_nexttok = nexttok;

    // Do all the reductions first, that way when we shift in the next step we can pick up
    // all shift transitions on the current input symbol, whether they originated in the current state
    // set or were added due to a reduce in this step.
    while( m_for_actor.size() || m_for_actor_delayed.size() )
    {
        if( m_for_actor.empty() )
        {
            m_for_actor.push_back(m_for_actor_delayed.back());
            m_for_actor_delayed.pop_back();
        }
        for( size_t i = 0; i < m_for_actor.size(); ++i )
        {
            pnode *state = m_for_actor[i];
            state->m_processed = true;
            if( state->m_reject )
                continue;
            queried_actions curaction(m_g, state->m_state, GTypeSymbol, m_nexttok, (AType)(ATypeShift|ATypeReduce|ATypeAccept|ATypeReject));
            queried_actions endaction;
            if( curaction == endaction )
                continue;
            for( ; curaction != endaction; ++curaction )
            {
                switch(curaction->m_atype)
                {
                case ATypeShift:
                    {
                        lrshift shft;
                        shft.m_prevstate = state;
                        state->AddRef();
                        shft.m_shiftto = curaction->m_shiftto->m_number;
                        shft.m_token = m_nexttok;
                        m_for_shifter.push_back(shft);
                        break;
                    }
                case ATypeReduce:
                case ATypeAccept:
                case ATypeReject:
                case ATypeReduce|ATypeAccept:
                case ATypeReduce|ATypeReject:
                case ATypeAccept|ATypeReject:
                case ATypeReduce|ATypeAccept|ATypeReject:
                    {
                        m_reducepath.clear();
                        Production * p = curaction->m_reduceby;
                        int reduceid = make_reduce_id(state, *curaction, p);
                        do_reductions(state, curaction->m_reduceby->m_items.size(), p, curaction->m_atype, reduceid);
                        break;
                    }
                }
            }
        }
        m_for_actor.clear();
    }
}

void glr_parser::update_shifters()
{
    // Shift for all the states, including states we just picked up due to reductions.
    m_for_actor_delayed.clear();
    for( vector<lrshift>::iterator curshift = m_for_shifter.begin(), endshift = m_for_shifter.end(); curshift != endshift; ++curshift )
    {
        if( curshift->m_prevstate->m_reject )
        {
            // Skipped shift due to later reject.
            curshift->m_prevstate->DecRef();
            continue;
        }
        pnode *nextstate = FindOrAddState(m_for_actor_delayed, curshift->m_shiftto, 0);
        drop_rejected_paths(nextstate);
        nlink *plink = FindOrAddLink(nextstate, curshift->m_prevstate, false, 0);
        if( ! plink->m_t )
            plink->m_t = terminal_node(curshift->m_token);
        curshift->m_prevstate->DecRef();
    }
}

void glr_parser::update_end()
{
    for( vector<pnode*>::iterator curstate = m_activestacks.begin(), endstate = m_activestacks.end(); curstate != endstate; ++curstate )
        (*curstate)->DecRef();
    m_activestacks.clear();
    m_activestacks = m_for_actor_delayed;
    m_for_actor.clear();
    m_for_actor_delayed.clear();
    m_nexttok = -1;
}

void glr_parser::update_states(int nexttok)
{
    m_cull_nonambig = false;
    update_start();
    update_step(nexttok);
    update_shifters();
    update_end();
}

void glr_parser::update_states(const vector<int> &nexttoks)
{
    m_cull_nonambig = true;
    update_start();
    for( vector<int>::const_iterator curtok = nexttoks.begin(), endtok = nexttoks.end(); curtok != endtok; ++curtok )
    {
        int nexttok = *curtok;
        update_step(nexttok);
    }
    update_shifters();
    update_end();
}

// Get the list of symbols that come next.
// Try to minimize the effective nondeterminism.
void glr_parser::next_symbols(hashset<int> &symbols)
{
    for( vector<pnode*>::iterator curstate = m_activestacks.begin(), endstate = m_activestacks.end(); curstate != endstate; ++curstate )
    {
        queried_actions curaction(m_g, (*curstate)->m_state, GTypeSymbol, -1, (AType)(ATypeShift|ATypeReduce|ATypeAccept|ATypeReject));
        queried_actions endaction;
        for( ; curaction != endaction; ++curaction )
        {
            LRAction &action = *curaction;
            int s = action.m_s->GetIndex();
            symbols.insert(s);
        }
    }
}

static int cmpfirst(const pair<int,int> *lhs, const pair<int,int> *rhs)
{
    return lhs->first - rhs->first;
}

static int cmpsecond(const pair<int,int> *lhs, const pair<int,int> *rhs)
{
    return rhs->second - lhs->second;
}

static void walk_ambiguous_paths(parsetree *accepttree, vector<amgmap> &ambiguous_reductions, amgpath &curpath)
{
    if( curpath.m_amgnodes.size() == 0 )
    {
        if( curpath.m_mappings.size() )
            ambiguous_reductions.push_back(curpath.m_mappings);
        return;
    }
    int permutations = 1;
    vector<int> divs;
    for( size_t i = 0, n = curpath.m_amgnodes.size(); i < n; ++i )
    {
        int div = curpath.m_amgnodes[i]->m_subtrees.size();
        divs.push_back(div);
        permutations *= div;
    }
    for( int permutation = 0; permutation < permutations; ++permutation )
    {
        amgpath nextpath;
        nextpath.m_mappings = curpath.m_mappings;
        int i = permutation;
        int div = 0, mod = 0, choice = 0;
        for( int place = 0, places = divs.size(); place < places; ++place )
        {
            parsetree *ppath = curpath.m_amgnodes[place];
            int div = divs[place];
            choice = i % div;
            i /= div;
            parsetree *subpath = ppath->m_subtrees[choice];
            nextpath.m_mappings[ppath] = choice;
            subpath->collect_ambiguous_nodes_shallow(nextpath.m_amgnodes);
        }
        walk_ambiguous_paths(accepttree, ambiguous_reductions, nextpath);
    }
}

String time_elapsed_string(double sec)
{
#ifdef _WIN32
    switch( _fpclass(sec) )
    {
    case _FPCLASS_SNAN:
    case _FPCLASS_QNAN:
        return "?";
    case _FPCLASS_PINF:
        return "Forever";
    case _FPCLASS_NINF:
        return "Ages ago";
    default:
        break;
    }
#else
    if( sec > FLT_MAX )
      return "Forever";
    else if( sec < -FLT_MAX )
      return "Ages ago";
    else if( sec != sec )
      return "?";
#endif
    String ret;
    bool bNeg = false;
    if( sec < 0 )
    {
        sec *= -1.0;
        bNeg = true;
    }
    double secs = sec;
    int years = (int)floor(secs / (365.0*24.0*60.0*60.0));
    secs -= (double)years * (365.0*24.0*60.0*60.0);
    int days = (int)floor(secs / (24.0*60.0*60.0));
    secs -= (double)days * (24.0*60.0*60.0);
    int hours = (int)floor(secs / (60.0*60.0));
    secs -= (double)hours * (60.0*60.0);
    int mins = (int)floor(secs / 60.0);
    secs -= (double)mins * 60.0;
    int isec = (int)floor(secs);
    if( years )
    {
        char tmbuf[512];
        sprintf(tmbuf, "%d y, %d d, %02d:%02d:%02d", years, days, hours, mins, isec);
        ret = tmbuf;
    }
    else if( days )
    {
        char tmbuf[512];
        sprintf(tmbuf, "%d d, %02d:%02d:%02d", days, hours, mins, isec);
        ret = tmbuf;
    }
    else
    {
        char tmbuf[512];
        sprintf(tmbuf, "%02d:%02d:%02d", hours, mins, isec);
        ret = tmbuf;
    }
    if( bNeg )
        ret += " ago";
    return ret;
}

bool glr_parser::do_ambiguities_search(searchinfo &info, int depth)
{
    if( info.m_poptions->m_bMaxDepth && depth >= info.m_poptions->m_iMaxDepth )
        return true;
    if( info.m_poptions->m_bMaxPaths && info.m_pathsprocessed >= info.m_poptions->m_iMaxPaths )
        return true;
    if( depth == 0 )
    {
        srand( (unsigned int)time(NULL) );
        m_cull_nonambig = false;
        info.m_endmarkerid = m_g->m_endmarker->GetIndex();
        info.m_starttime = GetSystemTicks();
    }

    bool bContinueSearch = true;
    bool bSearchSubnodes = true;
    hashset<int> nextsymbols;
    info.m_sgen->next_symbols(nextsymbols);

    if( nextsymbols.size() == 0 )
        return true;

    if( info.m_poptions->m_bProbabalisticSearch )
    {
        // Pick a few.  Never throw out the terminal token, make sure there is at least one besides the terminal.
        int endmarkerid = info.m_endmarkerid;
        size_t picks = nextsymbols.size() * info.m_poptions->m_iPercent / 100;
        if( nextsymbols.size() * info.m_poptions->m_iPercent % 100 )
            ++picks;
        vector< pair<int,int> > probindices;
        for( hashset<int>::iterator cursymbol = nextsymbols.begin(), endsymbol = nextsymbols.end(); cursymbol != endsymbol; ++cursymbol )
        {
            if( *cursymbol == endmarkerid )
                probindices.push_back( pair<int,int>(endmarkerid, -1) );
            else
                probindices.push_back( pair<int,int>(*cursymbol, rand()) );
        }
        probindices.sort(cmpsecond);
        if( probindices[0].first == endmarkerid && picks < nextsymbols.size() )
            ++picks;
        probindices.resize(picks);
        probindices.sort(cmpfirst);
        nextsymbols.clear();
        for( size_t i = 0; i < picks; ++i )
            nextsymbols.insert(probindices[i].first);
    }

    info.m_fanouts.push_back(nextsymbols.size());
    info.m_pos.push_back(0);

    uint64_t now = GetSystemTicks();
    if( info.m_pathsprocessed && TicksToSeconds(now-info.m_lastupdate) >= 1.0 )
    {
        if( ! info.m_poptions->m_icontinue )
        {
            ob_puts(info.m_poptions->m_outputstream, "Search Cancelled\n");
            return false;
        }
        double estpct = 0.0;
        for( int i = info.m_pos.size()-1; i >= 0; --i )
            estpct = ((double)info.m_pos[i] / (double)info.m_fanouts[i]) + (estpct / (double)info.m_fanouts[i]);
        uint64_t elapsed = now - info.m_starttime;
        double elapsedsec = TicksToSeconds(elapsed);
        double totalsearchsec = elapsedsec / estpct;
        double remainingsec = totalsearchsec - elapsedsec;
        String str;
        String strelapsed = time_elapsed_string(elapsedsec);
        String strremaining = time_elapsed_string(remainingsec);
        char buf[512];
        sprintf(buf, "Searched %" SCNu64 " paths. %s elapsed, ~%5.2f%% done, %s remaining...", info.m_pathsprocessed, strelapsed.c_str(), estpct*100.0, strremaining.c_str());
        str = buf;
        info.m_poptions->m_statusupdate(str.c_str());
        info.m_lastupdate = now;
    }

    for( hashset<int>::iterator cursymbol = nextsymbols.begin(), endsymbol = nextsymbols.end(); cursymbol != endsymbol; ++cursymbol )
    {
        vector<pnode*> curstate = m_activestacks;
        for( vector<pnode*>::iterator curpnode = curstate.begin(), endpnode = curstate.end(); curpnode != endpnode; ++curpnode )
            (*curpnode)->AddRef();
        int symbol = *cursymbol;
        int nextreduceid = m_nextreduceid;
        info.m_pathsprocessed++;
        info.m_acceptstring.push_back(symbol);
        if( bContinueSearch && info.m_poptions->m_bMaxPaths && info.m_pathsprocessed >= info.m_poptions->m_iMaxPaths )
            bSearchSubnodes = false;
        update_states(symbol);
        
        if( m_accepttree )
        {
            amgpath curpath;
            vector<amgmap> ambiguous_reductions;
            m_accepttree->collect_ambiguous_nodes_shallow(curpath.m_amgnodes);
            walk_ambiguous_paths(m_accepttree, ambiguous_reductions, curpath);
            run_reduce_tests(ambiguous_reductions);
                
            if( ambiguous_reductions.size() > 1 )
            {
                ob_printf(info.m_poptions->m_outputstream, "Found %d ways to accept string [", ambiguous_reductions.size());
                for( vector<int>::const_iterator curs = info.m_acceptstring.begin(), ends = info.m_acceptstring.end(); curs != ends; )
                {
                    m_g->m_symbols[*curs]->Print(info.m_poptions->m_outputstream);
                    if( ++curs != ends )
                        ob_putc(info.m_poptions->m_outputstream, ' ');
                }
                ob_printf(info.m_poptions->m_outputstream, "] length %d (including terminator)\n", info.m_acceptstring.size());
                if( info.m_poptions->m_bShowBigTree )
                    m_accepttree->print_symbols(m_g, info.m_poptions->m_outputstream, 0);
                if( info.m_poptions->m_bShowLilTrees )
                {
                    for( vector<amgmap>::iterator curmap = ambiguous_reductions.begin(), endmap = ambiguous_reductions.end(); curmap != endmap; ++curmap )
                        m_accepttree->print_symbols(m_g, info.m_poptions->m_outputstream, curmap);
                }
                ob_flush(info.m_poptions->m_outputstream);
                if( info.m_poptions->m_bStopAtFirstAmbiguity )
                    bContinueSearch = false;
            }
            m_accepttree->DecRef();
            m_accepttree = 0;
        }

        // Also search for longer accept strings.
        if( bContinueSearch && bSearchSubnodes && m_activestacks.size() )
        {
            bContinueSearch = do_ambiguities_search(info, depth+1);
            info.m_pos.back()++;
        }

        for( vector<pnode*>::iterator curpnode = m_activestacks.begin(), endpnode = m_activestacks.end(); curpnode != endpnode; ++curpnode )
            (*curpnode)->DecRef();
        m_activestacks = curstate;
        info.m_acceptstring.pop_back();
        m_nextreduceid = nextreduceid;
        if( ! bContinueSearch || ! bSearchSubnodes )
            break;
    }

    info.m_fanouts.pop_back();
    info.m_pos.pop_back();

    if( depth == 0 )
    {
        info.m_endtime = GetSystemTicks();
        double totaltime = TicksToSeconds(info.m_endtime-info.m_starttime);
        String strtotaltime = time_elapsed_string(totaltime);
        ob_printf(info.m_poptions->m_outputstream, "Searched %I64d paths, took %s\n", info.m_pathsprocessed, strtotaltime.c_str());
    }

    if( ! bContinueSearch )
        return false;
    return true;
}

// Remove duplicate parse trees that result from conditional reductions from the forest, they
// do not genuinely contribute ambiguities, since only one action or the other is ever used.
void glr_parser::run_reduce_tests(vector<amgmap> &ambiguous_reductions)
{
    if( ambiguous_reductions.size() <= 1 )
        return; // don't bother.
    // Previously, each choice in a conflicting multiple conditional reduction was assigned an id.
    // Without further ambiguities, any given _combination_ of conditional reduction ids will happen
    // only once, and we can remove that parse tree from the parse forest.
    // Otherwise, there was an ambiguity of another sort and we leave the parse tree in.
    hashmap< vector<int>, vector<int> > reduction_ids_map;
    vector<int> reduction_ids;
    for( size_t i = 0, n = ambiguous_reductions.size(); i < n; ++i )
    {
        reduction_ids.clear();
        m_accepttree->collect_reduction_ids(ambiguous_reductions[i], reduction_ids);
        if( reduction_ids.size() == 0 )
            continue;
        reduction_ids.sort(simplecmp<int>);
        reduction_ids_map[reduction_ids].push_back(i);
    }
    vector<int> elim;
    for( hashmap< vector<int>, vector<int> >::iterator curids = reduction_ids_map.begin(), endids = reduction_ids_map.end(); curids != endids; ++curids )
    {
        if( curids->second.size() > 1 )
            continue;
        elim.push_back(curids->second[0]);
    }
    // Remove conditional ambiguities in reverse index order.
    elim.sort(simplecmp<int>);
    elim.reverse();
    for( size_t i = 0, n = elim.size(); i < n; ++i )
    {
        int e = elim[i];
        ambiguous_reductions.erase(ambiguous_reductions.begin()+e);
    }
}

void glr_parser::find_and_report_ambiguities(AmbiguitiesOptions *poptions)
{
    poptions->m_statusupdate("Searching For Ambiguities...");
    symbolgenerator_glrparser sgen(this);
    searchinfo info(poptions, &sgen);
    do_ambiguities_search(info, 0);
    poptions->m_statusupdate("Completed Search.");
}

void glr_parser::find_and_report_ambiguities_in_string(AmbiguitiesOptions *poptions, const vector<int> &symbols)
{
    poptions->m_statusupdate("Searching For Ambiguities...");

    ob_puts(poptions->m_outputstream, "Checking input string [");
    for( vector<int>::const_iterator cursymbol = symbols.begin(), endsymbol = symbols.end(); cursymbol != endsymbol; )
    {
        m_g->m_symbols[*cursymbol]->Print(poptions->m_outputstream);
        if( ++cursymbol != endsymbol )
            ob_putc(poptions->m_outputstream, ' ');
    }
    ob_puts(poptions->m_outputstream, "] for ambiguities.\n");

    symbolgenerator_string sgen(symbols);
    searchinfo info(poptions, &sgen);
    do_ambiguities_search(info, 0);
    poptions->m_statusupdate("Completed Search.");
}

// There is absolutely no silver bullet when it comes to ambiguities checking.
// This approach is to methodically test legal input strings (or a probabalistic
// subset) and feed them to the GLR algorithm, which can detect ambiguities, and
// then sort out the real ambiguities after the fact.
void CheckForAmbiguities(Grammar *g, AmbiguitiesOptions *poptions)
{
    glr_parser parser(g);
    poptions->m_bShowLilTrees = false;
    poptions->m_icontinue = 1;
    parser.find_and_report_ambiguities(poptions);
}

void CheckForAmbiguitiesInString(Grammar *g, AmbiguitiesOptions *poptions, const vector<int> &symbols)
{
    glr_parser parser(g);
    poptions->m_bShowLilTrees = true;
    poptions->m_icontinue = 1;
    parser.find_and_report_ambiguities_in_string(poptions, symbols);
}
