// Copyright (c) 2008-2011 Mighty Heave Software LLC.
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
#include "grammar.h"
#include "nfa.h"
#include "hashset.h"
#include "hashmap.h"
#include "emitC.h"
#include "skeleton/c_parser_out.h"

Statement *if_stmt(const char *label, vector<Statement*> *strue, vector<Statement*> *selseifs, vector<Statement*> *selse)
{
    Statement *s = new Statement();
    s->m_stype = IF_STMT;
    s->m_label = label;
    s->m_strue = strue;
    s->m_selseifs = selseifs;
    s->m_selse = selse;
    return s;
}

Statement *for_each_stmt(const char *label, vector<Statement*> *sloop)
{
    Statement *s = new Statement();
    s->m_stype = FOREACH_STMT;
    s->m_label = label;
    s->m_sloop = sloop;
    return s;
}

Statement *subst_stmt(int tok)
{
    Statement *s = new Statement();
    s->m_stype = SUBST_STMT;
    s->m_tok = tok;
    return s;
}

Statement *txt_stmt(const char *txt)
{
    Statement *s = new Statement();
    s->m_stype = TXT_STMT;
    s->m_txt = txt;
    return s;
}

static void addref(hashmap<String,int> &s2i, String &ss)
{
    hashmap<String,int>::iterator iter = s2i.find(ss);
    if( iter == s2i.end() )
        s2i[ss] = 1;
    else
        ++iter->second;
}

static bool decref(hashmap<String,int> &s2i, String &ss)
{
    hashmap<String,int>::iterator iter = s2i.find(ss);
    if( iter == s2i.end() )
        return 0;
    else if(--iter->second == 0)
    {
        s2i.erase(iter);
        return 0;
    }
    return 1;
}

void defines::define(const char *s)
{
    String ss = s;
    if( m_scopes.size() > 0 )
        addref(m_scopes.back(), ss);
    addref(m_defcnt, ss);
}

void defines::undefine(const char *s)
{
    String ss = s;
    if( m_scopes.size() > 0 )
    {
        if( decref(m_scopes.back(), ss) )
            decref(m_defcnt, ss);
        return;
    }
    decref(m_defcnt, ss);
}

int defines::is_defined(String &s)
{
    if( m_defcnt.find(s) == m_defcnt.end() )
        return false;
    return true;
}

void defines::push()
{
    m_scopes.resize(m_scopes.size()+1);
}

void defines::pop()
{
    if( m_scopes.size() == 0 )
        return;
    hashmap<String,int> &r = m_scopes.back();
    for( hashmap<String,int>::iterator cur = r.begin(), end = r.end(); cur != end; ++cur )
    {
        hashmap<String,int>::iterator iter = m_defcnt.find(cur->first);
        if( iter == m_defcnt.end() )
            continue;
        iter->second -= cur->second;
        if( iter->second <= 0 )
            m_defcnt.erase(iter);
    }
    m_scopes.pop_back();
}

int evaluate(vector<Statement*> *stmts, EmitGlobals *globals, defines *defs, outbuf *wr, outbuf *err, outbuf *trace)
{
    if( ! stmts )
        return 1;
    for( vector<Statement*>::iterator curstmt = stmts->begin(), endstmt = stmts->end(); curstmt != endstmt; ++curstmt )
        if( ! (*curstmt)->eval(globals, defs, wr, err, trace) )
            return 0;
    return 1;
}

Statement::Statement() : m_stype(NULL_STMT), m_tok(-1), m_strue(0), m_selseifs(0), m_selse(0), m_sloop(0)
{
}

static void clean(vector<Statement*>* &s)
{
    if( ! s ) return;
    for( vector<Statement*>::iterator cur = s->begin(), end = s->end(); cur != end; ++cur )
        delete *cur;
    delete s;
    s = 0;
}

Statement::~Statement()
{
    clean(m_strue);
    clean(m_selseifs);
    clean(m_selse);
    clean(m_sloop);
}

int Statement::evalloops(EmitGlobals *globals, defines *defs, outbuf *wr, outbuf *err, outbuf *trace)
{
    if( m_label == "parse-state" )
    {
        if( ! m_sloop || ! globals->m_g || globals->m_g->m_lrstates.size() == 0 )
            return 1;
        for( vector<LRStateSet*>::iterator cur = globals->m_g->m_lrstates.begin(), end = globals->m_g->m_lrstates.end();  cur != end; ++cur )
        {
            globals->m_state = *cur;
            if( ! evaluate(m_sloop, globals, defs, wr, err, trace) )
                return 0;
        }
        return 1;
    }
    else if( m_label == "parse-action" )
    {
        if( ! globals->m_state || globals->m_state->m_actions.size() == 0 )
            return 1;
        for( vector<LRAction>::iterator cur = globals->m_state->m_actions.begin(), end = globals->m_state->m_actions.end(); cur != end; ++cur )
        {
            defs->push();
            globals->m_action = cur;
            if( globals->m_action->m_atype == ATypeShift )
                defs->define("shift");
            else if( globals->m_action->m_atype & (ATypeReduce|ATypeAccept) )
                defs->define("reduceaccept");
            else if( globals->m_action->m_atype & (ATypeReject) )
                defs->define("reject");
            if( ! evaluate(m_sloop, globals, defs, wr, err, trace) )
                return 0;
            defs->pop();
        }
        return 1;
    }
    else if( m_label == "tokenizer-state")
    {
        for( vector<OrderedTokenSet*>::iterator cur = globals->m_g->m_TokenSets.begin(), end = globals->m_g->m_TokenSets.end(); cur != end; ++cur )
        {
            globals->m_tokenset = *cur;
            if( ! evaluate(m_sloop, globals, defs, wr, err, trace) )
                return 0;
        }
        return 1;
    }
    else if( m_label == "tokenizer-token")
    {
        if( ! globals->m_tokenset || ! globals->m_tokenset->m_symbols.size() )
            return 1;
        for( vector<Symbol*>::iterator cur = globals->m_tokenset->m_symbols.begin(), end = globals->m_tokenset->m_symbols.end(); cur != end; ++cur )
        {
            defs->push();
            globals->m_s = *cur;
            globals->m_tokensetaction = globals->m_tokenset->m_actions.FindTokenSetAction(globals->m_s);
            if( globals->m_tokensetaction )
            {
                if( globals->m_tokensetaction->m_action == TokenSetAction::PUSH )
                    defs->define("push");
                else if( globals->m_tokensetaction->m_action == TokenSetAction::POP )
                    defs->define("pop");
                else if( globals->m_tokensetaction->m_action == TokenSetAction::GOTO )
                    defs->define("goto");
                else if( globals->m_tokensetaction->m_action == TokenSetAction::IGNOREWS )
                    defs->define("ws");
            }
            if( ! evaluate(m_sloop, globals, defs, wr, err, trace) )
                return 0;
            defs->pop();
        }
        return 1;
    }
    else if( m_label == "lexer-state")
    {
        if( ! globals->m_tokenset || ! globals->m_tokenset->m_bigdfa->m_statecount )
            return 1;
        int scount = globals->m_tokenset->m_bigdfa->m_statecount;
        for( globals->m_lexstate = 0; globals->m_lexstate < scount; ++globals->m_lexstate )
        {
            defs->push();
            if(  globals->m_tokenset->m_bigdfa->m_transitions[globals->m_lexstate].size() )
                defs->define("out-transitions");
            if( ! evaluate(m_sloop, globals, defs, wr, err, trace) )
                return 0;
            defs->pop();
        }
        return 1;
    }
    else if( m_label == "compressed-break" )
    {
        if( ! globals->m_tokenset || ! globals->m_tokenset->m_compressed_dfa_toksetbreaks.size() )
            return 1;
        for( vector<int>::iterator cur = globals->m_tokenset->m_compressed_dfa_toksetbreaks.begin(), end = globals->m_tokenset->m_compressed_dfa_toksetbreaks.end(); cur != end; ++cur )
        {
            globals->m_compressed_break = *cur;
            if( ! evaluate(m_sloop, globals, defs, wr, err, trace) )
                return 0;
        }
        return 1;
    }
    else if( m_label == "compressed-lxitem" )
    {
        if( ! globals->m_tokenset || ! globals->m_tokenset->m_compressed_dfa_data.size() )
            return 1;
        for( vector<compressed_lxitem>::iterator cur = globals->m_tokenset->m_compressed_dfa_data.begin(), end = globals->m_tokenset->m_compressed_dfa_data.end(); cur != end; ++cur )
        {
            globals->m_compressed_lxitem = &*cur;
            if( ! evaluate(m_sloop, globals, defs, wr, err, trace) )
                return 0;
        }
        return 1;
    }
    else if( m_label == "transition-range")
    {
        if( ! globals->m_tokenset || globals->m_lexstate == -1 )
            return 1;
        vector<FrozenDfaTransition> &trns = globals->m_tokenset->m_bigdfa->m_transitions[globals->m_lexstate];
        for(vector<FrozenDfaTransition>::iterator cur = trns.begin(), end = trns.end(); cur != end; ++cur )
        {
            globals->m_range = cur;
            if( ! evaluate(m_sloop, globals, defs, wr, err, trace) )
                return 0;
        }
        return 1;
    }
    else if( m_label == "ctype")
    {
        hashset<String> ctypes;
        for( vector<Symbol*>::iterator cur = globals->m_g->m_symbols.begin(), end = globals->m_g->m_symbols.end(); cur != end; ++cur )
        {
            String ctype = (*cur)->m_CType;
            if( ! ctype.length() || ctypes.find(ctype) != ctypes.end() )
                continue;
            globals->m_ctype = ctype;
            ctypes.insert(ctype);
            if( ! evaluate(m_sloop, globals, defs, wr, err, trace) )
                return 0;
        }
        return 1;
    }
    else if( m_label == "symbol")
    {
        for( vector<Symbol*>::iterator cur = globals->m_g->m_symbols.begin(), end = globals->m_g->m_symbols.end(); cur != end; ++cur )
        {
            defs->push();
            globals->m_s = *cur;
            if( globals->m_s->m_CType.length() )
                defs->define("has-ctype");
            if( ! evaluate(m_sloop, globals, defs, wr, err, trace) )
                return 0;
            defs->pop();
        }
        return 1;
    }
    else if( m_label == "production")
    {
        for( vector<Production*>::iterator cur = globals->m_g->m_productions.begin(), end = globals->m_g->m_productions.end(); cur != end; ++cur )
        {
            defs->push();
            globals->m_p = *cur;
            if( globals->m_p->m_copyof )
                defs->define("copycat");
            globals->m_snip = &globals->m_p->m_ReduceCode;
            if( globals->m_snip->m_code.length() )
                defs->define("hasaction");
            if( ! evaluate(m_sloop, globals, defs, wr, err, trace) )
                return 0;
            defs->pop();
        }
        return 1;
    }
    else if( m_label == "production-symbol")
    {
        if( ! globals->m_p )
            return 1;
        for( vector<Symbol*>::iterator cur = globals->m_p->m_items.begin(), end = globals->m_p->m_items.end(); cur != end; ++cur )
        {
            globals->m_s = *cur;
            if( ! evaluate(m_sloop, globals, defs, wr, err, trace) )
                return 0;
        }
        return 1;
    }
    else if( m_label == "skip")
    {
        if( ! globals->m_p || globals->m_p->m_skips.size() == 0 )
            return 1;
        int lastskp = -1;
        for( vector<SkipRange>::iterator cur = globals->m_p->m_skips.begin(), end = globals->m_p->m_skips.end(); cur != end; ++cur )
        {
            for( int i = cur->m_start, n = cur->m_stop; i < n; ++i )
            {
                if( i <= lastskp )
                    continue;
                lastskp = globals->m_skipindex = i;
                if( ! evaluate(m_sloop, globals, defs, wr, err, trace) )
                    return 0;
            }
        }
        return 1;
    }
    else if( m_label == "compressed-parse-action" )
    {
        if( ! globals->m_state || globals->m_state->m_actionranges.size() == 0 )
            return 1;
        for( vector<LRActionRange>::iterator cur = globals->m_state->m_actionranges.begin(), end = globals->m_state->m_actionranges.end(); cur != end; ++cur )
        {
            defs->push();
            globals->m_actionrange = cur;
            globals->m_action_range_index = cur->m_rangeidx;
            if( globals->m_actionrange->m_atype == ATypeShift )
                defs->define("shift");
            else if( globals->m_actionrange->m_atype & (ATypeReduce|ATypeAccept) )
                defs->define("reduceaccept");
            else if( globals->m_actionrange->m_atype & (ATypeReject) )
                defs->define("reject");
            if( ! evaluate(m_sloop, globals, defs, wr, err, trace) )
                return 0;
            defs->pop();
        }
        return 1;
    }
    else if( m_label == "compressed-action-range" )
    {
        if( ! globals->m_g || ! globals->m_g->m_rangesbyidx.size() )
            return 1;
        globals->m_action_range_index = 0;
        for( vector<UIntSet>::iterator cur = globals->m_g->m_rangesbyidx.begin(), end = globals->m_g->m_rangesbyidx.end(); cur != end; ++cur )
        {
            globals->m_compressed_range = cur;
            globals->m_action_range_size = cur->paircount();
            if( ! evaluate(m_sloop, globals, defs, wr, err, trace) )
                return 0;
            globals->m_action_range_index++;
        }
    }
    else if( m_label == "action-range-item" )
    {
        if( ! globals->m_compressed_range )
            return 1;
        for( UIntSet::iterator cur = globals->m_compressed_range->begin(), end = globals->m_compressed_range->end(); cur != end; ++cur )
        {
            globals->m_action_range_item = &*cur;
            if( ! evaluate(m_sloop, globals, defs, wr, err, trace) )
                return 0;
        }
    }
    else
    {
        ob_printf(err, "unknown for-each loop %s", m_label.c_str());
        return 0;
    }
    return 1;
}

String FriendlyCType(String s)
{
    return s.replace("*", "_ptr").replace("&", "_ref").replace("<", "_lt").replace(">", "_gt");
}

static void itox(String &s, uint32_t c)
{
    if( ! c ) return;
    uint32_t c2 = c & 0x0f;
    itox(s, c2>>4);
    if( c2 < 10 )
        s += ('0'+c2);
    else
        s += ('A'+c2);
}

String SymbolFriendlyName(Symbol *s)
{
    String ret;
    if( s->GetSType() == LITERAL_SYMBOL )
        ret = "LIT";
    else if( s->GetSType() == REGEX_SYMBOL )
        ret = "RX";
    const char *name = s->m_Name.c_str();
    size_t namelen = s->m_Name.length();
    static struct chsub { int c; const char *s; } chsubs[] = {
        {0, "NUL"},
        {'!', "EX"},
        {'@', "AT"},
        {'#', "PND"},
        {'$', "DOL"},
        {'%', "PCT"},
        {'^', "CRT"},
        {'&', "AMP"},
        {'*', "ST"},
        {'(', "LPR"},
        {')', "RPR"},
        {'-', "M"},
        {'+', "P"},
        {'=', "EQ"},
        {'[', "LBR"},
        {']', "RBR"},
        {'{', "LCB"},
        {'}', "RCB"},
        {'|', "VSEP"},
        {'\\', "BSL"},
        {':', "COL"},
        {';', "SCOL"},
        {'\'', "SQT"},
        {'\"', "DQT"},
        {'<', "LT"},
        {'>', "GT"},
        {',', "COM"},
        {'.', "PRD"},
        {'/', "FSL"},
        {'?', "Q"},
        {'~', "TLD"},
        {'`', "BKQT"}
    };
    for( size_t i = 0; i < namelen; ++i )
    {
        uint32_t c = name[i];
        if( c >= 'a' && c <= 'z' )
            ret += c-'a'+'A';
        else if( (c >= 'A' && c <= 'Z') || (c >= '0' && c <= '9') || c == '_')
            ret += c;
        else
        {
            int i = 0;
            for( i = 0; i < sizeof(chsubs)/sizeof(chsubs[0]); ++i )
            {
                if( chsubs[i].c == c )
                {
                    ret += chsubs[i].s;
                    break;
                }
            }
            if( i == sizeof(chsubs)/sizeof(chsubs[0]) )
            {
                ret += "0x";
                itox(ret, c);
            }
        }
    }
    return ret;
}

String ProductionFriendlyName(Production *p)
{
    String ret = SymbolFriendlyName(p->m_nt);
    for( vector<Symbol*>::iterator cur = p->m_items.begin(), end = p->m_items.end(); cur != end; ++cur )
    {
        ret += '_';
        ret += SymbolFriendlyName(*cur);
    }
    return ret;
}

int DoNT(const char *sec, EmitGlobals *globals, defines *defs, outbuf *wr, outbuf *err, outbuf *trace)
{
    globals->m_sindex = 0;
    globals->m_s = globals->m_p->m_nt;
    int ret = globals->m_sk->emit(sec, globals, defs, wr, err, trace);
    globals->m_s = 0;
    globals->m_sindex = -1;
    return ret;
}

int DoSYM(const char *sec, const char *sym, EmitGlobals *globals, defines *defs, outbuf *wr, outbuf *err, outbuf *trace)
{
    const char *sindexstr = sym;
    while( *sindexstr && (*sindexstr < '0' || *sindexstr > '9') )
        ++sindexstr;
    globals->m_sindex = glr_atoi(sindexstr, 0, 0);
    if( globals->m_sindex < 1 || globals->m_sindex > (int)globals->m_p->m_items.size() )
    {
        const char *file = globals->m_p->m_ReduceCode.m_file.c_str();
        int line = globals->m_p->m_ReduceCode.m_line;
        ob_printf(err, "error %s:%d: Symbol %s Out of Range", file, line, sym);
        globals->m_sindex = -1;
        return 0;
    }
    globals->m_s = globals->m_p->m_items[globals->m_sindex-1];
    int ret = globals->m_sk->emit(sec, globals, defs, wr, err, trace);
    globals->m_s = 0;
    globals->m_sindex = -1;
    return ret;
}

SkeletonFile::SkeletonFile(Grammar *g) : m_g(g)
{
}

SkeletonFile::~SkeletonFile()
{
    clear();
}

void SkeletonFile::clear()
{
    m_sections.clear();
}

bool SkeletonFile::parse(inbuf *sk, outbuf *err, outbuf *trace)
{
    return Parse_SkeletonFile(sk, err, trace, this) ? true : false;
}

int SkeletonFile::emit(String sectionname, EmitGlobals *globals, defines *defs, outbuf *wr, outbuf *err, outbuf *trace)
{
    hashmap<String,vector<Statement*>*>::iterator iter = m_sections.find(sectionname);
    if( iter == m_sections.end() )
    {
        ob_printf(err, "No such section %s", sectionname.c_str());
        return 0;
    }
    
    vector<Statement*> *stmts = iter->second;
    if( ! stmts )
        return 1;
    globals->m_g = m_g;
    globals->m_sk = this;
    return evaluate(stmts, globals, defs, wr, err, trace);
}

static inbuf *new_roll_inbuf(void *vpdata, size_t size)
{
    return 0;
}

EmitGlobals::EmitGlobals()
{
    m_g = 0;
    m_p = 0;
    m_snip = 0;
    m_state = 0;
    m_action = 0;
    m_actionrange = 0;
    m_tokenset = 0;
    m_tokensetaction = 0;
    m_lexstate = -1;
    m_range = 0;
    m_s = 0;
    m_sindex = -1;
    m_max_production_symbols = 0;
    m_sk = 0;
    m_skipindex = -1;
    m_compressed_range = 0;
    m_action_range_item = 0;
    m_action_range_index = -1;
    m_action_range_size = -1;
    m_compressed_break = -1;
    m_compressed_lxitem = 0;
}

void EmitBuiltinSkeletonFile(outbuf *dst)
{
    ob_write(dst, skstr, sizeof(skstr)-1);
    ob_flush(dst);
}

int emitC(Grammar *g, emitoptions &options)
{
    const char *skfile = "<internal skeleton file>";
    inbuf *skeleton = 0;
    if( options.m_skeletonfile.length() )
    {
        skfile = options.m_skeletonfile.c_str();
        skeleton = file_inbuf(skfile);
    }
    else
        skeleton = string_inbuf(skstr, 1);
    if( ! skeleton )
    {
        ob_printf(g->m_err, "Unable to open skeleton file %s\n", skfile);
        return 0;
    }
    outbuf *out = 0;
    if( g->m_outfile.length() > 0 )
      out = new_file_outbuf(g->m_outfile.c_str());
    else
      out = new_h_outbuf("stdout",H_STDOUT,0,0);
    if( ! out )
    {
        ob_printf(g->m_err, "Unable to open output file %s\n", g->m_outfile.c_str());
        return 0;
    }

    SkeletonFile sk(g);
    EmitGlobals globals;
    defines defs = options.defined_symbols;
    int bRet = 1;

    for( vector<Production*>::iterator cur = g->m_productions.begin(), endp = g->m_productions.end(); cur != endp; ++cur )
    {
        if( (int)(*cur)->m_items.size() > globals.m_max_production_symbols )
            globals.m_max_production_symbols = (int)(*cur)->m_items.size();
    }
    if( g->m_globalctype.length() )
        defs.define("global-ctype");

    if( ! sk.parse(skeleton, g->m_err, g->m_verbose) )
    {
        bRet = 0;
        goto cleanup;
    }

    globals.m_snip = &g->m_StartBlock;
    if( globals.m_snip->m_code.length() )
    {
        if( ! sk.emit("SRC", &globals, &defs, out, g->m_err, g->m_verbose) )
        {
            bRet = 0;
            goto cleanup;
        }
    }

    if( ! sk.emit("MAIN", &globals, &defs, out, g->m_err, g->m_verbose) )
    {
        bRet = 0;
        goto cleanup;
    }

    globals.m_snip = &g->m_EndBlock;
    if( globals.m_snip->m_code.length() )
    {
        if( ! sk.emit("SRC", &globals, &defs, out, g->m_err, g->m_verbose) )
        {
            bRet = 0;
            goto cleanup;
        }
    }

cleanup:
    free_inbuf(skeleton);
    skeleton = 0;
    free_ob(out);
    out = 0;
    return bRet;
}


