// Copyright (c) 2008-2011 Mighty Heave Software LLC.
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
#include "grammar.h"
#include "rx.h"
#include <ctype.h>
#include <limits.h>
#include "nfa.h"
#include "tokensetmerge.h"
#include "chtest.h"
#include <string.h>
#include "chsets.h"
#include <algorithm>
#include "emitC.h"
#include "ints.h"
//#define VERIFY_LR_TABLE_COMPRESSION (1)
//#define VERIFY_SCANNER_COMPRESSION (1)
#ifndef _WIN32
#define stricmp strcasecmp
#endif

CodeSnip::CodeSnip() : m_line(0)
{
}

void CodeSnip::clear()
{
    m_line = 0;
    m_file = "";
    m_code = "";
}

CodeSnip::CodeSnip(String file, int line, String code)
{
    m_file = file;
	m_file = m_file.replace("\\", "/");
    m_line = line;
    m_code = code;
}

CodeSnip::~CodeSnip()
{
}

int CodeSnip::count()
{
    return m_code.length();
}

void ComputeFirstsFromSymbolString(vector<Symbol*>::const_iterator first, vector<Symbol*>::const_iterator end, SymbolSet &firsts, Symbol *epsilon)
{
    for( ; first != end; ++first )
    {
		Symbol *s = *first;
        firsts += s->m_firsts;
        if( firsts & epsilon )
        {
            firsts -= epsilon;
            continue;
        }
        return;
    }
    firsts += epsilon;
}

Assoc::Assoc() : m_parent(0), m_child(0), m_assocstyle(Assoc::NoStyle)
{
}

bool Assoc::operator==(const Assoc &rhs) const
{
    if( m_parent == rhs.m_parent && m_child == rhs.m_child && m_assocstyle == rhs.m_assocstyle && m_assoc_pos == rhs.m_assoc_pos && m_from_series == rhs.m_from_series )
        return true;
    return false;
}

bool Assoc::operator<(const Assoc &rhs) const
{
    int pindex = m_parent->GetIndex();
    int rpindex = rhs.m_parent->GetIndex();
    if( pindex < rpindex )
        return true;
    else if( rpindex < pindex )
        return false;
    int cindex = m_child->GetIndex();
    int rcindex = rhs.m_child->GetIndex();
    if( cindex < rcindex )
        return true;
    else if( rcindex< cindex )
        return false;
    if( m_assocstyle < rhs.m_assocstyle )
        return true;
    else if( rhs.m_assocstyle < m_assocstyle )
        return false;
    if( m_assoc_pos < rhs.m_assoc_pos )
        return true;
    else if( rhs.m_assoc_pos < m_assoc_pos )
        return false;
    if( m_from_series < rhs.m_from_series )
        return true;
    else if( rhs.m_from_series < m_from_series )
        return false;
    return false;
}

LRItem Assoc::GetFrom(int pos) const
{
    if( pos < 0 || pos >= (int)m_from_series.size() )
        return LRItem();
    return m_from_series[0-pos-1];
}

AssocItem::AssocItem() : m_pos(0)
{
}

AssocItem::AssocItem(const Assoc &assoc, size_t pos) : m_assoc(assoc), m_pos(pos)
{
}

bool AssocItem::IsActive() const
{
    return m_pos == m_assoc.m_from_series.size();
}

LRItem AssocItem::NextFrom() const
{
    return m_assoc.GetFrom(m_pos);
}

bool AssocItem::operator==(const AssocItem &rhs) const
{
    if( m_assoc == rhs.m_assoc && m_pos == rhs.m_pos )
        return true;
    return false;
}

bool AssocItem::operator<(const AssocItem &rhs) const
{
    if( m_assoc < rhs.m_assoc )
        return true;
    else if( rhs.m_assoc < m_assoc )
        return false;
    if( m_pos < rhs.m_pos )
        return true;
    else if( rhs.m_pos < m_pos )
        return true;
    return false;
}

void AssocItem::Print(outbuf *pio) const
{
    for( UIntSet::const_iterator currange = m_assoc.m_assoc_pos.begin(); currange != m_assoc.m_assoc_pos.end(); ++currange )
    {
        for( unsigned int curpos = currange->m_min; curpos <= currange->m_max; ++curpos )
        {
            if( ! IsActive() )
                ob_puts(pio, "[Inactive] ");
            switch( m_assoc.m_assocstyle )
            {
                case Assoc::Forbid:
                    ob_puts(pio, "%forbid [");
                    m_assoc.m_child->Print(pio);
                    ob_puts(pio, "]");
                    break;
                case Assoc::Allow:
                    ob_puts(pio, "%allow [");
                    m_assoc.m_child->Print(pio);
                    ob_puts(pio, "]");
                    break;
                case Assoc::AllowAny:
                    ob_puts(pio, "%allow");
                    break;
                case Assoc::ForbidAny:
                    ob_puts(pio, "%forbid");
                    break;
            }
            ob_puts(pio, " @ [");
            m_assoc.m_parent->Print(pio, curpos);
            for( vector<LRItem>::const_iterator fiter = m_assoc.m_from_series.begin(); fiter != m_assoc.m_from_series.end(); ++fiter )
            {
                ob_puts(pio, "] %from [");
                fiter->Print(pio);
            }
            ob_puts(pio, "]\n");
        }
    }
}

AssocByLabel::AssocByLabel()
: m_assoc(Assoc::None), m_lineno(-1)
{
}

AssocByLabel::AssocByLabel(ProductionOrLabel parentlabel, ProductionOrLabel childlabel, Assoc::AssocStyle assocstyle, Assoc::Association assoc, const vector<IntSetRange> &assoc_pos, const vector<LRItem> &from_series, const String fname, int lineno)
: m_assocstyle(assocstyle), m_assoc(assoc), m_parentlabel(parentlabel), m_childlabel(childlabel), m_assoc_pos(assoc_pos), m_fname(fname), m_lineno(lineno)
{
}

OptionalRange::OptionalRange() : m_start(-1), m_stop(-1)
{
}

OptionalRange::OptionalRange(int start, int stop) : m_start(start), m_stop(stop)
{
}

SkipRange::SkipRange(const OptionalRange &rhs) : m_start(rhs.m_start), m_stop(rhs.m_stop)
{
}

vector< vector<SkipRange> > OptionalRange::GetListPermutations(const vector<OptionalRange> &list)
{
	vector< vector<SkipRange> > permutations;
	vector<SkipRange> apermutation;
	permutations.append(apermutation);
	if( list.size() == 0 )
		return permutations;
	vector< vector<SkipRange> > itempermutations;
	vector< vector<SkipRange> > prevpermutations;
	for( vector<OptionalRange>::const_iterator cur = list.begin(); cur != list.end(); ++cur )
	{
		itempermutations = cur->GetPermutations();
		prevpermutations = permutations;
		permutations.resize(0);
		for( vector< vector<SkipRange> >::const_iterator i = prevpermutations.begin(); i != prevpermutations.end(); ++i )
		{
			for( vector< vector<SkipRange> >::const_iterator j = itempermutations.begin(); j != itempermutations.end(); ++j )
			{
				apermutation.resize(0);
				apermutation.append( *i );
				apermutation.append( *j );
				permutations.append( apermutation );
			}
		}
	}
	return permutations;
}

vector< vector<SkipRange> > OptionalRange::GetPermutations() const
{
	vector< vector<SkipRange> > permutations;
	permutations.append( OptionalRange::GetListPermutations(m_optionalranges) );
	vector<SkipRange> apermutation;
	apermutation.append( SkipRange(m_start, m_stop) );
	permutations.append(apermutation);
	return permutations;
}

bool OptionalRange::operator==(const OptionalRange &rhs)
{
    if( m_start == rhs.m_start || m_stop == rhs.m_stop && m_optionalranges == rhs.m_optionalranges )
        return true;
    return false;
}

bool OptionalRange::operator!=(const OptionalRange &rhs)
{
    if( *this == rhs )
        return false;
    return true;
}

PriorityGroup::PriorityGroup() : m_assoc(Assoc::None)
{
}
    
PriorityGroup::PriorityGroup(Assoc::Association assoc) : m_assoc(assoc)
{
}

PriorityGroup::PriorityGroup(Assoc::Association assoc, ProductionLabel *first, ...) : m_assoc(assoc) // NULL Terminated
{
	int count = 1;
    va_list args;
    va_start(args, first);
    while(1)
    {
        ProductionLabel *lbl = va_arg(args, ProductionLabel*);
        if( ! lbl )
            break;
        ++count;
    }
    va_end(args);
	m_labels.reserve(count);
    m_labels.append(first);
    va_start(args, first);
    while(1)
    {
        ProductionLabel *lbl = va_arg(args, ProductionLabel*);
        if( ! lbl )
            break;
        m_labels.append(lbl);
    }
    va_end(args);
}

PrioritySet::PrioritySet() : m_lineno(-1)
{
}

PrioritySet::PrioritySet(String fname, int lineno, vector<PriorityGroup> &groups)  : m_fname(fname), m_lineno(lineno), m_groups(groups)
{
}

PrioritySet &PrioritySet::operator=(const PrioritySet &rhs)
{
    m_groups = rhs.m_groups;
    m_grouppriorities = rhs.m_grouppriorities;
    m_fname = rhs.m_fname;
    m_lineno = rhs.m_lineno;
    return *this;
}

void PrioritySet::FixatePriorityIndices()
{
    // Something to make lookup much faster.
    m_grouppriorities.clear();
    int i = 0;
    for( vector<PriorityGroup>::iterator group = m_groups.begin(); group != m_groups.end(); ++group )
    {
        for( vector<ProductionLabel*>::iterator l = group->m_labels.begin(), endl = group->m_labels.end(); l != endl; ++l )
            m_grouppriorities.insert(*l, i);
        ++i;
    }
}

int PrioritySet::GetPriorityIndex(ProductionLabel *label)
{
    if( m_groups.size() > 0 && m_grouppriorities.size() == 0 )
    {
        // didn't fixate yet.
        for( vector<PriorityGroup>::iterator group = m_groups.begin(); group != m_groups.end(); ++group )
        {
            if( group->m_labels.find(label) != group->m_labels.end() )
                return group - m_groups.begin();
        }
        return -1;
    }
    // much faster way to do the same thing.
    hashmap<ProductionLabel*,int,64>::iterator iter = m_grouppriorities.find(label);
    if( iter == m_grouppriorities.end() )
        return -1;
    return iter->second;
}

ProductionLabel::ProductionLabel(Grammar *g, String Name) : m_g(g), m_Name(Name)
{
}

int ProductionLabel::GetIndex()
{
    return m_g->GetProductionLabelIndex(this);
}

ProductionOrLabel::ProductionOrLabel() : m_ProductionGroupStyle(IsNone), m_l(0), m_p(0)
{
}

ProductionOrLabel::ProductionOrLabel(Production *p) : m_ProductionGroupStyle(IsProduction), m_l(0), m_p(p)
{
}

ProductionOrLabel::ProductionOrLabel(ProductionLabel *l) : m_ProductionGroupStyle(IsLabel), m_l(l), m_p(0)
{
}

vector<Production*> ProductionOrLabel::GetProductions(Grammar *g)
{
    switch( m_ProductionGroupStyle )
    {
    case IsLabel:
    {
        vector<Production*> ps;
        for( vector<Production*>::iterator cur = g->m_productions.begin(); cur != g->m_productions.end(); ++cur )
            if( (*cur)->m_label == m_l )
                ps.push_back(*cur);
        return ps;
    }
    case IsProduction:
    {
        vector<Production*> ps;
        ps.push_back(m_p);
        return ps;
    }
    case IsAny:
        return g->m_productions;
    default:
        break;
    }

    return vector<Production*>();
}

Production::Production(Grammar *g, Symbol *nt, vector<Symbol*> &symbols, vector<sl_decl> &sldecls)
    :   m_g(g), m_nt(nt), m_items(symbols), m_sldecls(sldecls), m_rank(101), m_label(0), m_copyof(0),
        m_reject(false), m_finalized(false), m_bPredeclareItem(true), m_runtime_resolvable(false)
{
    m_nt->m_OwningProductions.append(this);
}

Production::~Production()
{
}

void Production::MarkReachableNonterminalsAndCheckForProductionlessNonterminals(hashset<Nonterminal*> &nonterminals, outbuf *err)
{
    for( vector<Symbol*>::iterator s = m_items.begin(); s != m_items.end(); ++s )
        (*s)->MarkReachableNonterminalsAndCheckForProductionlessNonterminals(nonterminals, err);
}

// Implementation Notes:
// Argument positions are relative child positions, so in E + E -> E,
// the first E is at relative child position 1 (or -2),
// the second E is at relative child position 2 (or -1)
// Use ANY_ASSOCIATION_POSITION (0) to signify that any position is a match.
// Use 1, 2, 3, ... to signify the 1st position (left associative), 2nd position, 3rd position, ...
// Use -1, -2, -3, ... to signify the 1st from right position (right associative), 2nd from right position, 3rd from right position, ...
// If the association exists after the call, returns 1.
// If the association cannot be added (usually indicating position is out of range), returns 0
bool Production::AddAssociation(Production *child, Assoc::AssocStyle assocstyle, Assoc::Association assoc, const vector<IntSetRange> &byposition_positions, const vector<LRItem> &from_series, const String fname, int lineno)
{
    if( assoc == Assoc::None )
        return true; // Quick out.

    Symbol *childnt = 0;

    if( child == 0 )
    {
        // for null child, absolute position must be given.
        if( assoc != Assoc::ByAbsPosition  )
            return false;
        // can only pass null child for 'any' style associations.
        if( assocstyle != Assoc::AllowAny || assocstyle != Assoc::ForbidAny )
            return false;
    }

    if( assoc == Assoc::ByAbsPosition )
    {
        if( byposition_positions.size() == 0 )
            return true; // Easy one.
        // Sanity check AbsPosition values.  Make sure they're in range.  Make sure they're all the same nonterminal.
        Symbol *childnt = 0;
        for( vector<IntSetRange>::const_iterator cur = byposition_positions.begin(), end = byposition_positions.end(); cur != end; ++cur )
        {
	        if( cur->m_min < 0 || cur->m_min >= (int)m_items.size() )
		        return false;
	        if( cur->m_max < 0 || cur->m_max >= (int)m_items.size() )
		        return false;
            if( ! childnt )
            {
                childnt = m_items[cur->m_min];
                if( ! childnt || childnt->GetSType() != NONTERMINAL_SYMBOL )
                    return false;
            }
            for( int i = cur->m_min; i <= cur->m_max; ++i )
                if( m_items[i] != childnt )
                    return false;
        }

        if( assocstyle == Assoc::AllowAny || assocstyle == Assoc::ForbidAny )
            child = 0;

        vector<Assoc>::iterator passoc = m_g->FindOrCreateAssociation(this, child, assocstyle, from_series, fname, lineno);

        for( vector<IntSetRange>::const_iterator cur = byposition_positions.begin(); cur != byposition_positions.end(); ++cur )
            passoc->m_assoc_pos += UIntSetRange(cur->m_min, cur->m_max);

        return true;
    }
    else
    {
	    // Collect the possible positions for nesting the child.
        Symbol *childnt = child->m_nt;
        vector<Symbol*> symbols;
        vector<int> positions;
        for( vector<Symbol*>::iterator s = m_items.begin(); s != m_items.end(); ++s )
        {
            if( *s != childnt )
                continue;
		    positions.append(s-m_items.begin());
        }

	    // No association is possible, for non-associative forbids, this just means we do not have to bother forbidding.
	    if( positions.size() == 0 )
	    {
		    if( assoc == Assoc::Non )
			    return true;
		    else
			    return false;
	    }

        vector<IntSetRange> byposition;

	    // Get the min and max positions that are forbidden by the various associations.
        if( assoc == Assoc::Left )
            byposition.push_back( IntSetRange(1,-1) );
        else if( assoc == Assoc::Right )
            byposition.push_back( IntSetRange(0,-2) );
        else if( assoc == Assoc::Non )
            byposition.push_back( IntSetRange(0,-1) );
        else
            byposition = byposition_positions;

	    // Adjust the range and sanity check.
        for( vector<IntSetRange>::iterator cur = byposition.begin(); cur != byposition.end(); ++cur )
        {
            if( cur->m_min < 0 )
		        cur->m_min += positions.size();
	        if( cur->m_min < 0 || cur->m_min >= (int)positions.size() )
		        return false; // Cannot fulfill the requested association forbid.
	        if( cur->m_max < 0 )
		        cur->m_max += positions.size();
	        if( cur->m_max < 0 || cur->m_max >= (int)positions.size() )
		        return false; // Cannot fulfill the requested association forbid.
        }

        if( assocstyle == Assoc::AllowAny || assocstyle == Assoc::ForbidAny )
            child = 0;

        vector<Assoc>::iterator passoc = m_g->FindOrCreateAssociation(this, child, assocstyle, from_series, fname, lineno);

        for( vector<IntSetRange>::const_iterator cur = byposition.begin(); cur != byposition.end(); ++cur )
        {
	        int minpos = positions[cur->m_min];
	        int maxpos = positions[cur->m_max];
            passoc->m_assoc_pos += UIntSetRange(minpos, maxpos);
        }

        return true;
    }

    return true;
}

void Production::AddAssociationByLabel(ProductionLabel *childlabel, Assoc::AssocStyle assocstyle, Assoc::Association assoc, const vector<IntSetRange> &byposition_positions, const vector<LRItem> &from_series, const String fname, int lineno)
{
    m_g->AddAssociationsByLabel(this, childlabel, assocstyle, assoc, byposition_positions, from_series, fname, lineno);
}

void Production::Print(outbuf *pio, int position) const
{
    m_nt->Print(pio);
    ob_puts(pio, " : ");
    for( vector<Symbol*>::const_iterator cursym = m_items.begin(); cursym != m_items.end(); ++cursym )
    {
        if( position == 0 )
		{
            ob_putc(pio, '.');
			position = -1;
		}
		else if( cursym != m_items.begin() )
            ob_putc(pio, ' ');
        Symbol *s = *cursym;
        s->Print(pio);
		--position;
    }
    if( position == 0 )
	{
        ob_putc(pio, '.');
		position = -1;
	}
}

bool Production::ChildAllowedAtPosition(const LRStateSet &lrstate, Production *child, int position) const
{
	if( child->m_copyof )
		child = child->m_copyof;
	if( m_copyof )
	{
		int adjustedposition = position;
		for( vector<SkipRange>::const_iterator cur = m_skips.begin(); cur != m_skips.end(); ++cur )
		{
			if( adjustedposition < cur->m_start )
				break;
			adjustedposition += (cur->m_stop-cur->m_start);
		}
		return m_copyof->ChildAllowedAtPosition(lrstate, child, adjustedposition);
	}
	// The nonterminal has to appear here.
	if( position >= (int)m_items.size() || child->m_nt != m_items[position] )
		return false;
    // Check if we override other associations with an allow any here.
    const Assocs *p1stOrderAnyItems = m_g->GetAssocs(this, 0);
    if( p1stOrderAnyItems && p1stOrderAnyItems->Has1stOrderAssoc(Assoc::AllowAny, position) )
        return true;
    const AssocItems *pAnyItems = lrstate.GetAssocItems(this, 0);
    if( pAnyItems && pAnyItems->HasActiveAssocItem(Assoc::AllowAny, position) )
        return true;
    // Check if we override other associations with an allow here.
    const Assocs *p1stOrderChildItems = m_g->GetAssocs(this, child);
    if( p1stOrderChildItems && p1stOrderChildItems->Has1stOrderAssoc(Assoc::Allow, position) )
        return true;
    const AssocItems *pChildItems = lrstate.GetAssocItems(this, child);
    if( pChildItems && pChildItems->HasActiveAssocItem(Assoc::Allow, position) )
        return true;
    // Check for priority conflicts
    if( m_label && child->m_label )
    {
        if( m_g->TestForPriorityConflict(m_label, child->m_label) )
            return false;
    }
    // Check for forbid all associations.
    if( p1stOrderAnyItems && p1stOrderAnyItems->Has1stOrderAssoc(Assoc::ForbidAny, position) )
        return false;
    if( pAnyItems && pAnyItems->HasActiveAssocItem(Assoc::ForbidAny, position) )
        return false;
    // Check for explicitly forbidden associations.
    if( p1stOrderChildItems && p1stOrderChildItems->Has1stOrderAssoc(Assoc::Forbid, position) )
        return false;
    if( pChildItems && pChildItems->HasActiveAssocItem(Assoc::Forbid, position) )
        return false;
    return true;
}

bool Production::IsErrorProduction()
{
    for( vector<Symbol*>::iterator cursymbol = m_items.begin(); cursymbol != m_items.end(); ++cursymbol )
    {
        if( *cursymbol == m_g->m_errorsymbol )
            return true;
    }
    return false;
}

int Production::GrammarIndexToPosition(int index)
{
    // No need to adjust for skips (although inserted ws would be neccessary).
    // This should only be used on indices of fully populated (no skip) productions.
    return index;
}

int Production::PositionToAssocPosition(int index)
{
    if( index < 0 || index >= (int)m_items.size() )
        return -1;
    Symbol *s = m_items[index];
    if( ! s || s->IsTerminal() )
        return -1;
    int pos = -1;
    int curindex = -1;
    while( curindex < index )
    {
        vector<Symbol*>::iterator cur = m_items.find(s, curindex+1);
        if( cur == m_items.end() )
            return -1;
        ++pos;
        curindex = cur - m_items.begin();
    }
    if( curindex == index )
        return pos;
    return -1;
}

void Production::ComputeFirsts(SymbolSet &ntfirsts)
{
    ComputeFirstsFromSymbolString(m_items.begin(), m_items.end(), ntfirsts, m_g->m_epsilon);
}

int Production::GetIndex()
{
    return m_g->GetProductionIndex(this);
}

vector< vector<SkipRange> > Production::GenerateOptionalPartPermutations()
{
	return OptionalRange::GetListPermutations(m_optionalranges);
}

void Production::CollectUsedItems(hashset<Production*> &usedproductions, hashset<Symbol*> &usedsymbols, hashset<OrderedTokenSet*> &usedtokensets)
{
    if( usedproductions.find(this) != usedproductions.end() )
        return;
    usedproductions.insert(this);
    for( vector<Symbol*>::iterator cur = m_items.begin(); cur != m_items.end(); ++cur )
        (*cur)->CollectUsedItems(usedproductions, usedsymbols, usedtokensets);
}

void Production::SetProps(AllProperties *props)
{
    if( ! props )
        return;
    // designed to be called non-null only once.  if you don't follow this, blame yourself.
    if( props->m_flags & PROP_FLAGS_REJECT )
        m_reject = true;
    if( props->m_flags & PROP_FLAGS_RANK )
        m_rank = props->m_rank;
    if( props->m_flags & PROP_FLAGS_ASSOC )
        AddAssociation(this, Assoc::Forbid, props->m_assoc, props->m_assoc_pos, props->m_from_series, props->m_ReduceCode.m_file, props->m_ReduceCode.m_line);
    if( props->m_flags && PROP_FLAGS_LABEL )
        m_label = props->m_label;
    if( props->m_flags & PROP_FLAGS_ACTION )
        m_ReduceCode = props->m_ReduceCode;
    if( props->m_flags & PROP_FLAGS_RUNTIME_RESOLVABLE )
        m_runtime_resolvable = true;
}

void Production::FinalizeDeclaration(int index)
{
    if( ! m_bPredeclareItem )
        return;
    m_bPredeclareItem = true;
    m_g->MoveProduction(this, index);
}

bool Production::IsRuntimeResolvable()
{
    if( m_copyof )
        return m_copyof->IsRuntimeResolvable();
    return m_runtime_resolvable;
}

Symbol *LRItem::GetSymbolAtPosition() const
{
    if( ! m_p )
        return 0;
	if( m_pos < 0 || m_pos >= (int)m_p->m_items.size() )
		return 0;
    return m_p->m_items[m_pos];
}

SymbolSet LRItem::GetFollowsAtPosition()
{
	SymbolSet follows;
	ComputeFirstsFromSymbolString(m_p->m_items.begin()+m_pos, m_p->m_items.end(), follows, m_p->m_g->m_epsilon);
	return follows;
}

bool LRItem::operator==(const LRItem &rhs) const
{
    return m_p == rhs.m_p && m_pos == rhs.m_pos;
}

bool LRItem::operator<(const LRItem &rhs) const
{
    if( m_p != rhs.m_p )
        return m_p->GetIndex() < rhs.m_p->GetIndex();
    return m_pos < rhs.m_pos;
}

int LRItem::cmp_lritem(const LRItem *lhs, const LRItem *rhs)
{
	if( lhs->m_p != rhs->m_p )
        return lhs->m_p->GetIndex() - rhs->m_p->GetIndex();
    return lhs->m_pos - rhs->m_pos;
}

void LRItem::Print(outbuf *pio) const
{
    if( m_p )
        m_p->Print(pio, m_pos);
}

// Follow the goto transitions backwards, then
// the goto transition forward to the items.  Also
// skip any reduced production (child) that conflicts
// with the parent's forid rules.
lritem_ptrset LRItem::GetPostReduceItems() const
{
	lritem_ptrset postreduceitems;
	vector<LRStateSet*> prevstates;
	lritem_ptrset tmp;
	if( m_pos != m_p->m_items.size() )
		return postreduceitems;
	postreduceitems.insert((LRItem*)this);
	// Follow the backwards transitions ...
	for( int i = 0, n = m_pos; i < n; ++i )
	{
		tmp.clear();
		for( lritem_ptrset::iterator curitem = postreduceitems.begin(), enditem = postreduceitems.end(); curitem != enditem; ++curitem )
        {
            lritem_ptrset previtems = (*curitem)->GetPrevItems();
            tmp.insert(previtems.begin(), previtems.end());
        }
		postreduceitems = tmp;
	}
	// ... to figure out which states we reduce to ...
	hashset<LRStateSet*> states;
	for( lritem_ptrset::iterator curitem = postreduceitems.begin(), enditem = postreduceitems.end(); curitem != enditem; ++curitem )
    {
        LRItem *lritem = *curitem;
		states.insert( lritem->m_ss );
    }
	// ... and get the valid lr items that cause the shift in those states ...
	tmp.clear();
	for( hashset<LRStateSet*>::iterator curstate = states.begin(), endstate = states.end(); curstate != endstate; ++curstate )
	{
		for( vector<LRItem>::iterator curitem = (*curstate)->m_items.begin(), enditem = (*curstate)->m_items.end(); curitem != enditem; ++curitem )
		{
			if( ! curitem->ChildAllowedAtPosition(m_p) )
				continue; // Ignore illegal reductions.
			tmp.insert(curitem);
		}
	}
	postreduceitems = tmp;
	// ... and follow the forwards transitions after the reduction.
	tmp.clear();
	for( lritem_ptrset::iterator curitem = postreduceitems.begin(), enditem = postreduceitems.end(); curitem != enditem; ++curitem )
    {
        lritem_ptrset nextitems = (*curitem)->GetNextItems(GTypeProduction, 0, m_p);
		tmp.insert(nextitems.begin(), nextitems.end());
    }
	postreduceitems = tmp;
	return postreduceitems;
}

lritem_ptrset LRItem::GetPostReduceItems_Verbose(outbuf *verbose) const
{
	lritem_ptrset postreduceitems;
	vector<LRStateSet*> prevstates;
	lritem_ptrset tmp;
	if( m_pos != m_p->m_items.size() )
		return postreduceitems;
	postreduceitems.insert((LRItem*)this);
	// Follow the backwards transitions ...
    if( verbose )
        ob_puts(verbose, "Following backwards transitions:\n");
	for( int i = 0, n = m_pos; i < n; ++i )
	{
		tmp.clear();
		for( lritem_ptrset::iterator curitem = postreduceitems.begin(), enditem = postreduceitems.end(); curitem != enditem; ++curitem )
        {
            lritem_ptrset previtems = (*curitem)->GetPrevItems();
            tmp.insert(previtems.begin(), previtems.end());
        }
		postreduceitems = tmp;
        if( verbose )
        {
            ob_printf(verbose, "After %d backward transition:\n", i+1);
	        for( lritem_ptrset::iterator curitem = postreduceitems.begin(), enditem = postreduceitems.end(); curitem != enditem; ++curitem )
            {
                ob_printf(verbose, "State %d, [", (*curitem)->m_ss->m_number);
                (*curitem)->Print(verbose);
                ob_puts(verbose, "]\n");
            }
        }
	}
    if( verbose )
        ob_puts(verbose, "Finished backwards transitions.\n");
	// ... to figure out which states we reduce to ...
	hashset<LRStateSet*> states;
	for( lritem_ptrset::iterator curitem = postreduceitems.begin(), enditem = postreduceitems.end(); curitem != enditem; ++curitem )
    {
        LRItem *lritem = *curitem;
		states.insert( lritem->m_ss );
    }
	// ... and get the valid lr items that cause the shift in those states ...
	tmp.clear();
	for( hashset<LRStateSet*>::iterator curstate = states.begin(), endstate = states.end(); curstate != endstate; ++curstate )
	{
		for( vector<LRItem>::iterator curitem = (*curstate)->m_items.begin(), enditem = (*curstate)->m_items.end(); curitem != enditem; ++curitem )
		{
			if( ! curitem->ChildAllowedAtPosition(m_p) )
            {
                if( verbose )
                {
                    ob_puts(verbose, "Discarding ");
                    ob_printf(verbose, "State %d, [", curitem->m_ss->m_number);
                    curitem->Print(verbose);
                    ob_puts(verbose, "]\n");
                }
				continue; // Ignore illegal reductions.
            }
            if( verbose )
            {
                ob_puts(verbose, "Keeping pre-shift item ");
                ob_printf(verbose, "State %d, [", curitem->m_ss->m_number);
                curitem->Print(verbose);
                ob_puts(verbose, "]\n");
            }
			tmp.insert(curitem);
		}
	}
	postreduceitems = tmp;
	// ... and follow the forwards transitions after the reduction.
	tmp.clear();
	for( lritem_ptrset::iterator curitem = postreduceitems.begin(), enditem = postreduceitems.end(); curitem != enditem; ++curitem )
    {
        lritem_ptrset nextitems = (*curitem)->GetNextItems(GTypeProduction, 0, m_p);
        if( verbose )
        {
            ob_puts(verbose, "Shifting ");
            ob_printf(verbose, " State %d, [", (*curitem)->m_ss->m_number);
            (*curitem)->Print(verbose);
            ob_puts(verbose, "] to");
            for( lritem_ptrset::iterator postitem = nextitems.begin(); postitem != nextitems.end(); ++postitem )
            {
                ob_printf(verbose, " State %d, [", (*postitem)->m_ss->m_number);
                (*postitem)->Print(verbose);
                ob_puts(verbose, "]");
            }
            ob_puts(verbose, "\n");
        }
		tmp.insert(nextitems.begin(), nextitems.end());
    }
	postreduceitems = tmp;
	return postreduceitems;
}

lritem_ptrset LRItem::GetPrevItems()
{
	lritem_ptrset previtems;
	if( m_ss == 0 || m_pos == 0 )
		return previtems; // What previous items?
	Symbol *s = m_p->m_items[m_pos-1];
	for( vector<GotoItem>::iterator curgoto = m_ss->m_backgotos.begin(), endgoto = m_ss->m_backgotos.end(); curgoto != endgoto; ++curgoto )
	{
		if( curgoto->m_gtype == GTypeSymbol )
		{
			if( curgoto->m_s->IsCharSet() && s->IsCharSet() )
			{
				if( (curgoto->m_s->m_firsts & s->m_firsts).count() == 0 )
					continue;
			}
			else if( curgoto->m_s != s )
				continue;
		}
		else // curgoto->m_gtype == GTypeProduction
		{
			if( s->IsTerminal() || curgoto->m_p->m_nt != s )
				continue;
		}
		for( vector<LRItem>::iterator curitem = curgoto->m_to->m_items.begin(), enditem = curgoto->m_to->m_items.end(); curitem != enditem; ++curitem )
		{
			if( curitem->m_p == m_p && curitem->m_pos == m_pos-1 )
				previtems.insert(curitem);
		}
	}
	return previtems;
}

lritem_ptrset LRItem::GetNextItems(GType gtype, Symbol *s, Production *p)
{
	lritem_ptrset nextitems;
	// Don't bother testing for conflicts, assumed that caller passes valid child production.
	if( m_ss == 0 /*|| (gtype == GTypeProduction && ! ChildAllowedAtPosition(p))*/ )
		return nextitems; // Conflicting nesting.
	bool bUseNonterminalForProductions = (gtype == GTypeSymbol) && (s->IsTerminal() == false);
	for( vector<GotoItem>::iterator curgoto = m_ss->m_gotos.begin(), endgoto = m_ss->m_gotos.end(); curgoto != endgoto; ++curgoto )
	{
		if( ( (curgoto->m_gtype == GTypeSymbol) && (curgoto->m_s == s) ) ||
			( (curgoto->m_gtype == GTypeProduction) && 
				( (curgoto->m_p == p) ||
				  (bUseNonterminalForProductions && (curgoto->m_p->m_nt == s) )
				)
			)
		  )
        {
		    for( vector<LRItem>::iterator curitem = curgoto->m_to->m_items.begin(), enditem = curgoto->m_to->m_items.end(); curitem != enditem; ++curitem )
			    if( curitem->m_p == m_p && curitem->m_pos == m_pos+1 )
			        nextitems.insert(curitem);
        }
	}
	return nextitems;
}

// Follow the goto transitions forwards,
// until the end of the line.
lritem_ptrset LRItem::GetReducibleItems() const
{
	lritem_ptrset reducibleitems;
	lritem_ptrset tmp;
	reducibleitems.insert((LRItem*)this);
	for( vector<Symbol*>::iterator cursymbol = m_p->m_items.begin()+m_pos, endsymbol = m_p->m_items.end(); cursymbol != endsymbol; ++cursymbol )
	{
		tmp.clear();
		for( lritem_ptrset::iterator curitem = reducibleitems.begin(), enditem = reducibleitems.end(); curitem != enditem; ++curitem )
        {
            lritem_ptrset nextitems = (*curitem)->GetNextItems(GTypeSymbol, *cursymbol, 0);
			tmp.insert(nextitems.begin(), nextitems.end());
        }
		reducibleitems = tmp;
	}
	return reducibleitems;
}

bool LRItem::ChildAllowedAtPosition(Production *childproduction) const
{
    return m_p->ChildAllowedAtPosition(*m_ss, childproduction, m_pos);
}

GotoItem::GotoItem()
	: m_gtype(GTypeNone), m_s(0), m_p(0), m_to(0)
{
}

GotoItem::GotoItem(GType gtype, Symbol *s, Production *p, LRStateSet *to)
	: m_gtype(gtype), m_s(s), m_p(p), m_to(to)
{
}

bool LRAction::operator==(const LRAction &rhs) const
{
    if(m_state != rhs.m_state)
        return false;
    if(m_gtype != rhs.m_gtype)
        return false;
    if( m_gtype == GTypeSymbol )
    {
        if(m_s != rhs.m_s )
            return false;
    }
    else if( m_p != rhs.m_p )
        return false;
    if( m_atype != rhs.m_atype )
        return false;
    if( m_atype == ATypeShift )
    {
        if(m_shiftto != rhs.m_shiftto )
            return false;
    }
    else if( m_reduceby != rhs.m_reduceby )
        return false;
    return true;
}

bool LRAction::operator<(const LRAction &rhs) const
{
    if( m_state != rhs.m_state )
        return m_state->m_number < rhs.m_state->m_number;
    if( m_gtype != rhs.m_gtype )
        return m_gtype < rhs.m_gtype;
    if( m_gtype == GTypeSymbol )
    {
        if(m_s != rhs.m_s )
            return m_s->GetIndex() < rhs.m_s->GetIndex();
    }
    else if( m_p != rhs.m_p )
        return m_p->GetIndex() < rhs.m_p->GetIndex();
    if( m_atype != rhs.m_atype )
        return m_atype < rhs.m_atype;
    if( m_atype == ATypeShift )
    {
        if(m_shiftto != rhs.m_shiftto )
            return m_shiftto->m_number < rhs.m_shiftto->m_number;
    }
    else if( m_reduceby != rhs.m_reduceby )
        return m_reduceby->GetIndex() < rhs.m_reduceby->GetIndex();
    return false;
}

int LRAction::cmp_lraction(const LRAction *lhs, const LRAction *rhs)
{
    if( lhs->m_state != rhs->m_state )
        return lhs->m_state->m_number - rhs->m_state->m_number;
    if( lhs->m_gtype != rhs->m_gtype )
        return lhs->m_gtype - rhs->m_gtype;
    if( lhs->m_gtype == GTypeSymbol )
    {
        if(lhs->m_s != rhs->m_s )
            return lhs->m_s->GetIndex() - rhs->m_s->GetIndex();
    }
    else if( lhs->m_p != rhs->m_p )
        return lhs->m_p->GetIndex() - rhs->m_p->GetIndex();
    if( lhs->m_atype != rhs->m_atype )
        return lhs->m_atype < rhs->m_atype;
    if( lhs->m_atype == ATypeShift )
    {
        if(lhs->m_shiftto != rhs->m_shiftto )
            return lhs->m_shiftto->m_number - rhs->m_shiftto->m_number;
    }
    if( lhs->m_reduceby == rhs->m_reduceby )
		return 0;
    return lhs->m_reduceby->GetIndex() - rhs->m_reduceby->GetIndex();
}

int LRAction::cmp_actionsymbol(const LRAction *lhs, const LRAction *rhs)
{
    return lhs->symbolidx()-rhs->symbolidx();
}

void LRAction::Print(outbuf *pstream) const
{
    if( m_gtype == GTypeSymbol )
        m_s->Print(pstream);
    else
	{
		ob_putc(pstream, '[');
        m_p->Print(pstream);
		ob_putc(pstream, ']');
	}
    ob_puts(pstream, ": ");
    switch( m_atype )
    {
    case ATypeShift:
        ob_printf(pstream, "Shift to State %d", m_shiftto->m_number);
        break;
    case ATypeReduce:
        ob_printf(pstream, "Reduce by production %d [", m_reduceby->GetIndex());
        m_reduceby->Print(pstream);
        ob_putc(pstream, ']');
        break;
    case ATypeAccept:
        ob_printf(pstream, "Accept by production %d [", m_reduceby->GetIndex());
        m_reduceby->Print(pstream);
        ob_putc(pstream, ']');
        break;
    case ATypeReject:
        ob_printf(pstream, "Reject by production %d [", m_reduceby->GetIndex());
        m_reduceby->Print(pstream);
        ob_putc(pstream, ']');
        break;
    }
}

const char *LRAction::static_ActionType(AType atype)
{
    switch( atype & (ATypeShift|ATypeReduce|ATypeAccept|ATypeReject) )
    {
        case ATypeShift:
            return "shift";
        case ATypeReduce:
            return "reduce";
        case ATypeAccept:
            return "accept";
        case ATypeReject:
            return "reject";
        default:
            break;
    }
    return "???";
}

const char *LRAction::ActionType() const
{
    return static_ActionType(m_atype);
}

bool LRAction::IsRuntimeResolvable() const
{
    if( (m_atype & (ATypeReduce|ATypeAccept|ATypeReject)) == 0 )
        return false;
    if( ! m_reduceby->IsRuntimeResolvable() )
        return false;
    return true;
}

int LRAction::symbolidx() const
{
    if( m_gtype == GTypeSymbol )
        return m_s->GetIndex();
    if( m_gtype == GTypeProduction )
        return m_p->m_g->m_symbols.size()+m_p->GetIndex();
    return -1;
}

int LRAction::actionidx() const
{
    if( m_atype == ATypeShift )
        return m_shiftto->m_number;
    if( m_atype == ATypeReduce || m_atype == ATypeAccept || m_atype ==  ATypeReject )
        return m_reduceby->GetIndex();
    return -1;
}

LRStateSet::LRStateSet(Grammar *g) : m_g(g), m_number(-1)
{
}

bool LRStateSet::operator==(const LRStateSet &rhs) const
{
    return (m_items == rhs.m_items) && (m_associtems == rhs.m_associtems);
}

void LRStateSet::Print(outbuf *pio) const
{
	if( m_number == -1 )
		ob_printf(pio, "State Items:\n", m_number);
	else
		ob_printf(pio, "State %d:\n", m_number);
    for( associtem_map::const_iterator iter0 = m_associtems.begin(); iter0 != m_associtems.end(); ++iter0 )
    {
        for( vector<AssocItem>::const_iterator iter = iter0->second.m_active_allow.begin(); iter != iter0->second.m_active_allow.end(); ++iter )
        {
            if( iter->m_assoc.m_from_series.size() == 0 )
                continue;
            iter->Print(pio);
        }
        for( vector<AssocItem>::const_iterator iter = iter0->second.m_inactive_allow.begin(); iter != iter0->second.m_inactive_allow.end(); ++iter )
        {
            if( iter->m_assoc.m_from_series.size() == 0 )
                continue;
            iter->Print(pio);
        }
        for( vector<AssocItem>::const_iterator iter = iter0->second.m_active_forbid.begin(); iter != iter0->second.m_active_forbid.end(); ++iter )
        {
            if( iter->m_assoc.m_from_series.size() == 0 )
                continue;
            iter->Print(pio);
        }
        for( vector<AssocItem>::const_iterator iter = iter0->second.m_inactive_forbid.begin(); iter != iter0->second.m_inactive_forbid.end(); ++iter )
        {
            if( iter->m_assoc.m_from_series.size() == 0 )
                continue;
            iter->Print(pio);
        }
    }
	for( vector<LRItem>::const_iterator iter = m_items.begin(); iter != m_items.end(); ++iter )
	{
		iter->Print(pio);
		ob_putc(pio, '\n');
	}
}

void LRStateSet::SLRClosure_Verbose(hashset<Production*> &addedproductions, outbuf *verbose)
{
    addedproductions.clear();
    if( m_items.size() == 0 )
        return;
    if( verbose )
		ob_puts(verbose, "Doing SLRClosure on a set.\n");
    for( size_t i = 0; i < m_items.size(); i++ )
    {
        LRItem lritem = m_items[i];
        Symbol *nextsymbol = lritem.GetSymbolAtPosition();
        if( ! nextsymbol || nextsymbol->IsTerminal() )
            continue;
		if( verbose )
		{
			ob_puts(verbose, "Considering expansions of ");
			lritem.Print(verbose);
			ob_putc(verbose, '\n');
		}
        for( vector<Production*>::iterator nextproduction = nextsymbol->m_OwningProductions.begin(), endproduction = nextsymbol->m_OwningProductions.end(); nextproduction != endproduction; ++nextproduction )
        {
            if( addedproductions.find(*nextproduction) != addedproductions.end() )
			{
				if( verbose )
				{
					ob_puts(verbose, "Skipping already present production ");
					(*nextproduction)->Print(verbose);
					ob_putc(verbose, '\n');
				}
                continue;
			}
            if( ! lritem.ChildAllowedAtPosition(*nextproduction) )
			{
				if( verbose )
				{
					ob_puts(verbose, "Child production ");
					(*nextproduction)->Print(verbose);
					ob_puts(verbose, " not allowed at ");
					lritem.Print(verbose);
					ob_putc(verbose, '\n');
				}
                continue;
			}
            addedproductions.insert(*nextproduction);
			LRItem newitem(this, *nextproduction, 0);
			if( verbose )
			{
				ob_puts(verbose, "Added Item ");
				newitem.Print(verbose);
				ob_putc(verbose, '\n');
			}
            m_items.append(newitem);
        }
    }
    // Not safe to remove this.  Don't be fooled by a few good results.
	m_items.unique(LRItem::cmp_lritem);
    if( verbose )
		ob_puts(verbose, "Set Closed.\n");
}

// This function is the cental loop.  Be very very fugral in non-verbose mode.
void LRStateSet::SLRClosure(hashset<Production*> &addedproductions)
{
    addedproductions.clear(); // re-use the same memory for additional speed.
    if( m_items.size() == 0 )
        return;
    for( size_t i = 0; i < m_items.size(); i++ )
    {
        LRItem lritem = m_items[i];
        Symbol *nextsymbol = lritem.GetSymbolAtPosition();
        if( ! nextsymbol || nextsymbol->IsTerminal() )
            continue;
        for( vector<Production*>::iterator nextproduction = nextsymbol->m_OwningProductions.begin(), endproduction = nextsymbol->m_OwningProductions.end(); nextproduction != endproduction; ++nextproduction )
        {
            if( addedproductions.find(*nextproduction) != addedproductions.end() )
                continue;
            if( ! lritem.ChildAllowedAtPosition(*nextproduction) )
                continue;
            addedproductions.insert(*nextproduction);
            m_items.append(LRItem(this, *nextproduction, 0));
        }
    }
    // Probably not safe to remove this.  Don't be fooled by a few good results.
	m_items.unique(LRItem::cmp_lritem);
}

void LRStateSet::AddGotoItem(GType gtype, Symbol *s, Production *p, LRStateSet *to)
{
    GotoItem go;
    go.m_gtype = gtype;
    go.m_s = s;
    go.m_p = p;
    go.m_to = to;
    m_gotos.append(go);
}

void LRStateSet::AddLRItem(Production *p, int position)
{
    m_items.append(LRItem(this, p, position));
}

#ifdef VERIFY_LR_TABLE_COMPRESSION
static void ActionFailReport(int number, Grammar *g, LRAction &action, AType atype, int symbolidx, int actionidx, const char *failure)
{
    g->m_errcount++;
    if( g->m_err )
        ob_printf(g->m_err, "compression error: %s: state %d, Compressed LR table data (atype=%d,symbol=%d,action=%d) does not match original (atype=%d,symbol=%d,action=%d): \n", failure, number, atype, symbolidx, actionidx, action.m_atype, action.symbolidx(), action.actionidx());
}

static bool TestAction(int number, Grammar *g, LRAction &action, AType atype, int symbolidx, int actionidx)
{
    if( action.m_atype != atype )
    {
        ActionFailReport(number, g, action, atype, symbolidx, actionidx, "bad atype");
        return false;
    }
    int actionsymbolidx = action.symbolidx();
    if( actionsymbolidx != symbolidx )
    {
        ActionFailReport(number, g, action, atype, symbolidx, actionidx, "bad symbol");
        return false;
    }
    int actionactionidx = action.actionidx();
    if( actionactionidx != actionidx )
    {
        ActionFailReport(number, g, action, atype, symbolidx, actionidx, "bad action");
        return false;
    }
    return true;
}
#endif

void LRStateSet::GenerateActions(rangemap &ranges, vector<UIntSet> &rangesbyidx)
{
    // Add reduce actions.
    int i = 0;
    for( vector<LRItem>::iterator curitem = m_items.begin(); curitem != m_items.end(); ++curitem, ++i )
    {
        if( curitem->GetSymbolAtPosition() )
            continue;
        Production *p = curitem->m_p;
        AType atype = ATypeNone;
        if( p->m_nt->IsStartSymbol() )
            atype = ATypeAccept;
        else
            atype = p->IsRejectProduction() ? ATypeReject : ATypeReduce;
		if( curitem->m_follows.m_chars.count() )
		{
			Symbol *cset = m_g->FindOrAddCharset(curitem->m_follows.m_chars);
			AddReduceAction(p, i, cset, atype);
		}
		for( sortedset<Symbol*>::const_iterator cur = curitem->m_follows.m_symbols.begin(); cur != curitem->m_follows.m_symbols.end(); ++cur )
            AddReduceAction(curitem->m_p, i, *cur, atype);
    }
    // Add shift actions.
    for( vector<GotoItem>::iterator curitem = m_gotos.begin(); curitem != m_gotos.end(); ++curitem )
        AddShiftAction(curitem->m_gtype, curitem->m_s, curitem->m_p, curitem->m_to);

    // Let's compress the lrtable now...
    m_actions.sort(LRAction::cmp_actionsymbol);
    LRAction firstaction, lastaction;
    lastaction.m_gtype = GTypeNone;
    lastaction.m_atype = ATypeNone;
    UIntSet range;
    m_actionranges.clear();
    for( vector<LRAction>::iterator curaction = m_actions.begin(), endaction = m_actions.end(); curaction != endaction; ++curaction )
    {
        LRAction &action = *curaction;
        int symbolidx = action.symbolidx();
        int actionidx = action.actionidx();
        if( action.m_atype == lastaction.m_atype && actionidx == firstaction.actionidx() && actionidx == lastaction.actionidx() )
        {
            range += symbolidx;
            continue;
        }
        else if(    action.m_atype == lastaction.m_atype
                &&  range.paircount() == 1
                &&  symbolidx == range.begin()->m_max+1
                &&  actionidx == (lastaction.actionidx()+1)
                &&  range.begin()->count() == (lastaction.actionidx()-firstaction.actionidx()+1) )
        {
            lastaction = action;
            range += symbolidx;
            continue;
        }
        else if( lastaction.m_atype != ATypeNone )
        {
            rangemap::iterator rangeiter = ranges.find(range);
            int rangeidx = -1;
            if( rangeiter == ranges.end() )
            {
                rangeidx = rangesbyidx.size();
                rangesbyidx.push_back(range);
                ranges[range] = rangeidx;
            }
            else
                rangeidx = rangeiter->second;
            LRActionRange actioncmd = {this, firstaction.m_atype, firstaction.actionidx(), lastaction.actionidx(), rangeidx};
            m_actionranges.push_back(actioncmd);
        }
        firstaction = action;
        lastaction = action;
        range.clear();
        range += symbolidx;
    }
    if( lastaction.m_atype != ATypeNone )
    {
        rangemap::iterator rangeiter = ranges.find(range);
        int rangeidx = -1;
        if( rangeiter == ranges.end() )
        {
            rangeidx = rangesbyidx.size();
            rangesbyidx.push_back(range);
            ranges[range] = rangeidx;
        }
        else
            rangeidx = rangeiter->second;
        LRActionRange actioncmd = {this, firstaction.m_atype, firstaction.actionidx(), lastaction.actionidx(), rangeidx};
        m_actionranges.push_back(actioncmd);
    }
#ifdef VERIFY_LR_TABLE_COMPRESSION
    vector<LRAction>::iterator testaction = m_actions.begin(), endtestaction = m_actions.end();
    for( vector<LRActionRange>::iterator curactionrange = m_actionranges.begin(), endactionrange = m_actionranges.end(); curactionrange != endactionrange; ++curactionrange )
    {
        if( curactionrange->m_firstactionidx == curactionrange->m_lastactionidx )
        {
            int actionidx = curactionrange->m_firstactionidx;
            UIntSet &range = rangesbyidx[curactionrange->m_rangeidx];
            for( UIntSet::iterator currange = range.begin(), endrange = range.end(); currange != endrange; ++currange )
            {
                for( int symbolidx = currange->m_min, endsymbolidx = currange->m_max; symbolidx <= endsymbolidx; ++symbolidx )
                {
                    if( testaction == endtestaction )
                    {
                        m_g->m_errcount++;
                        if( m_g->m_err )
                            ob_printf(m_g->m_err, "error: state %d, symbol while verifying compressed LR table, out of actions to test\n", m_number);
                        return;
                    }
                    if( ! TestAction(m_number, m_g, *testaction, curactionrange->m_atype, symbolidx, actionidx) )
                        return;
                    ++testaction;
                }
            }
        }
        else
        {
            int minactionidx = curactionrange->m_firstactionidx;
            UIntSet &range = rangesbyidx[curactionrange->m_rangeidx];
            if( range.paircount() != 1 )
            {
                m_g->m_errcount++;
                if( m_g->m_err )
                    ob_printf(m_g->m_err, "error: state %d, Compressed LR tables contains a variant action with more than one range pair.\n", m_number);
                return;
            }
            for( UIntSet::iterator currange = range.begin(), endrange = range.end(); currange != endrange; ++currange )
            {
                int minsymbolidx = currange->m_min;
                for( int symbolidx = currange->m_min, endsymbolidx = currange->m_max; symbolidx <= endsymbolidx; ++symbolidx )
                {
                    if( testaction == endtestaction )
                    {
                        m_g->m_errcount++;
                        if( m_g->m_err )
                            ob_printf(m_g->m_err, "error: state %d, while verifying compressed LR table, out of actions to test\n", m_number);
                        return;
                    }
                    int actionidx = minactionidx + (symbolidx-minsymbolidx);
                    if( ! TestAction(m_number, m_g, *testaction, curactionrange->m_atype, symbolidx, actionidx) )
                        return;
                    ++testaction;
                }
            }
        }
    }
    if( testaction != endtestaction )
    {
        m_g->m_errcount++;
        if( m_g->m_err )
            ob_printf(m_g->m_err, "error: state %d, while verifying compressed LR table, out of compressed actions\n", m_number);
        return;
    }
#endif
    m_actions.sort(LRAction::cmp_lraction);
}

void LRStateSet::AddReduceAction(Production *reduceby, int itemidx, Symbol *s, AType atype)
{
	vector<LRAction>::iterator action = m_actions.newitem();
    action->m_state = this;
    action->m_gtype = GTypeSymbol;
    action->m_s = s;
    action->m_p = 0;
    action->m_atype = atype;
    action->m_reduceby = reduceby;
    action->m_itemidx = itemidx;
}

void LRStateSet::AddShiftAction(GType gtype, Symbol *s, Production *p, LRStateSet *shiftto)
{
	vector<LRAction>::iterator action = m_actions.newitem();
    action->m_state = this;
    action->m_gtype = gtype;
    action->m_s = s;
    action->m_p = p;
    action->m_atype = ATypeShift;
    action->m_shiftto = shiftto;
    action->m_itemidx = -1;
}

const AssocItems *LRStateSet::GetAssocItems(const Production *parent, const Production *child) const
{
    associtem_map::KeyType key((Production*)parent,(Production*)child);
    associtem_map::const_iterator iter = m_associtems.find(key);
    if( iter == m_associtems.end() )
        return 0;
    return &iter->second;
}

bool AssocItems::HasActiveAssocItem(Assoc::AssocStyle assocstyle, int position) const
{
    const sortedset<AssocItem> *passocitems = 0;
    switch( assocstyle )
    {
    case Assoc::Allow:
    case Assoc::AllowAny:
        passocitems = &m_active_allow;
        break;
    case Assoc::Forbid:
    case Assoc::ForbidAny:
        passocitems = &m_active_forbid;
        break;
    default:
        return false;
    }
    for( vector<AssocItem>::const_iterator cur = passocitems->begin(), e = passocitems->end(); cur != e; ++cur )
    {
        if( cur->IsActive() && (cur->m_assoc.m_assoc_pos & position) )
            return true;
    }
    return false;
}

void AssocItems::AddAssocItem(const AssocItem &associtem)
{
    sortedset<AssocItem> *passocitems = 0;
    if( associtem.IsActive() )
    {
        switch( associtem.m_assoc.m_assocstyle )
        {
        case Assoc::Allow:
        case Assoc::AllowAny:
            passocitems = &m_active_allow;
            break;
        case Assoc::Forbid:
        case Assoc::ForbidAny:
            passocitems = &m_active_forbid;
            break;
        default:
            return;
        }
    }
    else
    {
        switch( associtem.m_assoc.m_assocstyle )
        {
        case Assoc::Allow:
        case Assoc::AllowAny:
            passocitems = &m_inactive_allow;
            break;
        case Assoc::Forbid:
        case Assoc::ForbidAny:
            passocitems = &m_inactive_forbid;
            break;
        default:
            return;
        }
    }
    passocitems->insert(associtem);
}

bool AssocItems::operator==(const AssocItems &rhs) const
{
    return m_active_allow == rhs.m_active_allow && m_inactive_allow == rhs.m_inactive_allow && m_active_forbid == rhs.m_active_forbid && m_inactive_forbid == rhs.m_inactive_forbid;
}

bool AssocItems::operator!=(const AssocItems &rhs) const
{
    return ! operator==(rhs);
}

vector<Assoc>::iterator Assocs::FindOrCreateAssociation(Production *parent, Production *child, Assoc::AssocStyle assocstyle, const vector<LRItem> &from_series, String fname, int lineno)
{
    vector<Assoc> *passocs = 0;
    if( from_series.size() == 0 )
    {
        switch( assocstyle )
        {
        case Assoc::Allow:
        case Assoc::AllowAny:
            passocs = &m_1st_order_allow;
            break;
        case Assoc::Forbid:
        case Assoc::ForbidAny:
            passocs = &m_1st_order_forbid;
            break;
        default:
            return 0;
        }
    }
    else
    {
        switch( assocstyle )
        {
        case Assoc::Allow:
        case Assoc::AllowAny:
            passocs = &m_gt_1st_order_allow;
            break;
        case Assoc::Forbid:
        case Assoc::ForbidAny:
            passocs = &m_gt_1st_order_forbid;
            break;
        default:
            return 0;
        }
    }

    for( vector<Assoc>::iterator cur = passocs->begin(); cur != passocs->end(); ++cur )
        if( cur->m_assocstyle == assocstyle && cur->m_from_series == from_series )
            return cur;

    vector<Assoc>::iterator passoc = passocs->newitem();
    passoc->m_parent = parent;
    passoc->m_child = child;
    passoc->m_assocstyle = assocstyle;
    passoc->m_from_series = from_series;
    passoc->m_fname = fname;
    passoc->m_lineno = lineno;
    return passoc;
}

bool Assocs::Has1stOrderAssoc(Assoc::AssocStyle assocstyle, int position) const
{
    const vector<Assoc> *passocs = 0;
    switch( assocstyle )
    {
    case Assoc::Allow:
    case Assoc::AllowAny:
        passocs = &m_1st_order_allow;
        break;
    case Assoc::Forbid:
    case Assoc::ForbidAny:
        passocs = &m_1st_order_forbid;
        break;
    default:
        return false;
    }
    for( vector<Assoc>::const_iterator cur = passocs->begin(); cur != passocs->end(); ++cur )
        if( cur->m_assocstyle == assocstyle && (cur->m_assoc_pos & position) )
            return true;
    return false;
}


Grammar::Grammar()
: m_endmarker(0), m_startsymbol(0), m_errorsymbol(0), m_initialstate(0), m_epsilon(0), m_InitialTokenState(0), m_ntcounter(1),
  m_RequestedGrammarSolution(DeterministicParser), m_bLogDfa(false),
  m_bUsePoundLine(true), m_err(0), m_verbose(0), m_verbosedb(0),
  m_bLogParse(false), m_bLogFirsts(false), m_bLogFollows(false), m_bLogClosure(false),
  m_errcount(0), m_wrncount(0), m_TokenSetMerge(0)
{
	m_startsymbol = new StartSymbol(this);
	m_symbols.append(m_startsymbol);
	m_endmarker = FindOrAddTerminal("$");
	m_epsilon = FindOrAddTerminal("<E>");
    m_errorsymbol = FindOrAddTerminal("<error>");
    m_EmitLanguage = "C";
    m_name = "parser";
    
    m_namedcharsets.push_back( NamedCharSet("[:alpha:]", named_charset_alpha()) );
    m_namedcharsets.push_back( NamedCharSet("[:lower:]", named_charset_lower()) );
    m_namedcharsets.push_back( NamedCharSet("[:upper:]", named_charset_upper()) );
    m_namedcharsets.push_back( NamedCharSet("[:alnum:]", named_charset_alnum()) );
    m_namedcharsets.push_back( NamedCharSet("[:blank:]", named_charset_blank()) );
    m_namedcharsets.push_back( NamedCharSet("[:space:]", named_charset_space()) );
    m_namedcharsets.push_back( NamedCharSet("[:digit:]", named_charset_digit()) );
    m_namedcharsets.push_back( NamedCharSet("[:xdigit:]", named_charset_xdigit()) );
}

Grammar::~Grammar()
{
    m_err = 0;
    m_verbose = 0;
    m_verbosedb = 0;
    m_initialstate = 0;
    m_endmarker = 0;
    m_startsymbol = 0;
    m_epsilon = 0;
    for( vector<Production*>::iterator cur = m_productions.begin(), end = m_productions.end(); cur != end; ++cur )
        delete *cur;
    m_productions.clear();
    m_indexmap.clear();
    m_symbolidmap.clear();
    for( vector<Symbol*>::iterator cur = m_symbols.begin(), end = m_symbols.end(); cur != end; ++cur )
        delete *cur;
    m_symbols.clear();
    m_literals.clear();
    m_regexs.clear();
    m_charsets.clear();
    m_nonterminals.clear();
    m_terminals.clear();
    m_priorities.clear();
    m_InitialTokenState = 0;
    for( vector<LRStateSet*>::iterator cur = m_lrstates.begin(), end = m_lrstates.end(); cur != end; ++cur )
        delete *cur;
    m_lrstates.clear();
    for( vector<ProductionLabel*>::iterator cur = m_labels.begin(), end = m_labels.end(); cur != end; ++cur )
        delete *cur;
    m_labels.clear();
    for( vector<OrderedTokenSet*>::iterator cur = m_TokenSets.begin(), end = m_TokenSets.end(); cur != end; ++cur )
        delete *cur;
    m_TokenSets.clear();
    m_ActionConflicts.clear();
    m_ActionResolvedConflicts.clear();
    delete m_TokenSetMerge;
    m_TokenSetMerge = 0;
    m_assocs.clear();
    m_StartBlock.clear();
    m_EndBlock.clear();
    for( hashmap<String,Rx*>::iterator cur = m_tokfragmap.begin(), end = m_tokfragmap.end(); cur != end; ++cur )
        delete cur->second;
    m_tokfragmap.clear();
}

Nonterminal *Grammar::AddAnonymousNonterminal(OrderedTokenSet *tokstate)
{
    return FindOrAddNonterminal("", tokstate);
}

Nonterminal *Grammar::FindOrAddNonterminal(String name, OrderedTokenSet *tokstate, bool bCreateIfNotFound)
{
    if( name.length() == 0 )
    {
        if( ! bCreateIfNotFound )
            return 0;
        Nonterminal *s = new Nonterminal(this, name, tokstate);
        char buf[64];
        sprintf(buf, "<Nt_%d>", m_ntcounter++);
        s->m_Name = buf;
        m_symbols.append(s);
        m_nonterminals.append(s);
        return s;
    }
    for( vector<Nonterminal*>::iterator cur = m_nonterminals.begin(); cur != m_nonterminals.end(); ++cur )
    {
        Nonterminal *s = *cur;
        if( s->m_Name != name || s->m_TokenState != tokstate )
            continue;
        return s;
    }
    if( ! bCreateIfNotFound )
        return 0;
    Nonterminal *s = new Nonterminal(this, name, tokstate);
    m_symbols.append(s);
    m_nonterminals.append(s);
    return s;
}

Literal *Grammar::FindOrAddLiteral(String name, bool bCaseInsensitive, OrderedTokenSet *tokstate, bool bCreateIfNotFound)
{
    if( bCaseInsensitive )
    {
        const char *mbstr = name.c_str();
        size_t mblen = name.length();
        size_t read_count = 0;
        int err = 0;
        unsigned int c = 0;
        int found_case_char = 0;
        while(mblen)
        {
            c = mbget(mbstr, mblen, &read_count, &err);
            if( err )
                break;
            mbstr += read_count;
            mblen -= read_count;
            if( glr_toupper(c) != glr_tolower(c) )
                continue;
            found_case_char = 1;
            break;
        }
        if( ! found_case_char )
            bCaseInsensitive = false;
    }
    for( vector<Literal*>::iterator cur = m_literals.begin(); cur != m_literals.end(); ++cur )
    {
        Literal *s = *cur;
        if( s->m_TokenState != tokstate )
            continue;
        if( bCaseInsensitive != s->m_bCaseInsensitive )
            continue;
        if( bCaseInsensitive )
        {
            if( stricmp(s->m_Name.c_str(), name.c_str()) != 0 )
                continue;
        }
        else
        {
            if( s->m_Name != name )
                continue;
        }
        return s;
    }
	if( ! bCreateIfNotFound )
		return 0;
    Literal *s = new Literal(this, name, tokstate, bCaseInsensitive);
    m_symbols.append(s);
    m_literals.append(s);
    return s;
}

Regex *Grammar::FindOrAddRx(String name, Rx *rx, bool bCaseInsensitive, OrderedTokenSet *tokstate, bool bCreateIfNotFound)
{
    Rx *finalrx = 0;
    if( rx )
    {
        finalrx = rx->Copy();
        if( bCaseInsensitive )
            finalrx->CaseDesensitize();
    }
    else
    {
        bCreateIfNotFound = false;
    }
    for( vector<Regex*>::iterator cur = m_regexs.begin(); cur != m_regexs.end(); ++cur )
    {
        Regex *s = *cur;
        if( finalrx )
        {
            if( ! s->m_rx->EqualTo(finalrx) )
                continue;
            delete finalrx;
            return s;
        }
        else
        {
            if( s->m_Name == name )
                return s;
        }
    }
	if( ! bCreateIfNotFound )
    {
        delete finalrx;
		return 0;
    }
    Regex *s = new Regex(this, name, tokstate, finalrx, bCaseInsensitive);
    m_symbols.append(s);
    m_regexs.append(s);
    return s;
}

Terminal *Grammar::FindOrAddTerminal(String name)
{
    for( vector<Terminal*>::iterator cur = m_terminals.begin(); cur != m_terminals.end(); ++cur )
    {
        Terminal *s = *cur;
        if( s->m_Name != name )
            continue;
        return s;
    }
    Terminal *s = new Terminal(this, name);
    m_symbols.append(s);
    m_terminals.append(s);
    return s;
}

CharSet *Grammar::FindOrAddCharset(UIntSet chars)
{
    for( vector<CharSet*>::iterator cur = m_charsets.begin(); cur != m_charsets.end(); ++cur )
    {
        CharSet *s = *cur;
        if( s->m_chars != chars )
            continue;
        return s;
    }
    CharSet *s = new CharSet(this, chars);
    m_symbols.append(s);
    m_charsets.append(s);
    return s;
}

// Add a production
// Note: X -> <Start> with no layout is the only acceptable production involving <Start>
Production *Grammar::AddProduction(Symbol *nt, vector<Symbol*> &callsymbols, vector<sl_decl> &callsldecls, AllProperties *props)
{
    vector<Symbol*> symbols;
    vector<sl_decl> sldecls;
	vector<OptionalRange> optionalranges;
    GetSymbolsAndRanges(callsymbols, callsldecls, symbols, sldecls, optionalranges);
    
    return AddProductionCore(nt, symbols, sldecls, optionalranges, props);
}

Production *Grammar::AddProductionCore(Symbol *nt, vector<Symbol*> &symbols, vector<sl_decl> &sldecls, vector<OptionalRange> &optionalranges, AllProperties *props)
{
    if( nt->IsTerminal() )
    {
        m_errcount++;
        if( m_err )
            ob_printf(m_err, "error: Attempted to assign production to terminal %s.\n", nt->m_Name.c_str());
        return 0;
    }

	if( symbols.find(m_startsymbol) != symbols.end() || (nt == m_startsymbol && symbols.size() != 1) )
	{
        m_errcount++;
        if( m_err )
            ob_puts(m_err, "error: <Start> : X ; is the only acceptable production involving <Start>\n");
		return 0;
	}

    Production *p = new Production(this, nt, symbols, sldecls);
    m_indexmap[p] = (int)m_productions.size();
	p->m_optionalranges = optionalranges;
    m_productions.append(p);
    p->SetProps(props);
    return p;
}

void Grammar::GetSymbolsAndRanges(vector<Symbol*> &callsymbols, vector<sl_decl> &callsldecls, vector<Symbol*> &symbols, vector<sl_decl> &sldecls, vector<OptionalRange> &optionalranges)
{
	vector<OptionalRange> stack;
	int curindex = 0;
    for( size_t i = 0, n = callsymbols.size(); i < n; ++i )
    {
        Symbol *item = callsymbols[i];
        sl_decl sldecl = sl_neutral;
        if( i < callsldecls.size() )
            sldecl = callsldecls[i];
		if( item == Symbol::DOT() )
            continue;
		if( item == Symbol::LB() )
		{
			OptionalRange newback;
			newback.m_start = curindex;
			stack.push_back(newback);
			continue;
		}
		else if( item == Symbol::RB() )
		{
			if( ! stack.size() )
				continue;
			OptionalRange back = stack.back();
			stack.pop_back();
			back.m_stop = curindex;
			// Don't bother adding empty brackets.
			if( back.m_start == back.m_stop )
				continue;
			if( stack.size() == 0 )
			{
				optionalranges.push_back(back);
				continue;
			}
			OptionalRange &prev = stack.back();
			// Don't bother adding brackets that are the same place as the enclosing brackets.
			if( prev.m_start == back.m_start && prev.m_stop == back.m_stop )
				continue;
			prev.m_optionalranges.push_back(back);
		}
		else
		{
			// Simple test to check pointer validity.
			// If you crash here check your call.
			item->IsTerminal();
			symbols.append(item);
            sldecls.append(sldecl);
			++curindex;
			continue;
		}
    }
	while( stack.size() )
	{
		OptionalRange back = stack.back();
		stack.pop_back();
		back.m_stop = curindex;
		// Don't bother adding empty brackets.
		if( back.m_start == back.m_stop )
			continue;
		if( stack.size() == 1 )
		{
			optionalranges.push_back(back);
			continue;
		}
		OptionalRange &prev = stack.back();
		// Don't bother adding brackets that are the same place as the enclosing brackets.
		if( prev.m_start == back.m_start && prev.m_stop == back.m_stop )
			continue;
		prev.m_optionalranges.push_back(back);
	}
}

Production *Grammar::FindOrAddProduction(Symbol *nt, vector<Symbol*> &callsymbols, vector<sl_decl> &callsldecls, bool &bNew)
{
    bNew = false;
    vector<Symbol*> symbols;
    vector<sl_decl> sldecls;
	vector<OptionalRange> optionalranges;
    GetSymbolsAndRanges(callsymbols, callsldecls, symbols, sldecls, optionalranges);
    for( vector<Production*>::iterator curproduction = m_productions.begin(); curproduction != m_productions.end(); ++curproduction )
    {
        if( (*curproduction)->m_nt != nt )
            continue;
        if( (*curproduction)->m_items != symbols )
            continue;
        if( (*curproduction)->m_optionalranges != optionalranges )
            continue;
        // got it.
        return *curproduction;
    }
    // Didn't find it, make it.
    bNew = true;
    return AddProductionCore(nt, symbols, sldecls, optionalranges, 0);
}

Production *Grammar::AddSimpleProduction(Symbol *nt, ...)
{
    va_list args;
    va_start(args, nt);
	Production *p = vAddSimpleProduction(nt, args);
    va_end(args);
	return p;
}

Production *Grammar::vAddSimpleProduction(Symbol *nt, va_list args)
{
    vector<Symbol*> symbols;
    vector<sl_decl> sldecls;

    while(1)
    {
        Symbol *item = va_arg(args, Symbol*);
        if( ! item )
            break;
		// Simple test to check pointer validity.
		// If you crash here check your call.
		item->IsTerminal();
		symbols.append(item);
        sldecls.append(sl_neutral);
    }

    return AddProduction(nt, symbols, sldecls, 0);
}

int Grammar::GetProductionIndex(Production *p)
{
    // Measurements show that getting production indices by brute force
    // is a major drag on time.  This snippet and the others that work
    // on these variables help keep the drag to a minimum.
    indexmap::iterator indexiter = m_indexmap.find(p);
    if( indexiter == m_indexmap.end() )
    {
        vector<Production*>::iterator piter = m_productions.find(p);
        if( piter == m_productions.end() )
            return -1;
        int index = piter - m_productions.begin();
        m_indexmap[p] = index;
        return index;
    }
    indexmap::kvpair &kv = *indexiter;
    return kv.second;
}

int Grammar::GetSymbolIndex(Symbol *s)
{
    vector<Symbol*>::iterator piter = m_symbols.find(s);
    if( piter == m_symbols.end() )
        return -1;
    return piter - m_symbols.begin();
}

int Grammar::GetProductionLabelIndex(ProductionLabel *pl)
{
    vector<ProductionLabel*>::iterator piter = m_labels.find(pl);
    if( piter == m_labels.end() )
        return -1;
    return piter - m_labels.begin();
}

int Grammar::GetOrderedTokenSetIndex(OrderedTokenSet *ots)
{
    vector<OrderedTokenSet*>::iterator piter = m_TokenSets.find(ots);
    if( piter == m_TokenSets.end() )
        return -1;
    return piter - m_TokenSets.begin();
}

void Grammar::AddStartState()
{
    m_initialstate = AddLRStateSet();
    // Set up the associations.  We don't bother adding 1st order allows, since
    // we know they're always on, it is much cheaper to simply check the m_1st_order_allow/m_1st_order_forbid list.
    for( vector<Production*>::iterator cur = m_startsymbol->m_OwningProductions.begin(); cur != m_startsymbol->m_OwningProductions.end(); ++cur )
    {
        Production *p = *cur;
        m_initialstate->AddLRItem(p, 0);
    }
}

void Grammar::ExpandTokensIntoProductions()
{
    // Generate the productions for this token set.
    for( vector<OrderedTokenSet*>::iterator curtokenset = m_TokenSets.begin(); curtokenset != m_TokenSets.end(); ++curtokenset )
    {
        (*curtokenset)->GenerateTokens();
        (*curtokenset)->AddBigDfaAsCfg();
    }

    // Make sure there aren't any problems before we continue.    
    if( m_errcount )
        return;

    // Process Rx symbols into productions and charset symbols.
    // Process literals into character productions.
    for( int i = 0, n = m_symbols.size(); i < n; ++i )
    {
        // Note: The m_symbols pointer may be reallocated between loops, so we must fetch
        // the next pointer each time, instead of using an iterator.
        m_symbols[i]->GenerateToken();
    }

    if( m_errcount )
        return;

	RemoveUnusedParts();

    if( m_errcount )
        return;
}

void Grammar::RemoveUnusedParts()
{
    // After all that fixing we've left our grammar a mess.
    // Go back and over the grammar and clean it up.
    // Cloned productions will end up in the clean version, but not the originals.
    // Cloned symbols will end up in the clean version, but mostly not the originals,
    // unless it participated in a production in an out-of-token-state context (nondeterministic tokenizing).
    hashset<Production*> usedproductions;
    hashset<Symbol*> usedsymbols;
    hashset<OrderedTokenSet*> usedtokensets;
    m_startsymbol->CollectUsedItems(usedproductions, usedsymbols, usedtokensets);
    // Token sets will be generated even if some tokens are not used, so make sure they are listed as used symbols.
    for( vector<OrderedTokenSet*>::iterator curTokenSet = m_TokenSets.begin(); curTokenSet != m_TokenSets.end(); ++curTokenSet )
    {
        if( usedtokensets.find(*curTokenSet) == usedtokensets.end() )
        {
            if( m_err )
                ob_printf(m_err, "warning: Token set <%s> is defined but not used\n", (*curTokenSet)->m_Name.c_str());
            m_wrncount++;
        }
    }
    // Remove unused symbols
	usedsymbols.insert(m_endmarker);
	usedsymbols.insert(m_startsymbol);
	usedsymbols.insert(m_epsilon);
    // Remove unused productions
    for( size_t i = 0; i < m_productions.size(); ++i )
    {
        Production *p = m_productions[i];
        if( usedproductions.find(p) == usedproductions.end() )
		{
			RemoveProduction(i);
			--i;
		}
    }
    for( size_t i = 0; i < m_symbols.size(); ++i )
    {
        Symbol *s = m_symbols[i];
        if( usedsymbols.find(s) == usedsymbols.end() )
        {
            RemoveSymbol(s);
            --i;
        }
    }
    // Remove unused tokensets
    for( size_t i = 0; i < m_TokenSets.size(); ++i )
    {
        OrderedTokenSet *ots = m_TokenSets[i];
        if( usedtokensets.find(ots) == usedtokensets.end() )
        {
            m_TokenSets.removeat(i, 1);
            delete ots;
            --i;
        }
    }
	if( m_TokenSets.size() == 1 )
		m_TokenSets[0]->m_bHideState = true;
}

void Grammar::GenerateOptionalPartPermutations()
{
	vector<Production*> productions = m_productions;
	vector< vector<SkipRange> > permutations;
	vector<Symbol*> symbols;
    vector<sl_decl> sldecls;

    for( vector<Production*>::iterator pp = productions.begin(); pp != productions.end(); ++pp )
	{
		Production *p = *pp;
		permutations.clear();
        permutations = p->GenerateOptionalPartPermutations();
		// The empty combination is always first, and is already covered by initial production
		if( permutations.size() <= 1 )
			continue;
		for( vector< vector<SkipRange> >::iterator permutation = permutations.begin()+1; permutation != permutations.end(); ++permutation )
		{
			Production *orig = p;
			if( p->m_copyof )
				orig = p->m_copyof;
			symbols = p->m_items;
            sldecls = p->m_sldecls;
            // Erase the parts that need erasing.  (Backwards, so the indices stay valid.)
			permutation->reverse();
			for( vector<SkipRange>::iterator cur = permutation->begin(); cur != permutation->end(); ++cur )
            {
				symbols.erase(symbols.begin()+cur->m_start, symbols.begin()+cur->m_stop);
                sldecls.erase(sldecls.begin()+cur->m_start, sldecls.begin()+cur->m_stop);
            }
            // Look to see if this is a duplicate of an existing production.
            bool bNew = false;
            Production *pclone = FindOrAddProduction(p->m_nt, symbols, sldecls, bNew);
			permutation->reverse();
            if( ! bNew && ! pclone->m_bPredeclareItem )
                continue;
            // It's not a duplicate, so add it.
			pclone->m_skips = *permutation;
			pclone->m_copyof = orig;
			orig->m_copycats.append(pclone);
			pclone->m_rank = orig->m_rank;
			pclone->m_reject = orig->m_reject;
			pclone->m_label = orig->m_label;
            pclone->FinalizeDeclaration(orig->GetIndex()+orig->m_copycats.size());
		}
	}
}

void Grammar::CheckForUnreachableAndProductionlessNonterminals()
{
    hashset<Nonterminal*> nonterminals;
    m_startsymbol->MarkReachableNonterminalsAndCheckForProductionlessNonterminals(nonterminals, m_err);
    for( vector<Nonterminal*>::iterator cur = m_nonterminals.begin(); cur != m_nonterminals.end(); ++cur )
    {
        if( nonterminals.find(*cur) == nonterminals.end() )
        {
            m_wrncount++;
            if( m_err )
                ob_printf(m_err, "warning: Nonterminal %s is unreachable\n", (*cur)->m_Name.c_str());
        }
    }
}

bool Grammar::Solve(outbuf *log)
{
    m_errcount = 0;
    m_wrncount = 0;

    if( m_productions.size() == 0 )
    {
        if( m_err )
            ob_puts(m_err, "No Productions Specified.");
        m_errcount++;
        return false;
    }

    // Take the shorthand associations and precedence and expand them into
    // full blown forbid rules.
    ApplyAssociations();
    if( m_errcount )
        return false;

    // Check that the grammar is well formed.
    CheckForUnreachableAndProductionlessNonterminals();
    if( m_errcount )
        return false;

    // Rewrite the grammar to expand productions including
    // optional parts to thier combinatorial equivalent.
    // This will not introduce ambiguities, but makes the grammar bigger.
    GenerateOptionalPartPermutations();
    if( m_errcount )
        return false;

    // For scannerless grammars.  We simulate the
    // effects of using push/pop/goto lexer state operators
    // by rewriting the grammar into an equivalent grammar.
    // In this pass however, we just get a list of the actions.
    m_TokenSetMerge = new TokenSetMergeAlgorithm(this, m_err);
    m_TokenSetMerge->Collect();

    // This is what we decided the best solution you can actually achieve.
    GrammarSolution BestAchievableGrammarSolution = UseBestAchievableSolution;

	// This is currently the best achievable.
    if( m_TokenSetMerge->m_TokenSetMiss.size() )
        BestAchievableGrammarSolution = NondeterministicParser_ScannerLess;
    
    // In a nutshell, we can only use a lexer if there are no token misses
    // because that means we are can be deterministic at the token level.
	if( m_RequestedGrammarSolution < BestAchievableGrammarSolution )
    {
	    if( m_TokenSetMerge->m_TokenSetMiss.size() == 0 )
			BestAchievableGrammarSolution = UseBestAchievableSolution;
		else
		{
			// Can't fix the lexer, and we were told to produce a deterministic scanner.  We're done.
			m_errcount++;
			if( m_err )
			{
				ob_puts(m_err, "error: TokenSet Misses were found while trying to produce a deterministic lexer and parser.\n");
				TokenStateMissReport(m_err);
			}
            return false;
		}
    }

    if( m_RequestedGrammarSolution == UseBestAchievableSolution )
        m_RequestedGrammarSolution = BestAchievableGrammarSolution;

    // If we are to continue creating a scannerless solution, we need to do the token set merge.
    if( m_RequestedGrammarSolution == NondeterministicParser_ScannerLess )
    {        
        // For Non-deterministic grammars.  Simulate the
        // effects of using push/pop/set lexer state operators
        // by rewriting the grammar into an equivalent grammar.
        m_TokenSetMerge->Apply();
        if( m_errcount )
            return false;
    }

	// Add the starting state.
	AddStartState();

	if( m_errcount )
		return false;

	if( m_RequestedGrammarSolution == NondeterministicParser_ScannerLess )
	{
		// Expand the regular expression and string tokens
		// into productions, taking token sets into account.
		ExpandTokensIntoProductions();
		if( m_errcount )
			return false;
	}

	// Compute the firsts for all the symbols.
	ComputeFirsts();
	if( m_errcount )
		return false;
	// Build the states and gotos that comprise the grammar states.
	// Requires ComputeFirsts has been done.
	BuildStatesAndGotos();
	if( m_errcount )
		return false;
	// Not sure if this is right.
	FixupFollowRestricts();
	if( m_errcount )
		return false;
	// Compute follows using a lookahead technique.
	// Requires that all states and state items are computed.
	ComputeFollows();
	if( m_errcount )
		return false;
	// Generate actions for the various states.  Requires that
	// states and gotos are calculated, and also the follow sets.
	GenerateActions();
	if( m_errcount )
		return false;

	CollectConflicts();

	// If it's not already determined to be scannerless, assign the best achievable solution.
	if( BestAchievableGrammarSolution == UseBestAchievableSolution )
	{
		if( m_ActionConflicts.size() )
			BestAchievableGrammarSolution = NondeterministicParser;
		else
			BestAchievableGrammarSolution = DeterministicParser;
	}
    
	if( m_RequestedGrammarSolution == UseBestAchievableSolution )
		m_RequestedGrammarSolution = BestAchievableGrammarSolution;

	if( m_RequestedGrammarSolution < BestAchievableGrammarSolution )
	{
		// Must have requested NondeterministicParser, since scannerless would have been
		// detected and weeded out before solving the parser part of the grammar.
		if( m_err )
			ob_puts(m_err, "error: Conflicts were found while trying to produce a requested deterministic parser; parser can be GLR at best.\n");
        m_errcount++;
        if( log )
            PrintLog(log);
		return false;
	}

    // It's got a (deterministic) scanner at least, so create the scanner tokens.
	if( BestAchievableGrammarSolution < NondeterministicParser_ScannerLess )
    {
        for( vector<OrderedTokenSet*>::iterator curtokenset = m_TokenSets.begin(); curtokenset != m_TokenSets.end(); ++curtokenset )
            (*curtokenset)->GenerateTokens();
        if( m_errcount )
            return false;
    }

    if( log )
        PrintLog(log);

    return true;
}

bool Grammar::Emit(emitoptions &options)
{
    return emitC(this, options) ? true : false;
}

void Grammar::ApplyAssociationsByLabel(ProductionOrLabel parentlabel, ProductionOrLabel childlabel, Assoc::AssocStyle assocstyle, Assoc::Association assoc, vector<IntSetRange> &assoc_pos, const vector<LRItem> &from_series, String fname, int lineno)
{
    vector<Production*> parentproductions = parentlabel.GetProductions(this);
    vector<Production*> childproductions = childlabel.GetProductions(this);
    for( vector<Production*>::iterator childproduction = childproductions.begin(); childproduction != childproductions.end(); ++childproduction )
    {
        for( vector<Production*>::iterator parentproduction = parentproductions.begin(); parentproduction != parentproductions.end(); ++parentproduction )
        {
            if( ! (*parentproduction)->AddAssociation(*childproduction, assocstyle, assoc, assoc_pos, from_series, fname, lineno) )
            {
				m_errcount++;
                if( m_err )
                {
                    ob_printf(m_err, "error: %s(%d): Couldn't add association:\n", fname.c_str(), lineno);
                    ob_puts(m_err, "    Child production: ");
                    (*childproduction)->Print(m_err);
                    ob_puts(m_err, "\n   Cannot fit into Parent production: ");
                    (*parentproduction)->Print(m_err);
                    ob_puts(m_err, "\n    ");
                    switch( assoc )
                    {
                        case Assoc::Left:
                            ob_puts(m_err, "with Left association");
                            break;
                        case Assoc::Right:
                            ob_puts(m_err, "with Right association");
                            break;
                        case Assoc::Non:
                            ob_puts(m_err, "with Non association");
                            break;
                        case Assoc::ByPosition:
                            ob_puts(m_err, "with position association = (");
                            for( vector<IntSetRange>::iterator cur = assoc_pos.begin(); cur != assoc_pos.end(); ++cur )
                                ob_printf(m_err, "%d through %d, ", cur->m_min, cur->m_max);
                            ob_putc(m_err, ')');
                            break;
                    }
                    ob_putc(m_err, '\n');
                }
            }
        }
    }
}

// Apply priority set associations.  Labels are applied to productions, and
// priorities applied to labels.  When associations are used on a group of labels
// that use the '=' operator between them, the association is applied to every
// combination of productions with those labels.
//
// The algorithm:
//
// For each set of priorities
//     For each priority group
//        For each combination of labels
//            For each combination of productions corresponsing to those labels
//                Apply the association.
void Grammar::ApplyAssociations()
{
    for( vector<PrioritySet>::iterator priority = m_priorities.begin(); priority != m_priorities.end(); ++priority )
        for( vector<PriorityGroup>::iterator prioritygroup = priority->m_groups.begin(); prioritygroup != priority->m_groups.end(); ++prioritygroup )
            for( vector<ProductionLabel*>::iterator parentlabel = prioritygroup->m_labels.begin(); parentlabel != prioritygroup->m_labels.end(); ++parentlabel )
                for( vector<ProductionLabel*>::iterator childlabel = prioritygroup->m_labels.begin(); childlabel != prioritygroup->m_labels.end(); ++childlabel )
                    ApplyAssociationsByLabel(*parentlabel, *childlabel, Assoc::Forbid, prioritygroup->m_assoc, prioritygroup->m_assoc_pos, prioritygroup->m_from_series, priority->m_fname, priority->m_lineno);
	for( vector<AssocByLabel>::iterator curassoc = m_assocbylabels.begin(), endassoc = m_assocbylabels.end(); curassoc != endassoc; ++curassoc )
		ApplyAssociationsByLabel(curassoc->m_parentlabel, curassoc->m_childlabel, curassoc->m_assocstyle, curassoc->m_assoc, curassoc->m_assoc_pos, curassoc->m_from_series, curassoc->m_fname, curassoc->m_lineno);
}

class GotoTransition
{
public:
    GotoTransition() : m_gtype(GTypeNone), m_p(0), m_s(0)
    {
    }

	LRStateSet *m_from;
	LRItem m_to;
    GType m_gtype;
    Production *m_p;
    Symbol *m_s;
    associtem_map m_associtems;
    
    bool operator==(const GotoTransition &rhs) const
    {
        if( ! HasSameTransition(rhs) )
            return false;
		if( m_from != rhs.m_from )
			return false;
        if( m_to != rhs.m_to )
            return false;
        if( m_associtems != rhs.m_associtems )
            return false;
        return true;
    }

	static int cmp_GotoTransition(const GotoTransition *lhs, const GotoTransition *rhs)
	{
        if( lhs->m_gtype != rhs->m_gtype )
			return lhs->m_gtype - rhs->m_gtype;
		if( lhs->m_from != rhs->m_from )
			return lhs->m_from->m_number - rhs->m_from->m_number;
        if( lhs->m_gtype == GTypeProduction )
        {
            if( lhs->m_p != rhs->m_p )
                return lhs->m_p->GetIndex() - rhs->m_p->GetIndex();
        }
        else
        {
            if( lhs->m_s != rhs->m_s )
				return lhs->m_s->GetIndex() - rhs->m_s->GetIndex();
        }
		return LRItem::cmp_lritem(&lhs->m_to,&rhs->m_to);
	}
    
    bool HasSameTransition(const GotoTransition &rhs) const
    {
        if( m_gtype != rhs.m_gtype )
            return false;
        if( m_gtype == GTypeProduction )
        {
            if( m_p != rhs.m_p )
                return false;
        }
        else
        {
            if( m_s != rhs.m_s )
                return false;
        }
        return true;
    }

    bool HasNullTransition() const
    {
        if( m_gtype == GTypeNone )
            return true;
        if( m_gtype == GTypeProduction )
        {
            if( ! m_p )
                return true;
        }
        else
        {
            if( ! m_s )
                return true;
        }
        return false;
    }
    
    void CopyTransition(const GotoTransition &rhs)
    {
        m_gtype = rhs.m_gtype;
        m_p = rhs.m_p;
        m_s = rhs.m_s;
        m_associtems = rhs.m_associtems;
    }

	void Print(outbuf *pio)
	{
		ob_printf(pio, "goto transition from state %d to [", m_from->m_number);
		m_to.Print(pio);
		if( m_gtype == GTypeSymbol )
		{
			ob_puts(pio, "] on ");
			m_s->Print(pio);
			ob_putc(pio, '\n');
		}
		else
		{
			ob_puts(pio, "] on [");
			m_p->Print(pio);
			ob_puts(pio, "]\n");
		}
	}
};

// We promote a assoc rule item whenever the previous state (lrss)
// contains the next item in the from series of a forbid rule (we'll call it NI),
// and the next state (tmplrstateset) contains an LRItem (now at position 1) which
// is a valid child of NI at the activating position.
void Grammar::UpdateRuleSet(LRStateSet &tmplrstateset, LRStateSet *lrss)
{
    associtem_map &src_assocmap = lrss->m_associtems;
    associtem_map &dst_assocmap = tmplrstateset.m_associtems;
    dst_assocmap = src_assocmap;

    // Attempt to promote gt 1st order associations that are at > 0th level.
    for( associtem_map::iterator cur = dst_assocmap.begin(), e = dst_assocmap.end(); cur != e; ++cur )
    {
        AssocItems &associtems = cur->second;
        sortedset<AssocItem> *psets[2] = {&associtems.m_inactive_allow, &associtems.m_inactive_forbid};
        for( int i = 0; i < 2; ++i )
        {
            sortedset<AssocItem> *pset = psets[i];
            for( sortedset<AssocItem>::iterator curassoc = pset->begin(), endassoc = pset->end(); curassoc != endassoc; ++curassoc )
            {
                LRItem frm = curassoc->NextFrom();
                if( lrss->m_items.find(frm) == lrss->m_items.end() )
                    continue;
                for( vector<LRItem>::iterator curitem = tmplrstateset.m_items.begin(); curitem != tmplrstateset.m_items.end(); ++curitem )
                {
                    if( curitem->m_pos != 1 )
                        continue;
                    if( ! frm.m_p->ChildAllowedAtPosition(*lrss, curitem->m_p, frm.m_pos) )
                        continue;
                    // Adding 1 won't alter the order.
                    curassoc->m_pos++;
                    if( curassoc->IsActive() )
                    {
                        associtems.AddAssocItem(*curassoc);
                        curassoc = pset->erase(curassoc);
                        --curassoc;
                    }
                    break;
                }
            }
        }
    }
    // Attempt to promote gt 1st order associations that are at their 0th level.
    // Technically, everything in m_assocs.m_1st_order_allow/m_assocs.m_1st_order_allow is active, but
    // we don't bother adding them.  Also, everything in m_assocs.m_gt_1st_order_allow/m_gt_1st_order_forbid
    // is inactive but considered present at it's 0 level.
    for( assoc_map::iterator cur = m_assocs.begin(), e = m_assocs.end(); cur != e; ++cur )
    {
        Assocs &assocs = cur->second;
        AssocItems *pAssocItems = 0;
        vector<Assoc> *vecs[2] = {&assocs.m_gt_1st_order_allow, &assocs.m_gt_1st_order_forbid};
        for( int i = 0; i < 2; ++i )
        {
            vector<Assoc> *pvec = vecs[i];
            for( vector<Assoc>::iterator curassoc = pvec->begin(), endassoc = pvec->end(); curassoc != endassoc; ++curassoc )
            {
                LRItem frm = curassoc->GetFrom(0);
                if( lrss->m_items.find(frm) == lrss->m_items.end() )
                    continue;
                for( vector<LRItem>::iterator curitem = tmplrstateset.m_items.begin(); curitem != tmplrstateset.m_items.end(); ++curitem )
                {
                    if( curitem->m_pos != 1 )
                        continue;
                    if( ! frm.m_p->ChildAllowedAtPosition(*lrss, curitem->m_p, frm.m_pos) )
                        continue;
                    if( ! pAssocItems )
                        pAssocItems = &dst_assocmap[cur->first];
                    AssocItem ai(*curassoc, 1);
                    pAssocItems->AddAssocItem(ai);
                }
            }
        }
    }
}

void Grammar::BuildStatesAndGotos()
{
    // These are here as "optimization", so that we reuse the same memory.
    hashset<Production*> addedproductions;
    LRStateSet tmplrstateset(this);
    vector<GotoTransition> transitions;
    SymbolSet BigCSet;
    SymbolSet LittleCSet;

    // Build states and goto transitions.
    // We previously recorded the items for the m_initialstate.
    outbuf *verbose = m_bLogClosure ? m_verbose : 0;
    if( verbose )
        m_initialstate->SLRClosure_Verbose(addedproductions, verbose);
    else
        m_initialstate->SLRClosure(addedproductions);
    
    for( int i = 0; i != m_lrstates.size(); ++i )
    {
        LRStateSet *lrss = m_lrstates[i];
        // Collect new nonterminal items, and collect terminals.
        BigCSet.clear();
        transitions.clear();
        for( vector<LRItem>::iterator lritem = lrss->m_items.begin(); lritem != lrss->m_items.end(); ++lritem )
        {
            Symbol *s = lritem->GetSymbolAtPosition();
            if( ! s )
                continue; // At end
            if( s->IsTerminal() )
            {
			    if( s->IsCharSet() )
			    {
				    BigCSet += s->m_firsts;
				    continue; // Character of character set, so just collect.
			    }
                GotoTransition trns;
                trns.m_from = lrss;
                trns.m_to = *lritem;
                trns.m_to.m_pos++;
                trns.m_gtype = GTypeSymbol;
                trns.m_s = s;
                trns.m_associtems = lrss->m_associtems;
                transitions.append(trns);
			    continue;
            }
            for( vector<Production*>::iterator nextproduction = s->m_OwningProductions.begin(); nextproduction != s->m_OwningProductions.end(); ++nextproduction )
            {
                if( ! lritem->ChildAllowedAtPosition(*nextproduction) )
                    continue;
                GotoTransition trns;
                trns.m_from = lrss;
                trns.m_to = *lritem;
                trns.m_to.m_pos++;
                trns.m_gtype = GTypeProduction;
                trns.m_p = *nextproduction;
                trns.m_associtems = lrss->m_associtems;
                transitions.append(trns);
            }
        }
        // Partition the big character set into new character sets, and collect new terminal items.
        for( UIntSet::iterator curseg = BigCSet.m_chars.begin(); curseg != BigCSet.m_chars.end(); ++curseg )
        {
            unsigned int curch = curseg->m_min;
            unsigned int lastch = curseg->m_max;
            unsigned int capch = lastch;
            while( curch <= lastch )
            {
                vector<LRItem> matching_substates;
                for( vector<LRItem>::iterator lritem = lrss->m_items.begin(); lritem != lrss->m_items.end(); ++lritem )
                {
                    Symbol *s = lritem->GetSymbolAtPosition();
                    if( ! s )
                        continue; // At end
                    LittleCSet.clear();
                    if( ! s->IsTerminal() )
                        continue;
                    LittleCSet += s->m_firsts;
                    UIntSet::iterator ipair = LittleCSet.m_chars.findintersecting(curch, capch);
                    if( ipair == LittleCSet.m_chars.end() )
                        continue;
                    if( ipair->m_min > curch )
                    {
                        capch = ipair->m_min-1;
                        continue;
                    }
                    if( ipair->m_max < capch )
                        capch = ipair->m_max;
                    matching_substates.append(*lritem);
                }
                Symbol *newcset = FindOrAddCharset(UIntSet()%UIntSetRange(curch,capch));
                for( vector<LRItem>::iterator isubstate = matching_substates.begin(); isubstate != matching_substates.end(); ++isubstate )
                {
                    GotoTransition trns;
					trns.m_from = lrss;
                    trns.m_to = *isubstate;
                    trns.m_to.m_pos++;
                    trns.m_gtype = GTypeSymbol;
                    trns.m_s = newcset;
                    trns.m_associtems = lrss->m_associtems;
                    transitions.append(trns);
                }
                curch = capch + 1;
                capch = lastch;
            }
        }
        // sort transitions by transition symbol
		transitions.unique(GotoTransition::cmp_GotoTransition);
		GotoTransition curtrnsstate;
		tmplrstateset.m_items.clear();
        tmplrstateset.m_associtems.clear();
		for( vector<GotoTransition>::iterator item = transitions.begin(); item != transitions.end(); ++item )
		{
			if( ! curtrnsstate.HasSameTransition(*item) )
			{
				if( ! curtrnsstate.HasNullTransition() )
				{
                    UpdateRuleSet(tmplrstateset, lrss);
                    if( verbose )
                        tmplrstateset.SLRClosure_Verbose(addedproductions, verbose);
                    else
					    tmplrstateset.SLRClosure(addedproductions);
					LRStateSet *nextstate = FindLRStateSet(tmplrstateset);
					if( ! nextstate )
						nextstate = AddLRStateSet(tmplrstateset);
					tmplrstateset.m_items.clear();
                    tmplrstateset.m_associtems.clear();
					lrss->AddGotoItem(curtrnsstate.m_gtype, curtrnsstate.m_s, curtrnsstate.m_p, nextstate);
				}
				curtrnsstate.CopyTransition(*item);
			}
			tmplrstateset.m_items.append(item->m_to);
		}
		if( tmplrstateset.m_items.size() )
		{
            UpdateRuleSet(tmplrstateset, lrss);
            if( verbose )
			    tmplrstateset.SLRClosure_Verbose(addedproductions, verbose);
            else
                tmplrstateset.SLRClosure(addedproductions);
			LRStateSet *nextstate = FindLRStateSet(tmplrstateset);
			if( ! nextstate )
				nextstate = AddLRStateSet(tmplrstateset);
			lrss->AddGotoItem(curtrnsstate.m_gtype, curtrnsstate.m_s, curtrnsstate.m_p, nextstate);
		}
    }

    // Combinine gotos on character sets that shift to the same state.
	for( vector<LRStateSet*>::iterator curstate = m_lrstates.begin(), endstate = m_lrstates.end(); curstate != endstate; ++curstate )
	{
		sortedset<LRStateSet*> built_tostate;
		SymbolSet building_cset;
		for( int i = 0; i < (int)(*curstate)->m_gotos.size(); ++i )
		{
			GotoItem *curitem = (*curstate)->m_gotos.begin()+i;
			if( curitem->m_gtype != GTypeSymbol || ! curitem->m_s->IsCharSet() )
				continue;
			if( built_tostate.find(curitem->m_to) != built_tostate.end() )
				continue;
			LRStateSet *to = curitem->m_to;
			built_tostate.insert(to);
			building_cset.clear();
			int anotherindex = i;
			while( anotherindex < (int)(*curstate)->m_gotos.size() )
			{
				GotoItem *anotheritem = (*curstate)->m_gotos.begin()+anotherindex;
				if( anotheritem->m_to != to || anotheritem->m_gtype != GTypeSymbol || ! anotheritem->m_s->IsCharSet() )
				{
					++anotherindex;
					continue;
				}
				building_cset += anotheritem->m_s->m_firsts;
				(*curstate)->m_gotos.removeat(anotherindex, 1);
				if( i == anotherindex )
					--i;
				else
					++anotherindex;
			}
			if( building_cset.m_chars.count() == 0 )
				continue;
			Symbol *newcset = FindOrAddCharset(building_cset.m_chars);
			(*curstate)->AddGotoItem(GTypeSymbol, newcset, 0, to);
		}
	}

	// Done with that part, now add the backwards paths for the gotos
	for( vector<LRStateSet*>::iterator curstate = m_lrstates.begin(), endstate = m_lrstates.end(); curstate != endstate; ++curstate )
		for( vector<GotoItem>::iterator curgoto = (*curstate)->m_gotos.begin(), endgoto = (*curstate)->m_gotos.end(); curgoto != endgoto; ++curgoto )
			curgoto->m_to->m_backgotos.append( GotoItem(curgoto->m_gtype, curgoto->m_s, curgoto->m_p, *curstate) );
}

LRStateSet *Grammar::AddLRStateSet()
{
    LRStateSet *ret = new LRStateSet(this);
    ret->m_number = m_lrstates.size();
    m_lrstates.append(ret);
    return ret;
}

LRStateSet *Grammar::AddLRStateSet(const LRStateSet &copy)
{
    LRStateSet *ret = AddLRStateSet();
	for( vector<LRItem>::const_iterator curitem = copy.m_items.begin(); curitem != copy.m_items.end(); ++curitem )
		ret->AddLRItem(curitem->m_p, curitem->m_pos);
    ret->m_associtems = copy.m_associtems;
    return ret;
}

LRStateSet *Grammar::FindLRStateSet(const LRStateSet &match)
{
    for( vector<LRStateSet*>::iterator cur = m_lrstates.begin(); cur != m_lrstates.end(); ++cur )
    {
        if( **cur == match )
            return *cur;
    }
    return 0;
}

void Grammar::ComputeFirsts()
{
    outbuf *verbose = m_bLogFirsts ? m_verbose : 0;
    if( verbose )
        ob_puts(verbose, "Computing FIRSTS\n");

	for( vector<Symbol*>::iterator cursymbol = m_symbols.begin(); cursymbol != m_symbols.end(); ++cursymbol )
		(*cursymbol)->ResetFirsts();

    // FIRSTS
	int iteration = 0;
    int go = 1;
	SymbolSet tmp;
    while(go)
    {
		if( verbose )
			ob_printf(verbose, "Iteration %d\n", iteration);
        go = 0;
        for( vector<Production*>::iterator p = m_productions.begin(); p != m_productions.end(); ++p )
        {
            Symbol *nt = (*p)->m_nt;
			if( verbose )
			{
				ob_puts(verbose, "Computing FIRSTS for [");
				(*p)->Print(verbose);
				ob_puts(verbose, "], FIRST(");
				nt->Print(verbose);
				ob_puts(verbose, ") = ");
				nt->m_firsts.Print(verbose);
				ob_putc(verbose, '\n');
			}
            int prevcount = nt->m_firsts.count();
            tmp.clear();
            (*p)->ComputeFirsts(tmp);
            nt->m_firsts += tmp;
			if( verbose )
			{
				ob_puts(verbose, "Adding ");
				tmp.Print(verbose);
				ob_puts(verbose, " to ");
				ob_puts(verbose, "FIRST(");
				nt->Print(verbose);
				ob_puts(verbose, ") -> ");
				nt->m_firsts.Print(verbose);
				ob_putc(verbose, '\n');
			}
            int newcount = nt->m_firsts.count();
            if( prevcount != newcount )
                go = 1;
        }
		iteration++;
    }

	if( m_errcount )
		return;

	if( verbose )
		PrintFirsts(verbose);
}

void Grammar::PrintFirsts(outbuf *p) const
{
    if( ! p )
        return;
    SymbolSet tmp;
    for( vector<Symbol*>::const_iterator s = m_symbols.begin(); s != m_symbols.end(); ++s )
    {
        if( (*s)->IsCharSet() )
            continue;
        ob_puts(p, "FIRST(");
        (*s)->Print(p);
        ob_puts(p, ") = ");
        tmp = (*s)->m_firsts;
        tmp.FlattenCharsets();
        tmp.Print(p);
        ob_putc(p, '\n');
    }
}

void Grammar::FixupFollowRestricts()
{
    SymbolSet follow_restrict;
    for( vector<Symbol*>::iterator s = m_symbols.begin(); s != m_symbols.end(); ++s )
    {
        if( (*s)->IsTerminal() )
            continue;
        follow_restrict = (*s)->m_follow_restrict;
        for( vector<Symbol*>::iterator s2 = m_symbols.begin(); s2 != m_symbols.end(); ++s2 )
        {
            if( (*s2)->IsTerminal() )
                continue;
            if( ! (follow_restrict & *s2) )
                continue;
            follow_restrict -= *s2;
            follow_restrict += (*s2)->m_firsts;
            follow_restrict.FlattenCharsets();
        }
    }
}

void Grammar::PrintFollowRestricts(outbuf *p) const
{
    SymbolSet tmp;
    for( vector<Symbol*>::const_iterator s = m_symbols.begin(); s != m_symbols.end(); ++s )
    {
        ob_puts(p, "FOLLOW_RESTRICT(");
        (*s)->Print(p);
        ob_puts(p, ") = ");
        tmp = (*s)->m_follow_restrict;
        tmp.FlattenCharsets();
        tmp.Print(p);
        ob_putc(p, '\n');
    }
}

class followdep
{ 
public:
	SymbolSet *m_from;
    LRItem *m_fromlritem;
	SymbolSet *m_to;
    LRItem *m_tolritem;
	String m_label;
	int m_sortindex;
    SymbolSet m_copied;

	followdep() : m_from(0), m_to(0), m_sortindex(-1)
	{
	}

	followdep(SymbolSet *from, SymbolSet *to, int sortindex)
	: m_from(from), m_to(to), m_sortindex(sortindex), m_fromlritem(0), m_tolritem(0)
	{
	}

	followdep(SymbolSet *from, LRItem *fromlritem, SymbolSet *to, LRItem *tolritem, int sortindex)
	: m_from(from), m_fromlritem(fromlritem), m_to(to), m_tolritem(tolritem), m_sortindex(sortindex)
	{
	}

	static int bysortindex(const followdep *lhs, const followdep *rhs)
	{
		if( rhs->m_sortindex < lhs->m_sortindex )
			return 1;
		if( lhs->m_sortindex < rhs->m_sortindex )
			return -1;
		return 0;
	}

	bool operator<(const followdep &rhs) const
	{
		if( m_from < rhs.m_from )
			return true;
		else if( m_from > rhs.m_from )
			return false;
		else if( m_to < rhs.m_to )
			return true;
		else if( m_to > rhs.m_to )
			return false;
		return false;
	}

	bool operator==(const followdep &rhs) const
	{
		if( m_from != rhs.m_from || m_to != rhs.m_to )
			return false;
		return true;
	}

	bool update()
	{
		int precount = m_to->count();
		*m_to += *m_from;
		int postcount = m_to->count();
		if( precount == postcount )
			return false;
		return true;
	}

	bool update(SymbolSet &copied)
	{
		copied.clear();
		copied = *m_from - *m_to;
		int precount = m_to->count();
		*m_to += *m_from;
		int postcount = m_to->count();
		if( precount == postcount )
			return false;
		return true;
	}
};

static void AddFollow(outbuf *ss, verboseinfo *verbosedb, int dstid, int srcid, int isdep, Symbol *s)
{
    const char *sname = 0;
    if( s )
    {
        obs_clear(ss);
        s->Print(ss);
        sname = obs_getstr(ss);
    }
    verbosedb->AddFollow(dstid,srcid,isdep,sname);
}

static void AddFollows(outbuf *ss, verboseinfo *verbosedb, int dstid, int srcid, int isdep, SymbolSet &copied)
{
    bool bAdded = false;
    if( copied.m_chars.count() )
    {
        CharSet cs(0, copied.m_chars);
        AddFollow(ss, verbosedb, dstid, srcid, isdep, &cs);
        bAdded = true;
    }
    for( sortedset<Symbol*>::iterator cursym = copied.m_symbols.begin(), endsym = copied.m_symbols.end(); cursym != endsym; ++cursym )
    {
        AddFollow(ss, verbosedb, dstid, srcid, isdep, *cursym);
        bAdded = true;
    }
    if( ! bAdded )
        AddFollow(ss, verbosedb, dstid, srcid, isdep, 0);
}

// Our goal here is to find the follow sets for each reduceable item in every lr state.
// We do that by following each reduce path and looking at the symbols that follow the
// reduced nonterminal where it lies in the lr state item.  Following the reduce path
// is accomplished by taking the goto state transitions backwards along the reduced
// production.  Also, keep in mind that the follow symbols within an lr state item may
// yield an epsilon, in which case we need to know the follow sets for the reducable
// items that following the goto transitions forward will yield, this means that our
// approach is iterative until closure of the follow sets.  The last thing to keep
// in mind is that a particular reduction may be forbidden under the rules for our grammar
// and in the items which would result in a conflicting placement are ignored.
//
// The initial state is set up by adding the $ symbol to each xxx. -> <Start> lr state item.
// After that, we collect symbols into our follow sets the best we can and determine which
// follow sets depend on one another according to the criteria described above.  Finally,
// we iteratively copy according to the symbol set dependencies until closure is reached.
//
// The resulting follow sets obey the forbid (operator precedence) rules and have a lookahead
// of 1.
void Grammar::ComputeFollows_Verbose()
{
    outbuf *verbose = m_bLogFollows ? m_verbose : 0;
	vector<Symbol*> parentsymbols;
    if( verbose )
        ob_puts(verbose, "Computing FOLLOWS\n");
    if( m_verbosedb )
        m_verbosedb->clear();

	vector<followdep> deps;
    hashmap<LRItem*,int> item_rowids;
    outbuf *ss = new_string_outbuf();
	outbuf *verboselabel = new_string_outbuf();;

    // Do this first.  We end up needing the id's of lrstates we haven't encountered yet.
    if( m_verbosedb )
    {
	    for( vector<LRStateSet*>::iterator curstate = m_lrstates.begin(), endstate = m_lrstates.end(); curstate != endstate; ++curstate )
	    {
		    for( vector<LRItem>::iterator curitem = (*curstate)->m_items.begin(), enditem = (*curstate)->m_items.end(); curitem != enditem; ++curitem )
		    {
                int stateno = (*curstate)->m_number;
                int itemidx = curitem-curitem->m_ss->m_items.begin();
                obs_clear(ss);
                curitem->Print(ss);
                const char *displaystring = obs_getstr(ss);
                const char *allfollows = 0;
                item_rowids[curitem] = m_verbosedb->AddItem(stateno, itemidx, displaystring, allfollows);
            }
        }
    }

	for( vector<LRStateSet*>::iterator curstate = m_lrstates.begin(), endstate = m_lrstates.end(); curstate != endstate; ++curstate )
	{
		for( vector<LRItem>::iterator curitem = (*curstate)->m_items.begin(), enditem = (*curstate)->m_items.end(); curitem != enditem; ++curitem )
		{
			if( curitem->GetSymbolAtPosition() )
			{
				if( verbose )
				{
					ob_printf(verbose, "Skipping State %d, [", curitem->m_ss->m_number);
					curitem->Print(verbose);
					ob_puts(verbose, "] (not a reducable item)\n");
				}
				continue;
			}
			if( verbose )
			{
				ob_printf(verbose, "Computing Post Reduce Items for State %d, [", curitem->m_ss->m_number);
				curitem->Print(verbose);
				ob_puts(verbose, "]\n");
			}
			if( curitem->m_p->m_nt->IsStartSymbol() )
			{
				if( verbose )
				{
					ob_puts(verbose, "    Added ");
					m_endmarker->Print(verbose);
					ob_puts(verbose, " to [");
					curitem->Print(verbose);
					ob_puts(verbose, "] (nothing follows)\n");
				}
                if( m_verbosedb )
                    AddFollow(ss, m_verbosedb, item_rowids[curitem], 0, 0, m_endmarker);
				curitem->m_follows += m_endmarker;
				continue; // Nothing follows accept
			}
			lritem_ptrset postreduceitems = curitem->GetPostReduceItems_Verbose(verbose);
			for( lritem_ptrset::iterator curpritem = postreduceitems.begin(), endpritem = postreduceitems.end(); curpritem != endpritem; ++curpritem )
			{
				if( verbose )
				{
					ob_printf(verbose, "    State %d, [", (*curpritem)->m_ss->m_number);
					(*curpritem)->Print(verbose);
					ob_puts(verbose, "] is a legal post reduction state.");
				}
				if( verbose || m_verbosedb )
				{
					SymbolSet followsatpos = (*curpritem)->GetFollowsAtPosition();
					SymbolSet copied;
					copied = followsatpos - curitem->m_follows;
					copied -= m_epsilon;
					curitem->m_follows += followsatpos;
					if( copied.count() )
					{
                        if( verbose )
                        {
						    ob_puts(verbose, " (Added ");
						    copied.Print(verbose);
						    ob_putc(verbose, ')');
                        }
                        if( m_verbosedb )
                            AddFollows(ss, m_verbosedb, item_rowids[curitem], item_rowids[*curpritem], 0, copied);
					}
					else
					{
                        if( verbose )
						    ob_puts(verbose, " (No Symbols Added)");
					}
				}
				else
				{
					curitem->m_follows += (*curpritem)->GetFollowsAtPosition();
				}
				if( ! (curitem->m_follows & m_epsilon) )
				{
					if( verbose )
						ob_puts(verbose, " (FOLLOW set will NOT propagate to parent production)\n");
					continue;
				}
				if( verbose )
					ob_puts(verbose, " (FOLLOW set will propagate to parent production)\n");
				curitem->m_follows -= m_epsilon;
				lritem_ptrset reducibleitems = (*curpritem)->GetReducibleItems();
				for( lritem_ptrset::iterator curreducibleitem = reducibleitems.begin(), endreducibleitem = reducibleitems.end(); curreducibleitem != endreducibleitem; ++curreducibleitem )
				{
					deps.append( followdep(&(*curreducibleitem)->m_follows, *curreducibleitem, &curitem->m_follows, curitem, (int)deps.size()) );
					if( verbose )
					{
                        obs_clear(verboselabel);
						ob_printf(verboselabel, "FOLLOW(State %d, [", curitem->m_ss->m_number);
						curitem->Print(verboselabel);
						ob_puts(verboselabel, "]) <- ");
						ob_printf(verboselabel, "FOLLOW(State %d, [", (*curreducibleitem)->m_ss->m_number);
						(*curreducibleitem)->Print(verboselabel);
						ob_puts(verboselabel, "])");
						deps.back().m_label = obs_getstr(verboselabel);
					}
				}
			}
		}
	}

	deps.unique( simplecmp<followdep> );
	deps.sort( followdep::bysortindex );

	SymbolSet copied;
	int go = 1;
	while( go )
	{
		if( verbose )
			ob_puts(verbose, "Start Update loop\n");
		go = 0;
		for( vector<followdep>::iterator curdep = deps.begin(); curdep != deps.end(); ++curdep )
		{
			if( verbose || m_verbosedb )
			{
				if( curdep->update(copied) )
					go = 1;
                curdep->m_copied += copied;
                if ( verbose )
                {
				    if( copied.count() )
				    {
				        ob_printf(verbose, "Updated %s, added ", curdep->m_label.c_str());
				        copied.Print(verbose);
				        ob_putc(verbose, '\n');
				    }
				    else
				        ob_printf(verbose, "Updated %s, (No Updates)\n", curdep->m_label.c_str());
                }
			}
			else
			{
				if( curdep->update() )
					go = 1;
			}
		}
	}

	for( vector<LRStateSet*>::iterator curstate = m_lrstates.begin(), endstate = m_lrstates.end(); curstate != endstate; ++curstate )
	{
		for( vector<LRItem>::iterator curitem = (*curstate)->m_items.begin(), enditem = (*curstate)->m_items.end(); curitem != enditem; ++curitem )
		{
			if( curitem->m_follows.m_chars.count() == 0 )
				continue;
			Symbol *cset = FindOrAddCharset(curitem->m_follows.m_chars);
            SymbolSet& follows = curitem->m_follows;
			follows.m_chars.clear();
			follows += cset;
		}
	}

    if( ! m_verbosedb )
        return;

    for( vector<followdep>::iterator curdep = deps.begin(); curdep != deps.end(); ++curdep )
    {
        m_verbosedb->AddDep(item_rowids[curdep->m_tolritem], item_rowids[curdep->m_fromlritem]);
        AddFollows(ss, m_verbosedb, item_rowids[curdep->m_tolritem], item_rowids[curdep->m_fromlritem], 1, curdep->m_copied);
    }

	for( vector<LRStateSet*>::iterator curstate = m_lrstates.begin(), endstate = m_lrstates.end(); curstate != endstate; ++curstate )
	{
		for( vector<LRItem>::iterator curitem = (*curstate)->m_items.begin(), enditem = (*curstate)->m_items.end(); curitem != enditem; ++curitem )
		{
			if( curitem->GetSymbolAtPosition() )
                continue;
            obs_clear(ss);
            curitem->m_follows.Print(ss);
            m_verbosedb->UpdateItem(item_rowids[curitem], obs_getstr(ss));
        }
    }
}

void Grammar::ComputeFollows()
{
    outbuf *verbose = m_bLogFollows ? m_verbose : 0;
    if( verbose || m_verbosedb )
    {
        ComputeFollows_Verbose();
        return;
    }
	vector<Symbol*> parentsymbols;
	vector<followdep> deps;
	for( vector<LRStateSet*>::iterator curstate = m_lrstates.begin(), endstate = m_lrstates.end(); curstate != endstate; ++curstate )
	{
		for( vector<LRItem>::iterator curitem = (*curstate)->m_items.begin(), enditem = (*curstate)->m_items.end(); curitem != enditem; ++curitem )
		{
			if( curitem->GetSymbolAtPosition() )
				continue;
			if( curitem->m_p->m_nt->IsStartSymbol() )
			{
				curitem->m_follows += m_endmarker;
				continue; // Nothing follows accept
			}
			lritem_ptrset postreduceitems = curitem->GetPostReduceItems();
			for( lritem_ptrset::iterator curpritem = postreduceitems.begin(), endpritem = postreduceitems.end(); curpritem != endpritem; ++curpritem )
			{
				curitem->m_follows += (*curpritem)->GetFollowsAtPosition();
				if( ! (curitem->m_follows & m_epsilon) )
					continue;
				curitem->m_follows -= m_epsilon;
				lritem_ptrset reducibleitems = (*curpritem)->GetReducibleItems();
				for( lritem_ptrset::iterator curreducibleitem = reducibleitems.begin(), endreducibleitem = reducibleitems.end(); curreducibleitem != endreducibleitem; ++curreducibleitem )
					deps.append( followdep(&(*curreducibleitem)->m_follows, &curitem->m_follows, (int)deps.size()) );
			}
		}
	}

	deps.unique( simplecmp<followdep> );
	deps.sort( followdep::bysortindex );
	SymbolSet copied;
	int go = 1;
	while( go )
	{
		go = 0;
		for( vector<followdep>::iterator curdep = deps.begin(); curdep != deps.end(); ++curdep )
		{
			if( curdep->update() )
				go = 1;
		}
	}

	for( vector<LRStateSet*>::iterator curstate = m_lrstates.begin(), endstate = m_lrstates.end(); curstate != endstate; ++curstate )
	{
		for( vector<LRItem>::iterator curitem = (*curstate)->m_items.begin(), enditem = (*curstate)->m_items.end(); curitem != enditem; ++curitem )
		{
			if( curitem->m_follows.m_chars.count() == 0 )
				continue;
			Symbol *cset = FindOrAddCharset(curitem->m_follows.m_chars);
            SymbolSet& follows = curitem->m_follows;
			follows.m_chars.clear();
			follows += cset;
		}
	}
}


void Grammar::PrintFollows(outbuf *pstream) const
{
    SymbolSet tmp;
	for( vector<LRStateSet*>::const_iterator curstate = m_lrstates.begin(), endstate = m_lrstates.end(); curstate != endstate; ++curstate )
	{
		for( vector<LRItem>::iterator curitem = (*curstate)->m_items.begin(), enditem = (*curstate)->m_items.end(); curitem != enditem; ++curitem )
		{
			if( curitem->GetSymbolAtPosition() )
				continue;
			ob_printf(pstream, "FOLLOW(State %d, Rule %d: [", (*curstate)->m_number, curitem->m_p->GetIndex());
			curitem->Print(pstream);
			ob_puts(pstream, "]) = ");
			tmp = curitem->m_follows;
			tmp.FlattenCharsets();
			tmp.Print(pstream);
			ob_putc(pstream, '\n');
		}
	}
}

void Grammar::GenerateActions()
{
    // Generate the initial action table.
    m_ranges.clear();
    m_rangesbyidx.clear();
    for( vector<LRStateSet*>::iterator curstate = m_lrstates.begin(); curstate != m_lrstates.end(); ++curstate )
    {
        (*curstate)->GenerateActions(m_ranges, m_rangesbyidx);
        if( m_errcount )
            break;
    }
}

static bool IsResolvedConflict(vector<LRAction>::iterator action, vector<LRAction>::iterator otheraction)
{
    if( action->IsRuntimeResolvable() && otheraction->IsRuntimeResolvable() && action->m_atype == otheraction->m_atype && (action->m_atype == ATypeReduce || action->m_atype == ATypeAccept) )
        return true;
    return false;
}

void Grammar::CollectConflicts()
{
    SymbolSet cset_intersection, flat_lhs, flat_rhs;

    for( vector<LRStateSet*>::iterator state = m_lrstates.begin(); state != m_lrstates.end(); ++state )
    {
        for( vector<LRAction>::iterator action = (*state)->m_actions.begin(); action != (*state)->m_actions.end(); ++action )
        {
			for( vector<LRAction>::iterator otheraction = action+1; otheraction != (*state)->m_actions.end(); ++otheraction )
            {
				if( action->m_gtype != otheraction->m_gtype )
					continue;
                if( action->m_gtype == GTypeProduction )
				{
					if( action->m_p != otheraction->m_p )
						continue;
                    if( IsResolvedConflict(action, otheraction) )
                        m_ActionResolvedConflicts.push_back( ActionConflict(action->m_state, action->m_atype, otheraction->m_atype, GTypeProduction, SymbolSet(), otheraction->m_p, action->m_itemidx, otheraction->m_itemidx) );
                    else
                        m_ActionConflicts.push_back( ActionConflict(action->m_state, action->m_atype, otheraction->m_atype, GTypeProduction, SymbolSet(), otheraction->m_p, action->m_itemidx, otheraction->m_itemidx) );
					continue;
				}
				// else curgtype == GTypeSymbol
				flat_lhs = action->m_s->m_firsts;
				flat_lhs.FlattenCharsets();
				flat_rhs = otheraction->m_s->m_firsts;
				flat_rhs.FlattenCharsets();
                cset_intersection = flat_lhs & flat_rhs;
				if( ! cset_intersection.count() )
					continue;
                if( IsResolvedConflict(action, otheraction) )
                    m_ActionResolvedConflicts.push_back( ActionConflict(action->m_state, action->m_atype, otheraction->m_atype, GTypeSymbol, cset_intersection, 0, action->m_itemidx, otheraction->m_itemidx) );
                else
                    m_ActionConflicts.push_back( ActionConflict(action->m_state, action->m_atype, otheraction->m_atype, GTypeSymbol, cset_intersection, 0, action->m_itemidx, otheraction->m_itemidx) );
            }
        }
    }
}

void Grammar::TokenStateMissReport(outbuf *pio) const
{
    if( ! m_TokenSetMerge )
        return;
    m_TokenSetMerge->Report(pio);
}

void Grammar::PrintLog(outbuf *pio) const
{
    // Token State Miss Report
    TokenStateMissReport(pio);

    // Conflict report
    ob_printf(pio, "Conflict Report (%d Conflicts)\n", (int)m_ActionConflicts.size());
    for( vector<ActionConflict>::const_iterator cur = m_ActionConflicts.begin(), end = m_ActionConflicts.end(); cur != end; ++cur )
    {
        ob_printf(pio, "State %d has a %s/%s conflict on symbol ", cur->m_state->m_number, cur->ActionType(), cur->OtherActionType());
        if( cur->m_gtype == GTypeSymbol )
            cur->m_set.Print(pio);
        else
            cur->m_p->Print(pio);
        ob_putc(pio, '\n');
    }
    ob_putc(pio, '\n');

    // Resolved Conflict report
    ob_printf(pio, "Resolved Conflict Report (%d Resolved Conflicts)\n", (int)m_ActionResolvedConflicts.size());
    for( vector<ActionConflict>::const_iterator cur = m_ActionResolvedConflicts.begin(), end = m_ActionResolvedConflicts.end(); cur != end; ++cur )
    {
        ob_printf(pio, "State %d has a resolved %s/%s conflict on symbol ", cur->m_state->m_number, cur->ActionType(), cur->OtherActionType());
        if( cur->m_gtype == GTypeSymbol )
            cur->m_set.Print(pio);
        else
            cur->m_p->Print(pio);
        ob_putc(pio, '\n');
    }
    ob_putc(pio, '\n');

    // Full grammar report
	ob_puts(pio, "Full grammar report:\n");
    ob_printf(pio, "%d State(s), %d Production(s)\n", m_lrstates.size(), m_productions.size());
    for( vector<Production*>::const_iterator p = m_productions.begin(); p != m_productions.end(); ++p )
    {
        ob_printf(pio, "%d : ", (*p)->GetIndex());
        (*p)->Print(pio);
        ob_putc(pio, '\n');
    }
    ob_puts(pio, "\n");

    // States and actions report
    for( vector<LRStateSet*>::const_iterator state = m_lrstates.begin(); state != m_lrstates.end(); ++state )
    {
        (*state)->Print(pio);
        ob_putc(pio, '\n');
        ob_puts(pio, "Actions:\n");
        for( vector<LRAction>::iterator action = (*state)->m_actions.begin(); action != (*state)->m_actions.end(); ++action )
        {
            action->Print(pio);
            ob_putc(pio, '\n');
        }
        ob_putc(pio, '\n');
    }
}

OrderedTokenSet *Grammar::FindTokenSet(const String name)
{
    if( ! name.length() )
        return 0;
    for( vector<OrderedTokenSet*>::iterator cur = m_TokenSets.begin(); cur != m_TokenSets.end(); ++cur )
    {
        OrderedTokenSet *pts = *cur;
        if( stricmp(name.c_str(), pts->m_Name.c_str()) == 0 )
            return pts;
    }
    return 0;
}

OrderedTokenSet *Grammar::FindOrAddTokenSet(const String name)
{
    if( ! name.length() )
        return 0;
    OrderedTokenSet *ret = FindTokenSet(name);
    if( ret )
        return ret;
    ret = new OrderedTokenSet(this);
    ret->m_Name = name;
    m_TokenSets.append(ret);
	if( stricmp(name.c_str(), "Initial") == 0 )
		m_InitialTokenState = ret;
	return ret;
}

void Grammar::AddDfa(Nfa &dfa, vector<Symbol*> &accepttokens, OrderedTokenSet *tokstate)
{
    vector<Symbol*> nts;
    vector<int> from;
    vector<EdgeSet> edges;
    nts.resize(accepttokens.size());
    for( int i = 0, n = accepttokens.size(); i < n; ++i )
    {
        Symbol *newnt = AddAnonymousNonterminal(tokstate);
        newnt->m_only_child = true;
        nts[i] = newnt;
    }
    // Add the rule to accept the start state
    Symbol *starttok = nts[dfa.m_startstates[0]];
    AddSimpleProduction(starttok, 0);
    // Add the productions for the state transitions.
    for( int i = 0, n = dfa.m_transitions.size(); i < n; ++i )
    {
        NfaTransition &cur = dfa.m_transitions[i];
        Symbol *tsym = FindOrAddCharset(cur.m_on);
        AddSimpleProduction(nts[cur.m_to], nts[cur.m_from], tsym, 0);
    }
    // Add the rules to accept the tokens, and follow restrictions so we accept the longest one.
    for( int i = 0, n = accepttokens.size(); i < n; ++i )
    {        
        Symbol *tok = accepttokens[i];
        if( ! tok )
            continue;
        from.clear();
        from.append(i);
        dfa.discover_edges(from, edges);
        if( edges.size() == 0 )
            AddSimpleProduction(tok, nts[i], 0);
        else
        {
            Symbol *intermediate_nt = AddAnonymousNonterminal(tokstate);
            //intermediate_nt->only_child = true;
            AddSimpleProduction(intermediate_nt, nts[i], 0);
            AddSimpleProduction(tok, intermediate_nt, 0);
			UIntSet edgechars;
			for( vector<EdgeSet>::iterator curedge = edges.begin(), endedge = edges.end(); curedge != endedge; ++curedge )
				edgechars += curedge->m_inputs;
            Symbol *edgesym = FindOrAddCharset(edgechars);
            intermediate_nt->AddFollowRestriction(edgesym);
        }
    }
}

void Grammar::PrintProductions(outbuf *pio) const
{
    int i = 1;
    ob_printf(pio, "%d Productions\n", m_productions.size());
    for( vector<Production*>::const_iterator cur = m_productions.begin(); cur != m_productions.end(); ++cur )
    {
        ob_printf(pio, "%3d : ", i++);
        (*cur)->Print(pio);
        ob_putc(pio, '\n');
    }
}

void Grammar::RemoveProduction(int index)
{
	Production *p = m_productions[index];
	p->m_nt->m_OwningProductions.remove(p);
    m_productions.removeat(index, 1);
    delete p;
    // How annoying, we've invalidated the map.
    m_indexmap.clear();
}

void Grammar::RemoveSymbol(Symbol *s)
{
    m_symbols.remove(s);
    m_literals.remove((Literal*)s);
    m_regexs.remove((Regex*)s);
    m_charsets.remove((CharSet*)s);
    m_nonterminals.remove((Nonterminal*)s);
    m_terminals.remove((Terminal*)s);
    delete s;
}

UIntSet Grammar::LookupNamedCharSet(String name, bool *pbFound)
{
    if( pbFound )
        *pbFound = false;
    for( vector<NamedCharSet>::iterator cur = m_namedcharsets.begin(), end = m_namedcharsets.end(); cur != end; ++cur )
    {
        if( cur->m_name == name )
        {
            if( pbFound )
                *pbFound = true;
            return cur->chars();
        }
    }
    return UIntSet();
}

void Grammar::AddPriorities(PrioritySet &priorities)
{
	m_priorities.push_back(priorities);
    m_priorities.back().FixatePriorityIndices();
}

void Grammar::AddAssociationsByLabel(ProductionOrLabel parentlabel, ProductionOrLabel childlabel, Assoc::AssocStyle assocstyle, Assoc::Association assoc, const vector<IntSetRange> &assoc_pos, const vector<LRItem> &from_series, const String fname, int lineno)
{
	m_assocbylabels.push_back( AssocByLabel(parentlabel, childlabel, assocstyle, assoc, assoc_pos, from_series, fname, lineno) );
}

ProductionLabel *Grammar::FindOrAddLabel(String label)
{
    for( vector<ProductionLabel*>::iterator cur = m_labels.begin(); cur != m_labels.end(); ++cur )
    {
        ProductionLabel *pl = *cur;
        if( pl->m_Name == label )
            return pl;
    }
    ProductionLabel *pl = new ProductionLabel(this, label);
    m_labels.append(pl);
    return pl;
}

bool Grammar::TestForPriorityConflict(ProductionLabel *parentlabel, ProductionLabel *childlabel)
{
    for( vector<PrioritySet>::iterator priority = m_priorities.begin(); priority != m_priorities.end(); ++priority )
    {
        int parentindex = priority->GetPriorityIndex(parentlabel);
        if( parentindex == -1 )
            continue;
        int childindex = priority->GetPriorityIndex(childlabel);
        if( childindex == -1 )
            continue;
        // Lower values indicate higher precedence.
        if( parentindex < childindex )
            return true; // There is a priority conflict, i.e. it was declared that parent > child
    }
    return false;
}

vector<Assoc>::iterator Grammar::FindOrCreateAssociation(Production *parent, Production *child, Assoc::AssocStyle assocstyle, const vector<LRItem> &from_series, String fname, int lineno)
{
    assoc_map::KeyType key(parent, child);
    Assocs &assocs = m_assocs[key];
    return assocs.FindOrCreateAssociation(parent, child, assocstyle, from_series, fname, lineno);
}

const Assocs *Grammar::GetAssocs(const Production *parent, const Production *child) const
{
    assoc_map::KeyType key(parent, child);
    assoc_map::const_iterator assocs = m_assocs.find(key);
    if( assocs == m_assocs.end() )
        return 0;
    return &assocs->second;
}

void Grammar::MoveProduction(Production *p, int newindex)
{
    if( newindex < 0 )
        newindex = m_productions.size()-1;
    int oldindex = p->GetIndex();
    if( oldindex == newindex )
        return;
    m_productions.remove(p);
    if( newindex > (int)m_productions.size() )
        newindex = m_productions.size();
    if( newindex > oldindex )
        --newindex;
	m_productions.insertat(newindex, p);
    // How annoying, we've invalidated the map.
    m_indexmap.clear();
}

void Grammar::SetOption(const char *option, const char *op, const char *value)
{
    // TODO
}

bool Grammar::DeclareTokFrag(String name, Rx *rx)
{
    if( ! rx )
        return false;
    const char *s = name.c_str();
    size_t len = name.length();
    if( len && s[0] == '{' && s[len-1] == '}' )
        name = name.substr(1,-1);
    tokfragmap::iterator cur = m_tokfragmap.find(name);
    if( cur != m_tokfragmap.end() )
        return false;
    m_tokfragmap[name] = rx;
    return true;
}

Rx *Grammar::LookupTokFrag(String name)
{
    const char *s = name.c_str();
    size_t len = name.length();
    if( len && s[0] == '{' && s[len-1] == '}' )
        name = name.substr(1,-1);
    tokfragmap::iterator cur = m_tokfragmap.find(name);
    if( cur == m_tokfragmap.end() )
        return 0;
    return cur->second;
}

void AllProperties::operator|=(AllProperties &rhs)
{
    m_flags |= rhs.m_flags;
    if( rhs.m_flags & PROP_FLAGS_RANK )
        m_rank = rhs.m_rank;
    if( rhs.m_flags & PROP_FLAGS_ASSOC )
        m_assoc_pos.insert(m_assoc_pos.end(), rhs.m_assoc_pos.begin(), rhs.m_assoc_pos.end());
    if( rhs.m_flags & PROP_FLAGS_LABEL )
        m_label = rhs.m_label;
    if( rhs.m_flags & PROP_FLAGS_ACTION )
        m_ReduceCode = rhs.m_ReduceCode;
    if( rhs.m_flags & PROP_FLAGS_RUNTIME_RESOLVABLE )
        m_flags |= PROP_FLAGS_RUNTIME_RESOLVABLE;
    m_from_series.append(rhs.m_from_series);
}

NamedCharSet::NamedCharSet()
{
}

NamedCharSet::NamedCharSet(String name, const UIntSet &charset) : m_name(name), m_chars(charset)
{
}

UIntSet NamedCharSet::chars()
{
	return m_chars;
}

assocmapkey::assocmapkey(const assocmapkey &rhs) : m_parent(rhs.m_parent), m_child(rhs.m_child)
{
}

assocmapkey::assocmapkey(const Production *parent, const Production *child)
    : m_parent(parent), m_child(child)
{
}

bool assocmapkey::operator==(const assocmapkey &rhs) const
{
    return m_parent == rhs.m_parent && m_child == rhs.m_child;
}

bool assocmapkey::operator!=(const assocmapkey &rhs) const
{
    return ! operator==(rhs);
}

bool LRItem_eq_Exact::operator ()(LRItem* const &lhs, LRItem* const &rhs) const
{
    return lhs->m_ss == rhs->m_ss && lhs->m_p == rhs->m_p && lhs->m_pos == rhs->m_pos;
}

uint32_t LRItem_hash_Exact::operator()(LRItem* const &p) const
{
#if UINTPTR_MAX == 0xffffffff
    uint32_t u3[3] = {(uint32_t)p->m_ss, (uint32_t)p->m_p, (uint32_t)p->m_pos};
    return hash_uint32s(u3, 3, 101);
#elif UINTPTR_MAX == 0xffffffffffffffff
#define LOW_UINT32(uu) ((uint32_t)(((uint64_t)(uu))&0xffffffff))
#define HIGH_UINT32(uu) ((uint32_t)((((uint64_t)(uu))>>(uint64_t)32)&0xffffffff))
    uint32_t u5[5] = {LOW_UINT32(p->m_ss), HIGH_UINT32(p->m_ss), LOW_UINT32(p->m_p), HIGH_UINT32(p->m_p), (uint32_t)p->m_pos};
    return hash_uint32s(u5, 5, 101);
#else
    what?
#endif
}

SymbolSet::SymbolSet()
{
}

void Symbol::ResetFirsts()
{
	m_firsts.clear();
}


SymbolSet::SymbolSet(const SymbolSet &rhs)
{
    m_chars = rhs.m_chars;
    m_symbols = rhs.m_symbols;
}

SymbolSet &SymbolSet::operator=(const SymbolSet &rhs)
{
    m_chars = rhs.m_chars;
    m_symbols = rhs.m_symbols;
    return *this;
}

void SymbolSet::operator+=(const SymbolSet &rhs)
{
    m_chars += rhs.m_chars;
    m_symbols.insert(rhs.m_symbols.begin(), rhs.m_symbols.end());
}

void SymbolSet::operator+=(const UIntSet &rhs)
{
    m_chars += rhs;
}

void SymbolSet::operator+=(Symbol *rhs)
{
    m_symbols.insert(rhs);
}

void SymbolSet::operator-=(const SymbolSet &rhs)
{
    m_chars -= rhs.m_chars;
    if( m_symbols.size() && rhs.m_symbols.size() )
    {
        sortedset<Symbol*>::iterator last = std::set_difference(m_symbols.begin(), m_symbols.end(), rhs.m_symbols.begin(), rhs.m_symbols.end(), m_symbols.begin());
	    m_symbols.erase(last, m_symbols.end());
    }
}

SymbolSet SymbolSet::operator-(const SymbolSet &rhs) const
{
	SymbolSet ss = *this;
	ss -= rhs;
	return ss;
}

void SymbolSet::operator-=(const UIntSet &rhs)
{
    m_chars -= rhs;
}

void SymbolSet::operator-=(Symbol *rhs)
{
    m_symbols.erase(rhs);
}

bool SymbolSet::operator&(unsigned int rhs) const
{
    return m_chars & rhs;
}

bool SymbolSet::operator&(const Symbol *rhs) const
{
    return  m_symbols.find(const_cast<Symbol*>(rhs)) != m_symbols.end();
}

SymbolSet SymbolSet::operator&(const SymbolSet &rhs) const
{
	SymbolSet ret;
	ret.m_chars = m_chars & rhs.m_chars;
	if( m_symbols.size() > 0 && rhs.m_symbols.size() > 0 )
	{
		vector<Symbol*> out;
		out.resize( m_symbols.size() + rhs.m_symbols.size() );
		vector<Symbol*>::iterator outit = out.begin();
        vector<Symbol*>::iterator last = std::set_intersection(m_symbols.begin(), m_symbols.end(), rhs.m_symbols.begin(), rhs.m_symbols.end(), outit);
		for( vector<Symbol*>::iterator cur = out.begin(); cur != last; ++cur )
			ret.m_symbols.insert(*cur);
	}
	return ret;
}

void SymbolSet::operator&=(const SymbolSet &rhs)
{
	SymbolSet x = *this & rhs;
	*this = x;
}

bool SymbolSet::operator==(const SymbolSet &rhs) const
{
	return (m_chars == rhs.m_chars && m_symbols == rhs.m_symbols);
}

bool SymbolSet::operator!=(const SymbolSet &rhs) const
{
	return ! (*this == rhs);
}

void SymbolSet::clear()
{
	m_chars.clear();
	m_symbols.clear();
}

int SymbolSet::count()
{
    return m_chars.count() + m_symbols.size();
}

void SymbolSet::Print(outbuf *p) const
{
	bool bPrinted = false;
    for( sortedset<Symbol*>::const_iterator s = m_symbols.begin(); s != m_symbols.end(); ++s )
    {
		if( bPrinted )
			ob_putc(p, ',');
        (*s)->Print(p);
		bPrinted = true;
    }
	bool bPrintCharSet = m_chars.count() ? true : false;
	if( bPrinted && bPrintCharSet )
		ob_putc(p, ',');
	if( bPrintCharSet )
	{
		bPrinted = true;
		CharSet cset(0, m_chars);
		cset.Print(p);
	}
	if( ! bPrinted )
		ob_puts(p, "{}");
}

void SymbolSet::FlattenCharsets()
{
    sortedset<Symbol*> removed;    
    for( sortedset<Symbol*>::iterator s = m_symbols.begin(); s != m_symbols.end(); ++s )
    {
        if( ! (*s)->IsCharSet() )
            continue;
        *this += (*s)->m_firsts;
        removed.insert(*s);
    }
    if( m_symbols.size() && removed.size() )
    {
        sortedset<Symbol*>::iterator last = std::set_difference(m_symbols.begin(), m_symbols.end(), removed.begin(), removed.end(), m_symbols.begin());
	    m_symbols.erase(last, m_symbols.end());
    }
}

OrderedTokenSet::OrderedTokenSet(Grammar *g) : m_g(g), m_bHideState(false), m_ubigdfa(0), m_bigdfa(0)
{
}

void OrderedTokenSet::Print(outbuf *pio) const
{
	if( m_bHideState )
		return;
    ob_printf(pio, "<%s>", m_Name.c_str());
}

void OrderedTokenSet::vAddTokensAndRules(Symbol *s, TokenSetAction::TokenSetActionType action, OrderedTokenSet *to, const char *file, int lineno, ...)
{
    if( action != TokenSetAction::NOACTION )
        m_actions.m_actions.push_back(TokenSetAction(file, lineno, s, action, to));
    m_symbols.push_back(s);
    va_list args;
    va_start(args, lineno);
    while(1)
    {
        s = va_arg(args, Symbol*);
        if( ! s )
            break;
        m_symbols.push_back(s);
        action = (TokenSetAction::TokenSetActionType)va_arg(args, int);
        to = va_arg(args, OrderedTokenSet*);
        file = va_arg(args, const char*);
		if( ! file )
			file = "";
        lineno = va_arg(args, int);
        if( action != TokenSetAction::NOACTION )
            m_actions.m_actions.push_back(TokenSetAction(file, lineno, s, action, to));
    }
    va_end(args);
}

void OrderedTokenSet::AddTokensAndRules(vector<TokenSetAction> &actions)
{
    for( vector<TokenSetAction>::iterator cur = actions.begin(); cur != actions.end(); ++cur )
    {
        if( cur->m_action != TokenSetAction::NOACTION )
            m_actions.m_actions.push_back(*cur);
        m_symbols.push_back(cur->m_sym);
    }
}

int OrderedTokenSet::GetIndex()
{
    return m_g->GetOrderedTokenSetIndex(this);
}

// Digest the list of nfas, and create a map from nfa state to accept symbols
void GetNfaAcceptStates(hashmap<int, Symbol*> &acceptstates, vector<Symbol*> &symbols, vector<Nfa> &nfas)
{
	acceptstates.clear();
	int stateadjust = 0;
	for( size_t i = 0, n = nfas.size(); i < n; ++i )
	{
		Nfa &nfa = nfas[i];
		Symbol *s = symbols[i];
		for( vector<int>::iterator cur = nfa.m_endstates.begin(), end = nfa.m_endstates.end(); cur != end; ++cur )
		{
			int endstate = *cur;
			acceptstates[endstate+stateadjust] = s;
		}
		stateadjust += nfa.statecount();
	}
}

// Create a map from dfa state to accept symbols
void GetDfaAcceptStates(hashmap<int, Symbol*> &acceptstates, Nfa &dfa, hashmap<int, Symbol*> &dfaacceptstates, vector< vector<int> > &allsubstates)
{
	acceptstates.clear();
	for( int node = 0, n = dfa.statecount(); node < n; ++node )
	{
		vector<int> &substates = allsubstates[node];
		for( vector<int>::iterator cur = substates.begin(), end = substates.end(); cur != end; ++cur )
		{
			int isubstate = *cur;
			hashmap<int, Symbol*>::iterator accepter = dfaacceptstates.find(isubstate);
			if( accepter == dfaacceptstates.end() )
				continue;
			acceptstates[node] = accepter->second;
			break;
		}
	}
}

class NfaLabelHelper : public NfaDiagramHelper
{
protected:
	hashmap<int, Symbol*> &m_acceptstates;
public:
	NfaLabelHelper(hashmap<int, Symbol*> &acceptstates)
		: m_acceptstates(acceptstates)
	{
	}

	// We want to print the state number, and if an end state, the token we accept.
	virtual String NodeLabel(Nfa &nfa, int node)
	{
        char buf[64];
        sprintf(buf, "%d", node);
        String ret = buf;
		hashmap<int, Symbol*>::iterator acceptsymbol = m_acceptstates.find(node);
		if( acceptsymbol != m_acceptstates.end() )
		{
			ret += "\\n(accept ";
            outbuf *out = new_string_outbuf();
			acceptsymbol->second->Print(out);
			String sname = obs_getstr(out);
            free_ob(out);
			ret += sname.replace("\\","\\\\\\\\").replace("\"","\\\"");
			ret += ")";
		}
		return ret;
	}

	virtual String TransitionLabel(Nfa &nfa, NfaTransition &trns)
	{
        outbuf *out = new_string_outbuf();
		SymbolSet ss;
		ss += trns.m_on;
		ss.Print(out);
		String ret = obs_getstr(out);
        free_ob(out);
		return ret.replace("\\","\\\\\\\\").replace("\"","\\\"");
	}

	virtual String ETransitionLabel(Nfa &nfa, NfaETransition &etrns)
	{
		return "{}";
	}
};

class DfaLabelHelper : public NfaLabelHelper
{
	vector< vector<int> > &m_substates;
	hashmap<int, Symbol*> &m_dfaacceptstates;
public:
	DfaLabelHelper(	hashmap<int, Symbol*> &nfaacceptstates,
					hashmap<int, Symbol*> &dfaacceptstates,
					vector< vector<int> > &substates)
		: NfaLabelHelper(nfaacceptstates), m_dfaacceptstates(dfaacceptstates), m_substates(substates)
	{
	}

	// Give the state number, the aggregates nfa states, and the accepted token if any.
	virtual String NodeLabel(Nfa &nfa, int node)
	{
		String ret;
		bool bUsedComma = false;
        char buf[68];
        sprintf(buf, "%d : {", node);
        ret = buf;
		vector<int> &substates = m_substates[node];
		for( vector<int>::iterator cur = substates.begin(), end = substates.end(); cur != end; ++cur )
		{
			if( bUsedComma )
				ret += ",";
			else
				bUsedComma = true;
			int isubstate = *cur;
            char buf[64];
            sprintf(buf, "%d", isubstate);
			ret += buf;
		}
		ret += "}";
		hashmap<int, Symbol*>::iterator accepter = m_dfaacceptstates.find(node);
		if( accepter != m_dfaacceptstates.end() )
		{
			ret += "\\n(accept ";
			outbuf *out = new_string_outbuf();
			accepter->second->Print(out);
			String sname = obs_getstr(out);
            free_ob(out);
			ret += sname.replace("\\","\\\\\\\\").replace("\"","\\\"");
			ret += ")";
		}
		return ret;
	}
};

static int cmpstr(int &lmatch, vector<int> &data, int klhs, int llhs, int krhs, int lrhs, bool length_settles_tie)
{
    int L = (llhs < lrhs) ? llhs : lrhs;
    for( int i = 0; i < L; ++i )
    {
        int res = data[klhs+i] - data[krhs+i];
        if( res != 0 )
        {
            lmatch = i;
            return res;
        }
    }
    lmatch = L;
    if(length_settles_tie)
        return llhs-lrhs;
    return 0;
}

// data contains both the window and the lookahead data
// lxwndlen is the seperator between data and the look-ahead
// krhs might be before or after the seprator, but won't stradle it.
static void find_insert_point(int &out_longestmatch, int &out_longestkey, int &out_mid,
                       vector<int> &data, int lxwndlen, int krhs, vector<int> &strings, vector<int> &shifty)
{
    int low = 0;
    int high = (int)strings.size()-1;
    int mid = 0;
    int longestmatch = 0;
    int longestkey = -1;
    int key = 0, cres = 0, mlen = 0;
    int lrhs = (krhs < lxwndlen) ? (lxwndlen-krhs) : ((int)data.size()-krhs);
    while( low <= high )
    {
        mid = (low+high)/2;
        key = strings[mid];
        cres = cmpstr(mlen, data, krhs, lrhs, key, lxwndlen-key, true);
        if( mlen > longestmatch )
        {
            longestmatch = mlen;
            longestkey = key;
        }
        if( cres < 0 )
            high = mid-1;
        else if( cres > 0 )
            low = mid+1;
        else
            break;
    }
    if( low > high )
    {
        out_longestmatch = longestmatch;
        out_longestkey = longestkey;
        out_mid = low;
        return;
    }
    out_longestmatch = longestmatch;
    out_longestkey = longestkey;
    out_mid = mid;
}

static void find_longest_match(int &out_longestmatch, int &out_longestkey, vector<int> &data, int lxwndlen, vector<int> &strings, vector<int> &shifty)
{
    // find the longest match of the data in the look-ahead to the data in the window.
    // lxwndsize is then the seperator, and the start of the string we are trying to find in the window.
    int out_mid = 0;
    find_insert_point(out_longestmatch, out_longestkey, out_mid, data, lxwndlen, lxwndlen, strings, shifty);
}

static void addstring(vector<int> &data, vector<int> &strings, vector<int> &shifty, int lxwndlen, int slen)
{
    // every string has gotten longer, we may be in for a resorting.
    // only strings that tie the string that followed it in sorting order, aside from length may shift.
    // when we insert, we check if it ties the next string in the list, in which case we'll have to
    // watch for order change events.
    
    // move the changed strings forward...
    int scount = strings.size();
    for( int i = shifty.size()-1; i >= 0; --i )
    {
        int eqprev = shifty[i];
        int key = strings[eqprev];
        int eqnext = eqprev+1;
        int cres = 0, mlen = 0;
        while( eqnext < scount )
        {
            int knext = strings[eqnext];
            cres = cmpstr(mlen, data, key, lxwndlen-key, knext, lxwndlen-knext, false);
            if( cres <= 0 )
                break;
            // It's bigger, compare it to the next string.
            eqnext += 1;
        }
        // It's bigger than everything, or not equal to the next string.
        // In either case, it's not equal to anything in the shifty list so remove it.
        if( eqnext == scount || cres != 0 )
            shifty.erase(shifty.begin()+i);
        if( eqnext == eqprev+1 )
            continue;
        strings.erase(strings.begin()+eqprev);
        strings.insert(strings.begin()+eqnext-1, key);
    }
    
    // insert the new string
    int key = lxwndlen-slen;
    int longestmatch = 0, longestkey = 0, insertpt = 0;
    find_insert_point(longestmatch, longestkey, insertpt, data, lxwndlen, key, strings, shifty);
    strings.insert(strings.begin()+insertpt, key);
    insertpt += 1;

    // track it as a shifty string, if needed.
    if( insertpt < (int)strings.size() )
    {
        int cres = 0, mlen = 0;
        int iptkey = strings[insertpt];
        cres = cmpstr(mlen, data, iptkey, lxwndlen-iptkey, key, slen, false);
        // it's shifty, we'll need to add it to the watch list.
        // since we keep the watch list sorted, we may need to update indices that follow the shifty insert point.
        if( cres == 0 )
        {
            int low = 0;
            int high = (int)shifty.size()-1;
            int mid = 0;
            while( low <= high )
            {
                mid = (low+high)/2;
                int mpt = shifty[mid];
                if( insertpt < mpt )
                    high = mid-1;
                else if( insertpt > mpt )
                    low = mid+1;
                break;
            }
            for( int i = mid, n = shifty.size(); i < n; ++i )
                shifty[i] = shifty[i] + 1;
            shifty.insert(shifty.begin()+mid, insertpt);
        }
    }
}

#ifdef VERIFY_SCANNER_COMPRESSION
static void expand_dfa(vector<compressed_lxitem> &lxout, vector<int> &data)
{
    data.clear();
    vector<int> tmp;
    for( vector<compressed_lxitem>::iterator curlx = lxout.begin(), endlx = lxout.end(); curlx != endlx; ++curlx )
    {
        if( curlx->count )
        {
            vector<int>::iterator wndstart = data.begin()+curlx->off;
            vector<int>::iterator wndend = wndstart+curlx->lead;
            tmp.clear();
            tmp.insert(tmp.end(), wndstart, wndend);
            int count = curlx->count;
            int lead = curlx->lead;
            while( count )
            {
                if( lead < count )
                {
                    data.insert(data.end(), tmp.begin(), tmp.begin()+lead);
                    count -= lead;
                }
                else
                {
                    data.insert(data.end(), tmp.begin(), tmp.begin()+count);
                    count = 0;
                }
            }
        }
        if( curlx->next != -1 )
            data.push_back(curlx->next);
    }
}
#endif

static void compress_dfa(vector<compressed_lxitem> &lxout, vector<int> &data)
{
    lxout.clear();
    vector<int> strings;
    vector<int> shifty;
    int lxwndsize = 0;
    int datasize = (int)data.size();
    
    while( lxwndsize < datasize )
    {
        int longestmatch = 0;
        int longestkey = -1;
        int pos = lxwndsize;
        find_longest_match(longestmatch, longestkey, data, lxwndsize, strings, shifty);
        if( longestmatch == 0 )
        {
            int c = data[lxwndsize++];
            addstring(data, strings, shifty, lxwndsize, 1);
            compressed_lxitem clxi = {pos, 0, 0, 0, c};
            lxout.push_back(clxi);
        }
        else
        {
            vector<int>::iterator start_windowmatch, end_windowmatch, cur_windowmatch;
            vector<int>::iterator start_datamatch, end_datamatch, cur_datamatch;
            int curmatchlen = 0;
            lxwndsize += longestmatch;
            curmatchlen += longestmatch;
            while( lxwndsize < datasize )
            {
                start_windowmatch = data.begin()+longestkey;
                cur_windowmatch = start_windowmatch;
                end_windowmatch = start_windowmatch+longestmatch;
                start_datamatch = data.begin()+lxwndsize;
                cur_datamatch = start_datamatch;
                end_datamatch = start_datamatch+longestmatch;
                while( cur_windowmatch != end_windowmatch )
                {
                    if( *cur_datamatch != *cur_windowmatch )
                        break;
                    ++cur_datamatch;
                    ++cur_windowmatch;
                }
                int mlen = cur_windowmatch-start_windowmatch;
                lxwndsize += mlen;
                curmatchlen += mlen;
                if( cur_windowmatch != end_windowmatch )
                    break;
            }
            int c = -1;
            int newstrlen = curmatchlen;
            if( lxwndsize < datasize )
            {
                newstrlen += 1;
                c = data[lxwndsize++];
                addstring(data, strings, shifty, lxwndsize, 1);
            }
            addstring(data, strings, shifty, lxwndsize, newstrlen);
            compressed_lxitem clxi = {pos, longestkey, curmatchlen, longestmatch, c};
            lxout.push_back(clxi);
        }
    }
}

void OrderedTokenSet::GenerateTokens()
{
	outbuf *nfastrm = 0;
	outbuf *dfastrm = 0;
	if( m_g->m_bLogDfa )
	{
        String nfadot = m_g->m_fname_root+"_"+m_Name+"_nfa.dot";
        String dfadot = m_g->m_fname_root+"_"+m_Name+"_dfa.dot";
        nfastrm = new_file_outbuf(nfadot.c_str());
		dfastrm = new_file_outbuf(dfadot.c_str());
	}
    Nfa bignfa;
    vector<Nfa> nfas;
    nfas.resize(m_symbols.size());
    // Create nfa's for each token, and combine them into a ginormous nfa.
    vector<Symbol*> symbols;
    for( vector<Symbol*>::iterator cur = m_symbols.begin(); cur != m_symbols.end(); ++cur )
    {
        Nfa &nfa = nfas[cur-m_symbols.begin()];
        if( (*cur)->AddToNfa(nfa) )
            bignfa.AddNfa(nfa);
        symbols.push_back(*cur);
    }
	hashmap<int,Symbol*> nfaacceptstates;
	GetNfaAcceptStates(nfaacceptstates, symbols, nfas);
	if( nfastrm )
	{
		NfaLabelHelper *pHelper = new NfaLabelHelper(nfaacceptstates);
		bignfa.Print(nfastrm, pHelper, m_Name);
		delete pHelper;
		free_ob(nfastrm);
	}
    // Convert the ginormous nfa into a ginormous dfa.
    vector< vector<int> > substates;
    delete m_bigdfa;
    m_ubigdfa = new Nfa();
    bignfa.AsDfa(*m_ubigdfa, substates);
    m_bigdfa = new FrozenDfa();
    m_ubigdfa->Freeze(*m_bigdfa);
    Nfa &bigdfa = *m_ubigdfa;
	hashmap<int,Symbol*> dfaacceptstates;
	GetDfaAcceptStates(dfaacceptstates, bigdfa, nfaacceptstates, substates);
	if( dfastrm )
	{
		DfaLabelHelper *pHelper = new DfaLabelHelper(nfaacceptstates, dfaacceptstates, substates);
		bigdfa.Print(dfastrm, pHelper, m_Name);
		delete pHelper;
		free_ob(dfastrm);
	}
    // Figure out which token should be accepted for each dfa accept state (first match in supplied token order wins)
    m_accepttokens.clear();
    m_accepttokens.resize(bigdfa.statecount(), 0);
	for( vector<int>::iterator cur = bigdfa.m_endstates.begin(); cur != bigdfa.m_endstates.end(); ++cur )
    {
		int i = *cur;
		if( dfaacceptstates.find(i) == dfaacceptstates.end() )
			continue;
		m_accepttokens[i] = dfaacceptstates[i];
    }

    // Prep for compression.
    hashset<int> hs_toksetbreaks;
    for( vector< vector<FrozenDfaTransition> >::iterator transitionset = m_bigdfa->m_transitions.begin(), endtransitionset = m_bigdfa->m_transitions.end(); transitionset != endtransitionset; ++transitionset )
    {
        int laststop = -1;
        for( vector<FrozenDfaTransition>::iterator transition = transitionset->begin(), endtransition = transitionset->end(); transition != endtransition; ++transition )
        {
            if( transition->m_start != laststop+1 )
                hs_toksetbreaks.insert(laststop+1);
            hs_toksetbreaks.insert(transition->m_start);
            laststop = transition->m_stop;
        }
        if(laststop == INT_MAX)
            laststop = MAX_CODEPOINT;
        hs_toksetbreaks.insert(laststop+1);
    }
    vector<int> &toksetbreaks = m_compressed_dfa_toksetbreaks;
    toksetbreaks.clear();
    toksetbreaks.reserve(hs_toksetbreaks.size());
    for( hashset<int>::iterator cur = hs_toksetbreaks.begin(), end = hs_toksetbreaks.end(); cur != end; ++cur )
        toksetbreaks.push_back(*cur);
    toksetbreaks.sort(simplecmp<int>);

    // assign each break an index
    hashmap<int,int> toksetmap;
    for(int i = 0, n = toksetbreaks.size(); i < n; ++i )
        toksetmap[toksetbreaks[i]] = i;

    int statecount = m_bigdfa->m_statecount;
    vector<int> breaks;
    for( vector< vector<FrozenDfaTransition> >::iterator transitionset = m_bigdfa->m_transitions.begin(), endtransitionset = m_bigdfa->m_transitions.end(); transitionset != endtransitionset; ++transitionset )
    {
        int laststopidx = 0;
        for( vector<FrozenDfaTransition>::iterator transition = transitionset->begin(), endtransition = transitionset->end(); transition != endtransition; ++transition )
        {
            FrozenDfaTransition &t = *transition;
            int startidx = toksetmap[t.m_start];
            int to = t.m_to;
            if( to == -1 )
                to = statecount;
            // fill the gaps between the last range and this range.
            for( int i = laststopidx; i < startidx; ++i )
                breaks.push_back(statecount);
            // populate the range
            int stopidx = 0;
            hashmap<int,int>::iterator stopiter = toksetmap.find(t.m_stop+1);
            if( stopiter == toksetmap.end() )
                stopidx = toksetbreaks.size();
            else
                stopidx = stopiter->second;
            for( int i = startidx; i < stopidx; ++i )
                breaks.push_back(to);
            laststopidx = stopidx;
        }
        // fill till the end
        for( int i = laststopidx, n = toksetbreaks.size(); i < n; ++i )
            breaks.push_back(statecount);
    }
                    
    // now let's do some compression...
    compress_dfa(m_compressed_dfa_data, breaks);
#ifdef VERIFY_SCANNER_COMPRESSION
    vector<int> breaks_out;
    expand_dfa(m_compressed_dfa_data, breaks_out);
    if( breaks != breaks_out )
    {
        m_g->m_errcount++;
        if( m_g->m_err )
            ob_printf(m_g->m_err, "error: scanner %d, while verifying compressed scanner\n", m_Name.c_str());
        return;
    }
#endif
 }

void OrderedTokenSet::AddBigDfaAsCfg()
{
    // Now just do a simple DFA to CFG, plus add the non-follow rules that make it work so nice.
    m_g->AddDfa(*m_ubigdfa, m_accepttokens, this);
}

void OrderedTokenSet::CollectUsedItems(hashset<Production*> &usedproductions, hashset<Symbol*> &usedsymbols, hashset<OrderedTokenSet*> &usedtokensets)
{
	if( usedtokensets.find(this) != usedtokensets.end() )
		return;
	usedtokensets.insert(this);
    for( vector<Symbol*>::iterator cur = m_symbols.begin(); cur != m_symbols.end(); ++cur )
		(*cur)->CollectUsedItems(usedproductions, usedsymbols, usedtokensets);
}


TokenSetAction::TokenSetAction()
: m_action(NOACTION), m_actionset(0), m_lineno(-1), m_sym(0)
{
}

TokenSetAction::TokenSetAction(String file, int lineno, Symbol *sym, TokenSetActionType action, OrderedTokenSet *actionset)
: m_file(file), m_lineno(lineno), m_sym(sym), m_action(action), m_actionset(actionset)
{
}

bool TokenSetAction::operator==(const TokenSetAction &rhs) const
{
    if( m_action != rhs.m_action )
        return false;
    switch( m_action )
    {
        case NOACTION:
            break;
        case GOTO:
        case POP:
            if( m_sym != rhs.m_sym )
                return false;
            break;
        case PUSH:
            if( m_actionset != rhs.m_actionset || m_sym != rhs.m_sym )
                return false;
            break;
    }
    return true;
}

bool TokenSetAction::ApplyTo(Grammar *g, vector<OrderedTokenSet*> &statestack, outbuf *err)
{
    switch( m_action )
    {
        case GOTO:
            if( statestack.size() == 0 )
            {
                g->m_wrncount++;
                if( err )
                    ob_printf(err, "warning: %s(%d): Attempted a goto on an empty stack", m_file.c_str(), m_lineno);
                return false;
            }
            statestack.pop_back();
            statestack.push_back(m_actionset);
            break;
        case PUSH:
			statestack.push_back(m_actionset);
            break;
        case POP:
            if( statestack.size() <= 1 )
            {
                g->m_wrncount++;
                if( err )
                    ob_printf(err, "warning: %s(%d): Attempted a pop empty stack of depth %d\n", m_file.c_str(), m_lineno, statestack.size());
                return false;
            }
            statestack.pop_back();
            break;
    }
    return true;
}

void TokenSetAction::MergeTo(TokenSetActions &tsa)
{
    switch( m_action )
    {
        case GOTO:
            while( tsa.m_actions.size() > 0 && tsa.m_actions[-1].m_action == GOTO )
                tsa.m_actions.pop_back();
            tsa.m_actions.append(*this);
            tsa.m_actions[-1].m_sym = 0;
            break;
        case PUSH:
            tsa.m_actions.append(*this);
            tsa.m_actions[-1].m_sym = 0;
            break;
        case POP:
            while( tsa.m_actions.size() > 0 && tsa.m_actions[-1].m_action == GOTO )
                tsa.m_actions.pop_back();
            if( tsa.m_actions.size() == 0 || tsa.m_actions[-1].m_action == POP )
            {
                tsa.m_actions.append(*this);
                tsa.m_actions[-1].m_sym = 0;
            }
            else
                tsa.m_actions.pop_back();
            break;
    }
}

void TokenSetAction::Print(outbuf *pio) const
{
    switch( m_action )
    {
    case GOTO:
        ob_puts(pio, "    Goto ");
        m_actionset->Print(pio);
        ob_putc(pio, '\n');
        break;
    case PUSH:
        ob_puts(pio, "    Push ");
        m_actionset->Print(pio);
        ob_putc(pio, '\n');
        break;
    case POP:
        ob_puts(pio, "    Pop\n");
        break;
    default:
        break;
    }
}

TokenSetAction *TokenSetActions::FindTokenSetAction(Symbol *forsymbol)
{
    for( vector<TokenSetAction>::iterator curaction = m_actions.begin(); curaction != m_actions.end(); ++curaction )
    {
        if( curaction->m_sym == forsymbol )
            return &(*curaction);
    }
    return 0;
}

void TokenSetActions::MergeTo(TokenSetActions &tsa)
{
    for( vector<TokenSetAction>::iterator curaction = m_actions.begin(); curaction != m_actions.end(); ++curaction )
        curaction->MergeTo(tsa);
}

void TokenSetActions::ApplyTo(Grammar *g, vector<OrderedTokenSet*> &statestack, outbuf *err)
{
    for( vector<TokenSetAction>::iterator curaction = m_actions.begin(); curaction != m_actions.end(); ++curaction )
        curaction->ApplyTo(g, statestack, err);
}

void TokenSetActions::Print(outbuf *pio) const
{
    ob_puts(pio, "Actions:\n");
    if( m_actions.size() == 0 )
    {
        ob_puts(pio, "    (No Actions)\n");
        return;
    }
    for( vector<TokenSetAction>::const_iterator cur = m_actions.begin(); cur != m_actions.end(); ++cur )
        cur->Print(pio);
}

class Marker : public Symbol
{
public:
	Marker(const char *name) : Symbol(0, name, 0)
	{
	}
	SType GetSType() const { return TERMINAL_SYMBOL; }
};

// Don't call theseduring startup.  Or else KABOOM!!!
Symbol *Symbol::LB()
{
    static Marker _LB("[");
    return &_LB;
}

Symbol *Symbol::RB()
{
    static Marker _RB("]");
    return &_RB;
}

Symbol *Symbol::DOT()
{
    static Marker _DOT(".");
    return &_DOT;
}

Symbol::Symbol(Grammar *g, String name, OrderedTokenSet *tokstate)
{
    m_g = g;
    m_cloneof = 0;
    m_only_child = false;
    m_Name = name;
    m_TokenState = tokstate;
}

void Symbol::CollectUsedItems(hashset<Production*> &usedproductions, hashset<Symbol*> &usedsymbols, hashset<OrderedTokenSet*> &usedtokensets)
{
    if( usedsymbols.find(this) != usedsymbols.end() )
        return;
    usedsymbols.insert(this);
    if( m_TokenState )
		m_TokenState->CollectUsedItems(usedproductions, usedsymbols, usedtokensets);
    for( vector<Production*>::iterator cur = m_OwningProductions.begin(); cur != m_OwningProductions.end(); ++cur )
        (*cur)->CollectUsedItems(usedproductions, usedsymbols, usedtokensets);
}

void Symbol::Print(outbuf *pio) const
{
    ob_puts(pio, m_Name.c_str());
	if( m_TokenState )
		m_TokenState->Print(pio);
}

void Symbol::AddFollowRestriction(Symbol *s)
{
    if( IsTerminal() )
        return;
    if( s->IsCharSet() )
        m_follow_restrict += s->m_firsts;
    else
        m_follow_restrict += s;
}

void Symbol::MarkReachableNonterminalsAndCheckForProductionlessNonterminals(hashset<Nonterminal*> &nonterminals, outbuf *err)
{
}

int Symbol::GetIndex()
{
    return m_g->GetSymbolIndex(this);
}

Rx *Symbol::RxFromSymbol()
{
    return 0;
}

Literal::Literal(Grammar *g, String name, OrderedTokenSet *tokstate, bool CaseInsensitive)
: Symbol(g, name, tokstate), m_bCaseInsensitive(CaseInsensitive)
{
}

void Literal::ResetFirsts()
{
	Symbol::ResetFirsts();
	if( m_g->SolvingForScannerless() && m_Name.length() )
		m_firsts.m_chars += m_Name.c_str()[0];
	else
		m_firsts += this;
}

bool Literal::IsTerminal() const
{
	if( ! m_g->SolvingForScannerless() )
		return true;
	return false;
}

Rx *Literal::RxFromSymbol()
{
    Rx *ret = 0;
    const char *mbstr = m_Name.c_str();
    size_t mblen = m_Name.length();
    size_t read_count = 0;
    int err = 0;
    unsigned int c = 0;
    while(mblen)
    {
        c = mbget(mbstr, mblen, &read_count, &err);
        if( err )
            break;
        mbstr += read_count;
        mblen -= read_count;
        Rx *cset = new RxChars(c);
        if( ret )
            ret = new RxConcat(ret, cset);
        else
            ret = cset;
    }
    return ret;
}

Literal::~Literal()
{
}

void Literal::Print(outbuf *pio) const
{
	if( m_bCaseInsensitive )
		ob_puts(pio, "i\"");
	else
		ob_putc(pio, '\"');
    ob_puts(pio, m_Name.c_str());
    ob_putc(pio, '\"');
	if( m_TokenState )
		m_TokenState->Print(pio);
}

void Literal::GenerateToken()
{
    if( m_TokenState )
        return; // Was generated in OrderedTokenSet
    UIntSet chset;
    vector<Symbol*> symbols;
    vector<sl_decl> sldecls;
    const char *mbstr = m_Name.c_str();
    size_t mblen = m_Name.length();
    size_t read_count = 0;
    int err = 0;
    unsigned int c = 0;
    while(mblen)
    {
        int s0 = mbget(mbstr, mblen, &read_count, &err);
        if( err )
            break;
        mbstr += read_count;
        mblen -= read_count;
        chset.clear();
        if( m_bCaseInsensitive )
        {
            int lowch = glr_tolower(s0);
            int upch = glr_toupper(s0);
            chset += lowch;
            chset += upch;
        }
        else
        {
            chset += s0;
        }
        symbols.push_back( m_g->FindOrAddCharset(chset) );
        sldecls.push_back(sl_neutral);
    }
    m_g->AddProduction(this, symbols, sldecls, 0);
}

bool Literal::AddToNfa(Nfa &nfa)
{
    nfa.AddLiteral(m_Name.c_str(), m_bCaseInsensitive);
    return true;
}

Regex::Regex(Grammar *g, String name, OrderedTokenSet *tokstate, Rx *rx, bool bCaseInsensitive)
: Symbol(g, name, tokstate), m_rx(rx), m_bCaseInsensitive(bCaseInsensitive), m_bAliased(false)
{
}

Regex::~Regex()
{
    delete m_rx;
}

void Regex::Print(outbuf *pio) const 
{
	if( m_bCaseInsensitive )
		ob_puts(pio, "i/");
	else
		ob_putc(pio, '/');
    ob_puts(pio, m_Name.c_str());
    ob_putc(pio, '/');
	if( m_TokenState )
		m_TokenState->Print(pio);
}

void Rx_AddSymbols(Rx *rx, Grammar *g, Production *p)
{
    switch( rx->RxClassification() )
    {
    case Rx::Concat:
        {
            RxConcat *r = (RxConcat*)rx;
            // given [/ab/ -> rx]
            // add [a b -> /ab/]
            Rx_AddSymbols(r->m_lhs, g, p);
            Rx_AddSymbols(r->m_rhs, g, p);
            break;
        }
    case Rx::Union:
        {
            RxUnion *r = (RxUnion*)rx;
            // given a|b -> rx
            // add a -> a|b
            // add b -> a|b
            Symbol *a_or_b = g->AddAnonymousNonterminal();
            p->m_items.append(a_or_b); // a|b -> rx
            Production *p_a = g->AddSimpleProduction(a_or_b, 0);// p_a is [a -> a|b]
            Production *p_b = g->AddSimpleProduction(a_or_b, 0);// p_b is [b -> a|b]
            Rx_AddSymbols(r->m_lhs, g, p_a);
            Rx_AddSymbols(r->m_rhs, g, p_b);
            break;
        }
    case Rx::Star:
        {
            RxStar *r = (RxStar*)rx;
            // given [/a*/ -> rx]
            // add [-> /a*/]
            // add [/a+/ -> /a*/]
            // add [/a+/ a -> /a+/]
            // add [a -> /a+/]
            Symbol *a_star = g->AddAnonymousNonterminal();
            Symbol *a_plus = g->AddAnonymousNonterminal();
            p->m_items.append(a_star); // [/a*/ -> rx]
            g->AddSimpleProduction(a_star, 0);// [-> /a*/]
            g->AddSimpleProduction(a_star, a_plus, 0);// [/a+/ -> /a*/]
            Production *p2 = g->AddSimpleProduction(a_plus, 0);// p2 is [/a+/ a -> /a+/]
            Production *p3 = g->AddSimpleProduction(a_plus, 0);// p3 is [a -> /a+/]
            p2->m_items.append(a_plus);
            Rx_AddSymbols(r, g, p2);
            Rx_AddSymbols(r, g, p3);
            break;
        }
    case Rx::Plus:
        {
            RxPlus *r = (RxPlus*)rx;
            // given [/a+/ -> rx]
            // add [/a+/ a -> /a+/]
            // add [a -> /a+/]
            Symbol *a_plus = g->AddAnonymousNonterminal();
            p->m_items.append(a_plus);// [/a+/ -> rx]
            Production *p0 = g->AddSimpleProduction(a_plus, 0);// p0 is [/a+/ a -> /a+/]
            Production *p1 = g->AddSimpleProduction(a_plus, 0);// p1 is [a -> /a+/]
            p0->m_items.append(a_plus);
            Rx_AddSymbols(r, g, p0);
            Rx_AddSymbols(r, g, p1);
            break;
        }
    case Rx::Opt:
        {
            RxOpt *r = (RxOpt*)rx;
            // given [/a?/ -> rx]
            // add [-> /a?/]
            // add [a -> /a?/]
            Symbol *a_opt = g->AddAnonymousNonterminal();
            p->m_items.append(a_opt);// [/a?/ -> rx]
            g->AddSimpleProduction(a_opt, 0);// [-> /a?/]
            Production *p1 = g->AddSimpleProduction(a_opt, 0);// p1 is [a -> /a?/]
            Rx_AddSymbols(r, g, p1);
            break;
        }
    case Rx::Many:
        {
            RxMany *r = (RxMany*)rx;
            if( r->m_min == r->m_max ) // fixed number
            {
                for( int i = 0, n = r->m_min; i < n; ++i )
                    Rx_AddSymbols(r, g, p);
            }
            else if( r->m_min == -1 ) // zero through max
            {
                Symbol *s0 = g->AddAnonymousNonterminal();
                p->m_items.append(s0);
                for( int i = 0, n = r->m_max; i <= n; ++i )
                {
                    Production *pN = g->AddSimpleProduction(s0, 0);
                    for( int j = 0; j < i; ++j )
                        Rx_AddSymbols(r, g, pN);
                }
            }
            else if( r->m_max == -1 ) // min through any
            {
                Symbol *s0 = g->AddAnonymousNonterminal();
                Symbol *s1 = g->AddAnonymousNonterminal();
                p->m_items.append(s0);
                Production *p0 = g->AddSimpleProduction(s0, 0);
                p0->m_items.append(s1);
                for( int i = 0, n = r->m_min-1; i < n; ++i )
                    Rx_AddSymbols(r, g, p0);
                Production *p1 = g->AddSimpleProduction(s1, 0);
                Production *p2 = g->AddSimpleProduction(s1, 0);
                p1->m_items.append(s1);
                Rx_AddSymbols(r, g, p1);
                Rx_AddSymbols(r, g, p2);
            }
            else // min through max
            {
                Symbol *s0 = g->AddAnonymousNonterminal();
                p->m_items.append(s0);
                for( int i = r->m_min, n = r->m_max; i <= n; ++i )
                {
                    Production *pN = g->AddSimpleProduction(s0, 0);
                    for( int j = 0; j < i; ++j )
                        Rx_AddSymbols(r, g, pN);
                }
            }
            break;
        }
    case Rx::Chars:
        {
            RxChars *r = (RxChars*)rx;
            Symbol *s0 = g->FindOrAddCharset(r->m_chars);
            p->m_items.append(s0);
            break;
        }
    case Rx::Capture:
        {
            RxCapture *r = (RxCapture*)rx;
            Rx_AddSymbols(r->m_rx, g, p);
            break;
        }
    }
}

void Regex::GenerateToken()
{
    if( m_TokenState )
        return; // Was generated in OrderedTokenSet
    Production *p = m_g->AddSimpleProduction(this, 0);
    Rx_AddSymbols(m_rx, m_g, p);
}

bool Regex::AddToNfa(Nfa &nfa)
{
    nfa.AddRx(m_rx);
    return true;
}

void Regex::ResetFirsts()
{
	Symbol::ResetFirsts();
	if( ! m_g->SolvingForScannerless() )
		m_firsts += this;
}

bool Regex::IsTerminal() const
{
	if( ! m_g->SolvingForScannerless() )
		return true;
	return false;
}

bool Regex::SetFriendlyName(const char *friendlyname)
{
    if( ! friendlyname )
        friendlyname = "";
    if( m_bAliased )
    {
        if( m_Name == friendlyname )
            return true;
        return false;
    }
    m_Name = friendlyname;
    m_bAliased = true;
    return true;
}

Rx *Regex::RxFromSymbol()
{
    if( ! m_rx )
        return 0;
    return m_rx->Copy();
}

CharSet::CharSet(Grammar *g, UIntSet chars)
: Symbol(g, "", 0), m_chars(chars)
{
    m_firsts.m_chars = m_chars;
}

void CharSet::ResetFirsts()
{
	Symbol::ResetFirsts();
	m_firsts.m_chars = m_chars;
}

CharSet::~CharSet()
{
}

class NamedSetItem
{
public:
	NamedSetItem() {}
	NamedSetItem(const NamedSetItem &rhs)
	{
		namedset = rhs.namedset;
		name = rhs.name;
	}
	NamedSetItem &operator=(const NamedSetItem &rhs)
	{
		namedset = rhs.namedset;
		name = rhs.name;
		return *this;
	}
	NamedSetItem(const UIntSet &s, String n)
		: namedset(s), name(n) {}
	UIntSet namedset;
	String name;
};

static vector<NamedSetItem> *NamedSets()
{
	extern const UIntSet &named_charset_alpha();
	extern const UIntSet &named_charset_lower();
	extern const UIntSet &named_charset_upper();
	extern const UIntSet &named_charset_alnum();
	extern const UIntSet &named_charset_space();
	extern const UIntSet &named_charset_pspace();
	extern const UIntSet &named_charset_blank();
	extern const UIntSet &named_charset_digit();
	extern const UIntSet &named_charset_xdigit();
	static vector<NamedSetItem> namedsets;
	static bool bInitialized = false;
	if( ! bInitialized )
	{
		namedsets.push_back( NamedSetItem(named_charset_alnum(), "alnum") );
		namedsets.push_back( NamedSetItem(named_charset_alpha(), "alpha") );
		namedsets.push_back( NamedSetItem(named_charset_lower(), "lower") );
		namedsets.push_back( NamedSetItem(named_charset_upper(), "upper") );
		namedsets.push_back( NamedSetItem(named_charset_space(), "space") );
		namedsets.push_back( NamedSetItem(named_charset_pspace(), "pspace") );
		namedsets.push_back( NamedSetItem(named_charset_blank(), "blank") );
		namedsets.push_back( NamedSetItem(named_charset_xdigit(), "xdigit") );
		namedsets.push_back( NamedSetItem(named_charset_digit(), "digit") );
		bInitialized = true;
	}
	return &namedsets;
}

static void ob_rev_putc(outbuf *buf, int c, int maxc)
{
    if( ! c || ! maxc )
        return;
    ob_rev_putc(buf, c>>8, maxc-1);
    c &= 0x0f;
    if( c <= 9 )
        ob_putc(buf, c+'0');
    else
        ob_putc(buf, c-10+'a');
}

static void ob_printchar(outbuf *buf, int c)
{
    switch( c )
    {
    case 0:
        ob_puts(buf, "\\0");
        break;
    case '\a':
        ob_puts(buf, "\\a");
        break;
    case '\b':
        ob_puts(buf, "\\b");
        break;
    case '\t':
        ob_puts(buf, "\\t");
        break;
    case '\n':
        ob_puts(buf, "\\n");
        break;
    case '\v':
        ob_puts(buf, "\\v");
        break;
    case '\f':
        ob_puts(buf, "\\f");
        break;
    case '\r':
        ob_puts(buf, "\\r");
        break;
    case '\\':
        ob_puts(buf, "\\\\");
        break;
    default:
        if( c >= 31 && c <= 127 )
            ob_putc(buf, c);
		else if( c < 256 )
		{
			ob_puts(buf, "\\x");
            ob_rev_putc(buf, c, 2);
		}
		else
		{
			ob_puts(buf, "\\u");
            ob_rev_putc(buf, c, 8);
		}
        break;
    }
}

void CharSet::Print(outbuf *pio) const
{
	UIntSet cs = m_chars;
    UIntSet negcs = ~cs;
	vector<NamedSetItem> &namedsets = *NamedSets();
	bool bUseNamedSets = false;
	bool bNegUseNamedSets = false;
	vector<bool> usenamedset;
	vector<bool> negusenamedset;
	usenamedset.resize(namedsets.size());
	negusenamedset.resize(namedsets.size());
	for( size_t i = 0, n = namedsets.size(); i < n; ++i )
	{
		if( (cs & namedsets[i].namedset) != namedsets[i].namedset )
			usenamedset[i] = false;
        else
        {
		    cs -= namedsets[i].namedset;
		    usenamedset[i] = true;
		    bUseNamedSets = true;
        }
		if( (negcs & namedsets[i].namedset) != namedsets[i].namedset )
			negusenamedset[i] = false;
        else
        {
		    negcs -= namedsets[i].namedset;
		    negusenamedset[i] = true;
		    bNegUseNamedSets = true;
        }
	}
    if( cs.paircount() <= negcs.paircount() )
		ob_puts(pio, "[");
    else
    {
        // Use the negative charset.
		ob_puts(pio, "[^");
        cs = negcs;
        bUseNamedSets = bNegUseNamedSets;
        usenamedset = negusenamedset;
    }
	if( bUseNamedSets )
	{
		for( size_t i = 0, n = namedsets.size(); i < n; ++i )
		{
			if( usenamedset[i] )
			{
				ob_puts(pio, "[:");
				ob_puts(pio, namedsets[i].name.c_str());
				ob_puts(pio, ":]");
			}
		}
	}
    for( UIntSet::iterator cur = cs.begin(); cur != cs.end(); ++cur )
    {
        if( cur->m_min && strchr("[]-", cur->m_min) )
        {
            ob_putc(pio, '\\');
            ob_putc(pio, cur->m_min);
        }
        else
            ob_printchar(pio, cur->m_min);
        if( cur->m_min == cur->m_max )
            continue;
		if( cur->m_max - cur->m_min > 1 )
	        ob_putc(pio, '-');
        if( cur->m_max && strchr("[]-", cur->m_max) )
        {
            ob_putc(pio, '\\');
            ob_putc(pio, cur->m_max);
        }
        else
            ob_printchar(pio, cur->m_min);
    }
	ob_puts(pio, "]");
	if( m_TokenState )
		m_TokenState->Print(pio);
}

Nonterminal::Nonterminal(Grammar *g, String name, OrderedTokenSet *tokstate)
: Symbol(g, name, tokstate)
{
	m_ntfirsts += this;
}

Nonterminal::~Nonterminal()
{
}

void Nonterminal::MarkReachableNonterminalsAndCheckForProductionlessNonterminals(hashset<Nonterminal*> &nonterminals, outbuf *err)
{
    if( nonterminals.find(this) != nonterminals.end() )
        return;
    nonterminals.insert(this);
    if( m_OwningProductions.size() == 0 )
    {
        m_g->m_errcount++;
        if( err )
            ob_printf(err, "error: Nonterminal %s has no productions\n", m_Name.c_str());
        return;
    }
    else
    {
        for( vector<Production*>::iterator cur = m_OwningProductions.begin(); cur != m_OwningProductions.end(); ++cur )
            (*cur)->MarkReachableNonterminalsAndCheckForProductionlessNonterminals(nonterminals, err);
    }
}

StartSymbol::StartSymbol(Grammar *g)
: Nonterminal(g, "<Start>", 0)
{
}

StartSymbol::~StartSymbol()
{
}

Terminal::Terminal(Grammar *g, String name)
: Symbol(g, name, 0)
{
}

Terminal::~Terminal()
{
}

void Terminal::ResetFirsts()
{
	Symbol::ResetFirsts();
    m_firsts += this;
}

void verboseinfo::clear()
{
    m_rules.clear();
    m_conflicts.clear();
    m_stateitems.clear();
    m_deps.clear();
    m_follows.clear();
}

int verboseinfo::AddItem(int stateno, int itemidx, const char *displaystring, const char *allfollows)
{
    m_stateitems.resize(m_stateitems.size()+1);
    stateitem &back = m_stateitems.back();
    back.rowid = m_stateitems.size()-1;
    back.state = stateno;
    back.item = itemidx;
    back.displaystring = displaystring ? displaystring : "";
    back.allfollows = allfollows ? allfollows : "";
    return (int)m_stateitems.size()-1;
}

void verboseinfo::UpdateItem(int rowid, const char *allfollows)
{
    m_stateitems[rowid].allfollows = allfollows;
}

void verboseinfo::AddDep(int dstid, int srcid)
{
    m_deps.resize(m_deps.size()+1);
    dep &back = m_deps.back();
    back.dstid = dstid;
    back.srcid = srcid;
}

void verboseinfo::AddFollow(int dstid, int srcid, int isdep, const char *symbol)
{
    m_follows.resize(m_follows.size()+1);
    follow &back = m_follows.back();
    back.dstid = dstid;
    back.srcid = srcid;
    back.isdep = isdep;
    back.symbol = symbol ? symbol : "";
}
