// Copyright (c) 2008-2011 Mighty Heave Software LLC.
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
#ifndef __hashset_h
#define __hashset_h

#include "ints.h"
#include "str.h"
#include "templatehelp.h"
#include <stdlib.h>

#define HIGH_BITS(bitcount, value) ((bitcount) == 0) ? 0 : ((value) >> (32-(bitcount)));

template<>
class simplehash <String>
{
public:
    uint32_t operator()(const String &str) const
    {
        return hash_djb2(str.c_str());
    }
};

// On the benchmark machine, hashed buckets with 1/2 the number of hash slots as bucket entries seem
// to take slightly under 2/3 the time of the sorted buckets.  Furthermore, their performance stays
// about the same regardless of the bucket size.

// Hashed bucket performance (bucketsize = 128, ratio = 1/2): mass add/find 5,10,15,20,25 million entries.
// 7.74 sec, 16.32 sec, 24.52 sec, 33.58 sec, 90.98 sec (lots of thrashing on 25 mil)
// Hashed bucket performance (bucketsize = 65534, ratio = 1/2): mass add/find 5,10,15,20,25 million entries.
// 7.48 sec, 16.10 sec, 24.58 sec, 33.68 sec, 121.59 sec (varies, lots of thrashing on 25 mil)
//
// As expected, increasing the ratio boosts performance somewhat, at the cost of more memory.
// Also, on the benchmark machine, the thrashing becomes more pronounced at the 25 mil mark.
// Hashed bucket performance (bucketsize = 65534, ratio = 1/1): mass add/find 5,10,15,20,25 million entries.
// 6.90 sec, 14.52 sec, 21.30 sec, 30.87 sec, 162.93 sec (as usual, thrashing on 25 mil)
template<class T, unsigned int bucketsize, class eqT, int ratio_top = 1, int ratio_bottom = 1>
struct hs_nodebucket
{
public:
    int	             m_count;
    uint32_t	     m_bucket_bits;
    hs_nodebucket*   m_prev;
    hs_nodebucket*   m_next;
    uint16_t         m_freelist;

    struct bucketentry
    {
        bucketentry(uint32_t hashvalue, const T &value)
            : m_hashvalue(hashvalue), m_value(value), m_prev(0), m_next(0)
        {
        }

        bucketentry(const bucketentry &rhs)
            : m_hashvalue(rhs.m_hashvalue), m_value(rhs.m_value), m_prev(rhs.m_prev), m_next(rhs.m_next)
        {
        }

        uint32_t    m_hashvalue;
        T           m_value;
        uint16_t    m_prev;
        uint16_t    m_next;
    };
    uint16_t        m_hash[bucketsize*ratio_top/ratio_bottom];
    bucketentry		m_entries[bucketsize];

    class entrypos
    {
    public:
        int         m_idx;
        uint16_t    m_entryidx;

        entrypos() : m_idx(-1), m_entryidx(0xFFFF) {}
        entrypos(int idx) : m_idx(idx), m_entryidx(0xFFFF) {}
        entrypos(int idx, uint16_t entryidx) : m_idx(idx), m_entryidx(entryidx) {}

        bool operator==(const entrypos &rhs) const
        {
            return m_idx == rhs.m_idx && m_entryidx == rhs.m_entryidx;
        }
        bool operator!=(const entrypos &rhs) const
        {
            return ! operator==(rhs);
        }
        bool operator==(int idx)
        {
            if( idx == -1 && m_idx == -1 && m_entryidx == 0xFFFF )
                return true;
            return false;
        }
    };
    typedef entrypos insertpos;

public:
    void init()
    {
        m_freelist = 0;
        m_count = 0;
        m_bucket_bits = 0;
        m_prev = 0;
        m_next = 0;
        for( int i = 0; i < sizeof(m_hash)/sizeof(m_hash[0]); ++i )
            m_hash[i] = 0xFFFF;
        for( int i = 0; i < sizeof(m_entries)/sizeof(m_entries[0]); ++i )
        {
            if( i == sizeof(m_entries)/sizeof(m_entries[0])-1 )
                m_entries[i].m_next = 0xFFFF;
            else
                m_entries[i].m_next = i+1;
        }
    }

    void destroy()
    {
        // Clean out any buckets that still have data.
        int count = m_count;
        int freelist = m_freelist;
        if( ! count )
            return;
        for( int i = 0; i < sizeof(m_hash)/sizeof(m_hash[0]); ++i )
        {
            uint16_t pos = m_hash[i];
            if( pos == 0xFFFF )
                continue;
            while( pos != 0xFFFF )
            {
                uint16_t p = pos;
                bucketentry	*entry = m_entries+pos;
                pos = entry->m_next;
                entry->~bucketentry();
                entry->m_next = freelist;
                freelist = p;
                --count;
            }
            m_hash[i] = 0xFFFF;
            if( count == 0 )
                break;
        }
        m_freelist = freelist;
        m_count = 0;
    }

    void copy(const hs_nodebucket &rhs)
    {
        m_count = rhs.m_count;
        m_freelist = rhs.m_freelist;
        m_bucket_bits = rhs.m_bucket_bits;
        for( int i = 0; i < sizeof(m_entries)/sizeof(m_entries[0]); ++i )
            m_entries[i].m_next = rhs.m_entries[i].m_next;
        for( int i = 0; i < sizeof(m_hash)/sizeof(m_hash[0]); ++i )
        {
            uint16_t pos = m_hash[i] = rhs.m_hash[i];
            if( pos == 0xFFFF )
                continue;
            while( pos != 0xFFFF )
            {
                bucketentry	*entry = m_entries+pos;
                const bucketentry *rhsentry = rhs.m_entries+pos;
                new(entry) bucketentry(*rhsentry);
                pos = rhsentry->m_next;
            }
        }
    }

    bool find(uint32_t h, const T &t, const eqT &eq, insertpos *idx)
    {
        int index = h % sizeof(m_hash)/sizeof(m_hash[0]);
        idx->m_idx = index;
        idx->m_entryidx = 0xFFFF;
        uint16_t pos = m_hash[index];
        if( pos == 0xFFFF )
            return false;
        bucketentry	*entry = 0;
        while( pos != 0xFFFF )
        {
            entry = m_entries+pos;
            if( eq(t, entry->m_value) )
            {
                idx->m_entryidx = pos;
                return true;
            }
            uint16_t nextpos = entry->m_next;
            pos = nextpos;
        }
        return false;
    }

    bool isfull()
    {
        if( m_count == bucketsize )
            return true;
        return false;
    }

    insertpos insert(insertpos idx, uint32_t h, const T &t)
    {
        if( idx.m_idx == -1 )
        {
            int index = h % sizeof(m_hash)/sizeof(m_hash[0]);
            idx.m_idx = index;
        }
        uint16_t entryidx = m_freelist;
        bucketentry *entry = m_entries+entryidx;
        m_freelist = entry->m_next;
        memset(entry, 0, sizeof(bucketentry));
        new(entry) bucketentry(h, t);
        uint16_t head = m_hash[idx.m_idx];
        entry->m_prev = 0xFFFF;
        entry->m_next = head;
        m_hash[idx.m_idx] = entryidx;
        if( head != 0xFFFF )
            m_entries[head].m_prev = entryidx;
        ++m_count;
        return insertpos(idx.m_idx, entryidx);
    }

    void split(hs_nodebucket *next)
    {
#ifdef HASHCORE_EXTRA_CHECKING
        if( ! isfull() || next->m_count != 0 || m_bucket_bits != next->m_bucket_bits )
            assert();
#endif
        next->m_next = m_next;
        m_next = next;
        next->m_prev = this;
        // All entries that belong in the next bucket have a 1 in the low bit amongst the high bits.
        for( int i = 0; i < sizeof(m_hash)/sizeof(m_hash[0]); ++i )
        {
            uint16_t head = m_hash[i];
            uint16_t pos = head;
            if( pos == 0xFFFF )
                continue;
            while( pos != 0xFFFF )
            {
                uint16_t curpos = pos;
                bucketentry	*entry = m_entries+pos;
                pos = entry->m_next;
                uint32_t highbits = HIGH_BITS(m_bucket_bits,entry->m_hashvalue);
                if( ! (highbits&1) )
                    continue;
                // Remove from the chain
                if( entry->m_next != 0xFFFF )
                    m_entries[entry->m_next].m_prev = entry->m_prev;
                if( entry->m_prev != 0xFFFF )
                    m_entries[entry->m_prev].m_next = entry->m_next;
                if( curpos == head )
                    head = entry->m_next;
                // Move to next bucket
                next->insert(i, entry->m_hashvalue, entry->m_value);
                // free the slot
                entry->~bucketentry();
                entry->m_next = m_freelist;
                m_freelist = curpos;
            }
            m_hash[i] = head;
        }
        m_count -= next->m_count;
    }

    T *getptr(insertpos idxpos)
    {
#ifdef HASHCORE_EXTRA_CHECKING
        if( idxpos.m_idx < 0 || idxpos.m_idx >= sizeof(m_hash)/sizeof(m_hash[0]) )
            assert();
#endif
        if( idxpos.m_entryidx < 0 || idxpos.m_entryidx >= bucketsize )
        {
#ifdef HASHCORE_EXTRA_CHECKING
            assert();
#endif
            return 0;
        }
        return &m_entries[idxpos.m_entryidx].m_value;
    }

    bool erase(insertpos idxpos)
    {
        if( idxpos.m_idx < 0 || idxpos.m_idx >= sizeof(m_hash)/sizeof(m_hash[0]) )
        {
#ifdef HASHCORE_EXTRA_CHECKING
            assert();
#endif
            return false;
        }
        if( idxpos.m_entryidx < 0 || idxpos.m_entryidx >= bucketsize )
        {
#ifdef HASHCORE_EXTRA_CHECKING
            assert();
#endif
            return false;
        }
        // Remove from the chain
        bucketentry *entry = m_entries+idxpos.m_entryidx;
        if( entry->m_next != 0xFFFF )
            m_entries[entry->m_next].m_prev = entry->m_prev;
        if( entry->m_prev != 0xFFFF )
            m_entries[entry->m_prev].m_next = entry->m_next;
        uint16_t head = m_hash[idxpos.m_idx];
#ifdef HASHCORE_EXTRA_CHECKING
        if( head == 0xFFFF )
            assert();
#endif
        if( idxpos.m_entryidx == head )
            m_hash[idxpos.m_idx] = entry->m_next;
        // free the slot
        entry->~bucketentry();
        entry->m_next = m_freelist;
        m_freelist = idxpos.m_entryidx;
        --m_count;
        return true;
    }

    entrypos first()
    {
        if( m_count == 0 )
            return entrypos();
        for( int i = 0; i < sizeof(m_hash)/sizeof(m_hash[0]); ++i )
        {
            if( m_hash[i] == 0xFFFF )
                continue;
            return entrypos(i, m_hash[i]);
        }
        return entrypos();
    }

    entrypos next(entrypos pos)
    {
        if( m_entries[pos.m_entryidx].m_next != 0xFFFF )
            return entrypos(pos.m_idx, m_entries[pos.m_entryidx].m_next);
        for( int i = pos.m_idx+1; i < sizeof(m_hash)/sizeof(m_hash[0]); ++i )
        {
            if( m_hash[i] == 0xFFFF )
                continue;
            return entrypos(i, m_hash[i]);
        }
        return entrypos();
    }
};

template<class F, class S>
class pair
{
public:
    typedef F keytype;
    typedef S valuetype;
    F first;
    S second;

    pair() : first(), second() {}
    pair(F f, S s) : first(f), second(s) {}
    pair(const pair &rhs) : first(rhs.first), second(rhs.second) {}
    pair &operator=(const pair &rhs)
    {
        first = rhs.first;
        second = rhs.second;
        return *this;
    }
    bool operator==(const pair &rhs) const
    {
        return first == rhs.first && second == rhs.second;
    }
    bool operator!=(const pair &rhs) const
    {
        return first != rhs.first || second != rhs.second;
    }
};

template<class T, int bucketsize, class eqT, class hashT>
class hashcore
{
public:
    typedef hs_nodebucket<T,bucketsize,eqT> nodebucket;
    typedef typename nodebucket::insertpos insertpos;
    hashT m_hashT;
    eqT m_eqT;

    int m_bits_in_index;
    int m_size;
    // The list of buckets.  Even though they're stored as a linked list, we need O(1) time to fetch a bucket.
	nodebucket **m_buckets;

private:
    void init()
    {
        m_buckets = (nodebucket**)malloc(sizeof(nodebucket*));
        nodebucket *bucket = (nodebucket*)malloc(sizeof(nodebucket));
        bucket->init();
        m_buckets[0] = bucket;
        m_bits_in_index = 0;
        m_size = 0;
    }

public:
    hashcore() : m_bits_in_index(0), m_size(0), m_buckets(0)
    {
        init();
    }

    bool empty() const
    {
        return m_size == 0;
    }

	void clear()
	{
		if( ! m_buckets )
            return;
		nodebucket *curbucket = m_buckets[0];
		while( curbucket )
		{
			nodebucket *nextbucket = curbucket->m_next;
            curbucket->destroy();
			curbucket = nextbucket;
		}
        m_size = 0;
	}

	~hashcore()
	{
		nodebucket *curbucket = m_buckets[0];
		while( curbucket )
		{
			nodebucket *nextbucket = curbucket->m_next;
            curbucket->destroy();
            free(curbucket);
			curbucket = nextbucket;
		}
		free(m_buckets);
		m_buckets = 0;
        m_size = 0;
	}

	size_t size() const
    {
        return m_size;
    }

    // Mostly useless, except for testing that split/merge happened properly.
    int bucketcount() const
    {
        return (1 << m_bits_in_index);
    }

    class hashpos
    {
    public:
        nodebucket *m_bucket;
        insertpos m_idx;

        hashpos() : m_bucket(0), m_idx(-1) {}
        hashpos(nodebucket *bucket, insertpos idx) : m_bucket(bucket), m_idx(idx) {}
        bool operator==(const hashpos &rhs) const { return m_bucket == rhs.m_bucket && m_idx == rhs.m_idx; }
        bool operator!=(const hashpos &rhs) const { return m_bucket != rhs.m_bucket || m_idx != rhs.m_idx; }

        hashpos operator++()
        {
            if( ! m_bucket )
                return *this;
            m_idx = m_bucket->next(m_idx);
            if( m_idx == -1 )
            {
                m_bucket = m_bucket->m_next;
                while( m_bucket && m_bucket->m_count == 0 )
                    m_bucket = m_bucket->m_next;
                if( m_bucket )
                    m_idx = m_bucket->first();
            }
            return *this;
        }

        hashpos operator++(int)
        {
            hashpos ret = *this;
            ++(*this);
            return ret;
        }

        hashpos operator--()
        {
            if( ! m_bucket )
                return *this;
            m_idx = m_bucket->prev(m_idx);
            if( m_idx == -1 )
            {
                m_bucket = m_bucket->m_prev;
                while( m_bucket && m_bucket->m_count == 0 )
                    m_bucket = m_bucket->m_prev;
                if( m_bucket )
                    m_idx = m_bucket->last();
            }
            return *this;
        }

        hashpos operator--(int)
        {
            hashpos ret = *this;
            --(*this);
            return ret;
        }
    };

	hashpos find(const T &t) const
	{
		uint32_t hash = m_hashT(t);
		uint32_t h = HIGH_BITS(m_bits_in_index,hash);
		nodebucket *bucket = m_buckets[h];
        insertpos idx = -1;
		if( ! bucket->find(hash, t, m_eqT, &idx) )
            return hashpos();
		return hashpos(bucket, idx);
	}

	pair<hashpos,bool> insert(const T &t)
	{
		uint32_t hash = m_hashT(t);
        uint32_t h = HIGH_BITS(m_bits_in_index,hash);
		nodebucket *bucket = m_buckets[h];
        typename nodebucket::insertpos idx;
        // Look for the entry, if already present, return it.
		if( bucket->find(hash, t, m_eqT, &idx) )
            return pair<hashpos,bool>(hashpos(bucket, idx),false);

		// Split the bucket to make room for another entry.  When we split a bucket, we
    // may need to increase the number of bits in the hash.
		// In rare cases, the data may work out such that all entries are hashed to the same
		// bucket after a split.  When this happens, we need to make room again and we end up
		// making multiple splits.
		while( bucket->isfull() )
		{
			// We are going to split the bucket, we just need to see if there is room for the new bucket.
			int slots = (1 << (m_bits_in_index-bucket->m_bucket_bits));
			if( slots == 1 )
			{
                // The bucket occupies one bucket slot, so we must incrase the size of the bucket list.
				int oldbucketsize = (1 << m_bits_in_index);
				m_bits_in_index++;
				int newbucketsize = (1 << m_bits_in_index);
				m_buckets = (nodebucket**)realloc(m_buckets, sizeof(nodebucket*)*newbucketsize);
				for( int i = oldbucketsize-1; i >= 0; --i )
				{
					nodebucket *tmp = m_buckets[i];
					m_buckets[i*2+1] = m_buckets[i*2] = tmp;
				}
                h *= 2;
				slots *= 2;
			}
			// Assign half the bucket slots in the split bucket to the split off bucket.
			nodebucket *newbucket = (nodebucket*)malloc(sizeof(nodebucket));
            newbucket->init();
			bucket->m_bucket_bits++;
			newbucket->m_bucket_bits = bucket->m_bucket_bits;
			for( int i = h/slots*slots+slots/2, n = (h/slots+1)*slots; i < n; ++i )            
				m_buckets[i] = newbucket;
			// Distribute the nodes between the new and old buckets.
            bucket->split(newbucket);
            // renew the hash
            h = HIGH_BITS(m_bits_in_index,hash);
            bucket = m_buckets[h];
            idx = -1;
		}

        // We found a bucket with enough room, so add the entry.
        idx = bucket->insert(idx, hash, t);
        ++m_size;
		return pair<hashpos,bool>(hashpos(bucket,idx),true);
 	}

	bool erase(hashpos pos)
	{
        if( ! pos.m_bucket )
            return false;
        if( pos.m_bucket->erase(pos.m_idx) )
        {
            --m_size;
            return true;
        }
        return false;
	}

    bool erase(const T &t)
    {
        return erase(find(t));
    }

    hashpos begin() const
    {
        if( m_size == 0 )
            return hashpos();
        nodebucket *curbucket = m_buckets[0];
        while( curbucket && curbucket->m_count == 0 )
            curbucket = curbucket->m_next;
        if( ! curbucket )
            return hashpos();
        return hashpos(curbucket, curbucket->first());
    }

    hashpos end() const
    {
        return hashpos();
    }

    void insert(hashpos first, hashpos last)
    {
        while( first != last )
        {
            T *t = first.m_bucket->getptr(first.m_idx);
            insert(*t);
            ++first;
        }
    }

    hashcore &operator=(const hashcore &rhs)
    {
        m_buckets = (nodebucket**)realloc(m_buckets, sizeof(nodebucket*)*rhs.bucketcount());
        m_bits_in_index = rhs.m_bits_in_index;
        m_size = rhs.m_size;
        nodebucket **curfinalbuckets = m_buckets;
        nodebucket *prevlhsbucket = 0;
        nodebucket *lhsbucket = m_buckets[0];
        nodebucket *rhsbucket = rhs.m_buckets[0];
        while( rhsbucket )
        {
            if( ! lhsbucket )
            {
                lhsbucket = (nodebucket*)malloc(sizeof(nodebucket));
                lhsbucket->init();
                lhsbucket->m_prev = prevlhsbucket;
                prevlhsbucket->m_next = lhsbucket;
            }
            else
                lhsbucket->destroy();
            lhsbucket->copy(*rhsbucket);
            int slots = (1 << (m_bits_in_index-lhsbucket->m_bucket_bits));
            for( int i = 0; i < slots; ++i )
                *curfinalbuckets++ = lhsbucket;
            prevlhsbucket = lhsbucket;
            lhsbucket = lhsbucket->m_next;
            rhsbucket = rhsbucket->m_next;
        }
        // Remove the extra buckets.  If any.
        if( lhsbucket )
        {
            nodebucket *next = 0;
            prevlhsbucket->m_next = 0;
            while( lhsbucket )
            {
                next = lhsbucket->m_next;
                lhsbucket->destroy();
                free(lhsbucket);
                lhsbucket = next;
            }
        }
        return *this;
    }

    hashcore(const hashcore &rhs) : m_bits_in_index(0), m_size(0), m_buckets(0)
    {
        init();
        *this = rhs;
    }

    bool operator==(const hashcore &rhs) const
    {
        if( m_size != rhs.m_size )
            return false;
        for( hashpos iter = rhs.begin(), iterend = rhs.end(), myend = end(); iter != iterend; ++iter )
        {
            T *t = iter.m_bucket->getptr(iter.m_idx);
            hashpos myiter = find(*t);
            if( myiter == myend )
                return false;
            T *myt = myiter.m_bucket->getptr(myiter.m_idx);
            if( ! (*t == *myt) )
                return false;
        }
        return true;
    }

    bool operator!=(const hashcore &rhs) const
    {
        return ! operator==(rhs);
    }

};

template<class T, int bucketsize = 512, class eqT=simpleeq<T>, class hashT=simplehash<T> >
class hashset
{
private:
    typedef hashcore<T,bucketsize,eqT,hashT> hcore;
    typedef typename hcore::nodebucket nodebucket;
    typedef typename hcore::hashpos hashpos;
    hcore m_core;

public:
    class iterator
    {
    protected:
        hashpos m_pos;
    public:
        iterator() {}
        iterator(hashpos pos) : m_pos(pos) {}
        iterator(const iterator &rhs) { m_pos = rhs.m_pos; }
        iterator &operator=(const iterator &rhs) { m_pos = rhs.m_pos; return *this; }
        bool operator==(const iterator &rhs) const { return m_pos == rhs.m_pos; }
        bool operator!=(const iterator &rhs) const { return ! (*this == rhs); }
        iterator operator++() { return iterator(++m_pos); }
        iterator operator++(int)
        {
            iterator it = *this;
            ++(*this);
            return it;
        }

        iterator operator--() { return iterator(--m_pos); }

        iterator operator--(int)
        {
            iterator it = *this;
            --(*this);
            return it;
        }
        
        const T &operator*() const { return *m_pos.m_bucket->getptr(m_pos.m_idx); }
        const T *operator->() const { return m_pos.m_bucket->getptr(m_pos.m_idx); }
    };

    hashset() {}
    hashset(const hashset &rhs) : m_core(rhs.m_core) {}

    hashset &operator=(const hashset &rhs)
    {
        m_core = rhs.m_core;
        return *this;
    }

    void clear()
    {
        m_core.clear();
    }
    
    size_t size() { return m_core.size(); }

    iterator begin() const
    {
        return iterator(m_core.begin());
    }

    iterator end() const
    {
        return iterator(m_core.end());
    }

    iterator find(const T &t) const
    {
        hashpos pos = m_core.find(t);
        return iterator(pos);
    }

    pair<iterator,bool> insert(const T &t)
    {
        pair<hashpos,bool> pos = m_core.insert(t);
        return pair<iterator,bool>(iterator(pos.first),pos.second);
    }

    template<class InputIterator>
    void insert(InputIterator first, InputIterator last)
    {
        for( ; first != last; ++first )
        {
            const T &t = *first;
            insert(t);
        }
    }

    void erase(iterator iter)
    {
        m_core.erase(iter.m_curbucket, iter.m_idx);
    }

    void erase(const T &t)
    {
        m_core.erase(t);
    }

    bool operator==(const hashset &rhs) const
    {
        return m_core.operator==(rhs.m_core);
    }

    bool operator!=(const hashset &rhs) const
    {
        return ! m_core.operator==(rhs.m_core);
    }
};

#endif /* __hashset_h */
