// Copyright (c) 2008-2011 Mighty Heave Software LLC.
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
#include "intset.h"
#include <limits.h>
#ifdef _CRTDBG_MAP_ALLOC
#include <stdlib.h>
#include <crtdbg.h>
#endif

//#define WANT_UINTSET_SANITY_CHECK 1
#ifdef WANT_UINTSET_SANITY_CHECK
#define SANITY_CHECK SanityCheck();
#else
#define SANITY_CHECK
#endif

IntSetRange::IntSetRange() : m_min(0), m_max(-1) {}

IntSetRange::IntSetRange(unsigned int min, unsigned int max) : m_min(min), m_max(max)
{
}

bool IntSetRange::operator==(const IntSetRange &rhs) const
{
    return m_min == rhs.m_min && m_max == rhs.m_max;
}

bool IntSetRange::operator!=(const IntSetRange &rhs) const
{
    return ! operator==(rhs);
}

bool IntSetRange::operator&(const IntSetRange &rhs) const
{
    // If either has no elements, there is no intersection.
    if( ! count() || ! rhs.count() )
        return false;
    // rhs falls completely in our range.
	if( m_min <= rhs.m_min && m_max >= rhs.m_max )
		return true;
    // our min falls within rhs range.
    if( m_min >= rhs.m_min && m_min <= rhs.m_max )
        return true;
    // our max falls within rhs range.
    if( m_max >= rhs.m_min && m_max <= rhs.m_max )
        return true;
    return false;
}

size_t IntSetRange::count() const
{
    if( m_max < m_min )
        return 0;
    return m_max - m_min + 1;
}

UIntSetRange::UIntSetRange() : m_min(1), m_max(0) {}

UIntSetRange::UIntSetRange(unsigned int min, unsigned int max) : m_min(min), m_max(max)
{
}

bool UIntSetRange::operator==(const UIntSetRange &rhs) const
{
    return m_min == rhs.m_min && m_max == rhs.m_max;
}

bool UIntSetRange::operator!=(const UIntSetRange &rhs) const
{
    return ! operator==(rhs);
}

bool UIntSetRange::operator&(const UIntSetRange &rhs) const
{
    // If either has no elements, there is no intersection.
    if( ! count() || ! rhs.count() )
        return false;
    // rhs falls completely in our range.
	if( m_min <= rhs.m_min && m_max >= rhs.m_max )
		return true;
    // our min falls within rhs range.
    if( m_min >= rhs.m_min && m_min <= rhs.m_max )
        return true;
    // our max falls within rhs range.
    if( m_max >= rhs.m_min && m_max <= rhs.m_max )
        return true;
    return false;
}

UIntSetRange UIntSetRange::intersect(const UIntSetRange &rhs) const
{
    // If either has no elements, there is no intersection.
    if( ! count() || ! rhs.count() )
        return UIntSetRange();
    // rhs falls completely in our range.
	if( m_min <= rhs.m_min && m_max >= rhs.m_max )
        return UIntSetRange(rhs.m_min, rhs.m_max);
    // our min falls within rhs range.
    if( m_min >= rhs.m_min && m_min <= rhs.m_max )
        return UIntSetRange(m_min,((m_max<rhs.m_max)?m_max:rhs.m_max));
    // our max falls within rhs range.
    if( m_max >= rhs.m_min && m_max <= rhs.m_max )
        return UIntSetRange(((m_min>rhs.m_min)?m_min:rhs.m_min), m_max);
    return UIntSetRange();
}

unsigned int UIntSetRange::count() const
{
    if( m_max < m_min )
        return 0;
    return m_max - m_min + 1;
}

bool UIntSetRange::leftof(const UIntSetRange &rhs) const
{
	if( rhs.m_min > 0 && m_max < rhs.m_min-1 )
		return true;
	return false;
}

bool UIntSetRange::operator<(const UIntSetRange &rhs) const
{
    if( m_min < rhs.m_min )
        return true;
    else if( rhs.m_min < m_min )
        return false;
    if( m_max < rhs.m_max )
        return true;
    return false;
}

bool UIntSetRange::is_subset_of(const UIntSetRange &rhs) const
{
	if( m_min >= rhs.m_min && m_max <= rhs.m_max )
		return true;
	return false;
}

bool UIntSetRange::contiguous_with(const UIntSetRange &rhs) const
{
	if( rhs.m_min > m_max+1 || rhs.m_max+1 < m_min )
		return false;
	return true;
}

UIntSetRange UIntSetRange::contiguous_range(const UIntSetRange &rhs) const
{
	unsigned int low = m_min;
	if( rhs.m_min < low )
		low = rhs.m_min;
	unsigned int high = m_max;
	if( rhs.m_max > high )
		high = rhs.m_max;
	return UIntSetRange(low, high);
}

UIntSet::IntSetPacket::IntSetPacket() : m_ref(1)
{
}

UIntSet::IntSetPacket::IntSetPacket(const IntSetPacket &rhs) : m_ref(1)
{
    m_ranges = rhs.m_ranges;
}

UIntSet::IntSetPacket& UIntSet::IntSetPacket::operator=(const IntSetPacket &rhs)
{
    m_ranges = rhs.m_ranges;
    return *this;
}

int UIntSet::IntSetPacket::IncRef()
{
    return ++m_ref;
}

int UIntSet::IntSetPacket::DecRef()
{
    if( --m_ref == 0 )
    {
        delete this;
        return 0;
    }
    return m_ref;
}

int UIntSet::IntSetPacket::Refs()
{
    return m_ref;
}

UIntSet::UIntSet() : m_packet(0)
{
}

UIntSet::UIntSet(const UIntSet &rhs)
{
    m_packet = rhs.m_packet;
    if( m_packet )
        m_packet->IncRef();
	SANITY_CHECK
}

UIntSet::~UIntSet()
{
    erase();
}

void UIntSet::PrepareForInplaceUpdate()
{
    if( ! m_packet )
    {
        m_packet = new IntSetPacket();
    }
    else if( m_packet->Refs() > 1 )
    {
		SANITY_CHECK
        IntSetPacket *oldpacket = m_packet;
        m_packet = new IntSetPacket(*oldpacket);
        oldpacket->DecRef();
		SANITY_CHECK
    }
}

UIntSet &UIntSet::operator=(const UIntSet &rhs)
{
    if( rhs.m_packet )
        rhs.m_packet->IncRef();
    if( m_packet )
        m_packet->DecRef();
    m_packet = rhs.m_packet;
	SANITY_CHECK
    return *this;
}

bool UIntSet::operator==(const UIntSet &rhs) const
{
    if( m_packet == rhs.m_packet )
        return true;
    size_t rangecount = (m_packet ? m_packet->m_ranges.size() : 0);
    if( rangecount != (rhs.m_packet ? rhs.m_packet->m_ranges.size() : 0) )
        return false;
    for( unsigned int i = 0; i < rangecount; ++i )
    {
        if( m_packet->m_ranges[i] != rhs.m_packet->m_ranges[i] )
            return false;
    }
    return true;
}

bool UIntSet::operator<(const UIntSet &rhs) const
{
    if( m_packet == rhs.m_packet )
        return false;
    if( ! m_packet )
        return false;
    if( ! rhs.m_packet )
        return true;
    return m_packet->m_ranges < rhs.m_packet->m_ranges;
}

bool UIntSet::operator!=(const UIntSet &rhs) const
{
    return ! operator==(rhs);
}

void UIntSet::operator+=(unsigned int value)
{
    (*this) += UIntSetRange(value, value);
	SANITY_CHECK
}

void UIntSet::operator+=(const UIntSet &rhs)
{
    if( ! rhs.m_packet )
        return;
	if( ! m_packet )
	{
		m_packet = new IntSetPacket(*rhs.m_packet);
		SANITY_CHECK
		return;
	}
    vector<UIntSetRange>::const_iterator currhsrange = rhs.m_packet->m_ranges.begin(), endrhsrange = rhs.m_packet->m_ranges.end();
	int currange = 0;
	while( currhsrange != endrhsrange && currange != m_packet->m_ranges.size() )
	{
		UIntSetRange lhsrangevalue = m_packet->m_ranges[currange];
		if( lhsrangevalue.leftof(*currhsrange) )
		{
			// haven't reached a place to compare with rhs yet.
			++currange;
			continue;
		}
		if( currhsrange->leftof(lhsrangevalue) )
		{
			// rhs range is before current range, so add it.
			PrepareForInplaceUpdate();
			SANITY_CHECK
			m_packet->m_ranges.insertat(currange, *currhsrange);
			SANITY_CHECK
			++currhsrange; // because we inserted it.
			++currange; // because this now points at the inserted value, adjust.
			continue;
		}
		if( currhsrange->is_subset_of(lhsrangevalue) )
		{
			// rhs range fits within current range, so skip it.
			++currhsrange;
			continue;
		}
		if( currhsrange->contiguous_with(lhsrangevalue) )
		{
			// rhs range butts up against current range, so extend current range.
			PrepareForInplaceUpdate();
			UIntSetRange &reflhsrangevalue = m_packet->m_ranges[currange];
			// extend to match the rhs
			reflhsrangevalue = currhsrange->contiguous_range(reflhsrangevalue);
			// move past these ranges.
			++currange;
			++currhsrange;
			int beginremove = currange, endremove = currange;
			size_t endrange = m_packet->m_ranges.size();
			// keep comparing with the left and right hand sides, extending the range so
			// long as the range butts up with either left or right hand ranges.
			while(1)
			{
				if( endremove != endrange && reflhsrangevalue.contiguous_with(m_packet->m_ranges[endremove]) )
				{
					reflhsrangevalue = reflhsrangevalue.contiguous_range(m_packet->m_ranges[endremove++]);
					continue;
				}
				if( currhsrange != endrhsrange && reflhsrangevalue.contiguous_with(*currhsrange) )
				{
					reflhsrangevalue = reflhsrangevalue.contiguous_range(*currhsrange++);
					continue;
				}
				break;
			}
			// remove any ranges that we extended over.
			if( endremove > beginremove )
				m_packet->m_ranges.removeat(beginremove, endremove-beginremove);
			SANITY_CHECK
			continue;
		}
	}
	if( currhsrange != endrhsrange && currange == m_packet->m_ranges.size() )
	{
		// past the end, just dump the rest in.
		PrepareForInplaceUpdate();
		SANITY_CHECK
		m_packet->m_ranges.appendmany(currhsrange, endrhsrange-currhsrange);
		SANITY_CHECK
	}
	SANITY_CHECK
}

void UIntSet::operator+=(const UIntSetRange &range)
{
    // Really Easy case.
    if( ! range.count() )
        return;
    // Easy case.
    if( ! m_packet || m_packet->m_ranges.size() == 0 )
    {
        PrepareForInplaceUpdate();
        m_packet->m_ranges.append(range);
		SANITY_CHECK
        return;
    }
	// Optimize a common case.  Fits after current data.
	if( range.m_min > m_packet->m_ranges[-1].m_max+1 )
	{
		PrepareForInplaceUpdate();
		m_packet->m_ranges.append(range);
		SANITY_CHECK
		return;
	}

    // See if range fits between ranges or overlaps them.
	int currange = 0;
	size_t endrange = m_packet->m_ranges.size();
	size_t retainrange = endrange;
	size_t beginremove = endrange;
    while( currange != endrange )
    {
		UIntSetRange &r = m_packet->m_ranges[currange];
		if( range.is_subset_of(r) )
			return; // new data fits within a current range, no mod.
		if( range.contiguous_with(r) )
		{
			// found a continuous range that's not a subset, merge many on the right.
			PrepareForInplaceUpdate();
			vector<UIntSetRange>::iterator mergedrange = m_packet->m_ranges.begin()+currange;
			*mergedrange = range.contiguous_range(r);
			size_t firstremoval = ++currange;
			while( currange != endrange && mergedrange->contiguous_with(m_packet->m_ranges[currange]) )
				*mergedrange = mergedrange->contiguous_range(m_packet->m_ranges[currange++]);
			m_packet->m_ranges.removeat(firstremoval, currange-firstremoval);
			SANITY_CHECK
			return;
		}
		if( range.leftof(r) )
		{
			// First between datas.
			PrepareForInplaceUpdate();
			m_packet->m_ranges.insertat(currange, range);
			SANITY_CHECK
			return;
		}
		++currange;
	}
	SANITY_CHECK
}

void UIntSet::operator-=(unsigned int value)
{
	SANITY_CHECK
    *this -= UIntSetRange(value, value);
	SANITY_CHECK
}

void UIntSet::operator-=(const UIntSet &rhs)
{
	// Quick out.
    if( ! m_packet || m_packet->m_ranges.size() == 0 || ! rhs.m_packet || rhs.m_packet->m_ranges.size() == 0 )
        return;
    for( vector<UIntSetRange>::const_iterator prhs = rhs.m_packet->m_ranges.begin(); prhs != rhs.m_packet->m_ranges.end(); ++prhs )
        *this -= UIntSetRange(prhs->m_min, prhs->m_max);
	SANITY_CHECK
}

void UIntSet::operator-=(const UIntSetRange &range)
{
	
    if( ! m_packet || m_packet->m_ranges.size() == 0 || range.count() == 0 || range.m_max < m_packet->m_ranges[0].m_min || range.m_min > m_packet->m_ranges[-1].m_max )
		return; // Quick out
	int currange = 0;
	int endrange = (int)m_packet->m_ranges.size();
	int lowrange = endrange;
	int highrange = endrange;
	for( ; currange != endrange; ++currange )
	{
		UIntSetRange &sr = m_packet->m_ranges[currange];
		if(lowrange == endrange)
		{
			if(range & sr)
				highrange = lowrange = currange;
		}
		else if(range & sr)
			highrange = currange;
		else
			break;
	}
	if( lowrange == endrange )
		return;
	PrepareForInplaceUpdate();
	size_t ends = 0;
	bool bLowChops = (range.m_min > m_packet->m_ranges[lowrange].m_min);
	bool bHighChops = (range.m_max < m_packet->m_ranges[highrange].m_max);
	unsigned int oldmax = m_packet->m_ranges[highrange].m_max;

	if( bLowChops )
		m_packet->m_ranges[lowrange++].m_max = range.m_min-1;
	if( bLowChops && bHighChops && lowrange >= highrange )
	{
		// ugliest case.  We split existing data with the range.
		m_packet->m_ranges.insertat(lowrange, UIntSetRange(range.m_max+1, oldmax));
		SANITY_CHECK
		return;
	}
	if( bHighChops )
		m_packet->m_ranges[highrange--].m_min = range.m_max+1;
	if( highrange >= lowrange )
		m_packet->m_ranges.removeat(lowrange, highrange-lowrange+1);
	SANITY_CHECK
}

UIntSet &UIntSet::operator%(unsigned int value)
{
    *this += value;
    return *this;
}

UIntSet &UIntSet::operator%(const UIntSet &rhs)
{
    *this += rhs;
    return *this;
}

UIntSet &UIntSet::operator%(const UIntSetRange &rhs)
{
    *this += rhs;
    return *this;
}

UIntSet UIntSet::operator~() const
{
    UIntSet ret;
    if( ! m_packet || m_packet->m_ranges.size() == 0 )
    {
		SANITY_CHECK
        ret += UIntSetRange(0, INT_MAX);
        return ret;
    }
	// Take (first, last) now, because we're about to alter the data.
	unsigned int first = m_packet->m_ranges[0].m_min;
	unsigned int last = m_packet->m_ranges[-1].m_max;
	if( first > 0 )
		ret += UIntSetRange(0, first-1);
	// Ranges are never butted up together, because we always extend in +=.
    for( int i = 0, n = (int)m_packet->m_ranges.size()-1; i < n; ++i )
        ret += UIntSetRange(m_packet->m_ranges[i].m_max+1, m_packet->m_ranges[i+1].m_min-1);
	if( last < INT_MAX )
		ret += UIntSetRange(last+1, INT_MAX);
#ifdef WANT_UINTSET_SANITY_CHECK
	ret.SanityCheck();
#endif
	return ret;
}

bool UIntSet::operator&(unsigned int value) const
{
    if( ! m_packet )
        return false;
    for( vector<UIntSetRange>::const_iterator cur = m_packet->m_ranges.begin(); cur != m_packet->m_ranges.end(); ++cur )
        if( value >= cur->m_min && value <= cur->m_max )
            return true;
    return false;
}

size_t UIntSet::count() const
{
    if( ! m_packet )
        return 0;
	SANITY_CHECK
    unsigned int ret = 0;
    for( vector<UIntSetRange>::const_iterator cur = m_packet->m_ranges.begin(); cur != m_packet->m_ranges.end(); ++cur )
        ret += (cur->m_max - cur->m_min) + 1;
    return ret;
}

size_t UIntSet::paircount() const
{
    if( ! m_packet )
        return 0;
    return m_packet->m_ranges.size();
}

UIntSet UIntSet::operator&(const UIntSet &rhs) const
{
    UIntSet negrhs = ~rhs;
    UIntSet ret = *this;
    ret -= negrhs;
    return ret;
}

void UIntSet::operator&=(const UIntSet &rhs)
{
    UIntSet negrhs = ~rhs;
    *this -= negrhs;
}

UIntSet::iterator UIntSet::begin()
{
    if( ! m_packet )
        return UIntSetIterator(this, 0, true);
    return UIntSetIterator(this, m_packet->m_ranges.begin(), false);
}

UIntSet::iterator UIntSet::end()
{
    if( ! m_packet )
        return UIntSetIterator(this, 0, true);
    return UIntSetIterator(this, m_packet->m_ranges.end(), false);
}

UIntSet::const_iterator UIntSet::begin() const
{
    if( ! m_packet )
        return UIntSetConstIterator(this, 0, true);
    return UIntSetConstIterator(this, m_packet->m_ranges.begin(), false);
}

UIntSet::const_iterator UIntSet::end() const
{
    if( ! m_packet )
        return UIntSetConstIterator(this, 0, true);
    return UIntSetConstIterator(this, m_packet->m_ranges.end(), false);
}

UIntSet::iterator UIntSet::findintersecting(unsigned int low, unsigned int high)
{
    for( UIntSet::iterator r = begin(); r != end(); ++r )
    {
        if( low < r->m_min && high > r->m_max )
            return r;
        if( (low >= r->m_min && low <= r->m_max) || (high >= r->m_min && high <= r->m_max) )
            return r;
    }
    return end();
}

void UIntSet::clear()
{
    if( ! m_packet )
        return;
    if( m_packet->Refs() == 1 )
        m_packet->m_ranges.clear();
    else
    {
        m_packet->DecRef();
        m_packet = 0;
    }
}

void UIntSet::erase()
{
    if( m_packet )
    {
        m_packet->DecRef();
        m_packet = 0;
    }
}

bool UIntSet::TestValues(unsigned int *values, size_t valuecount) const
{
	if( ! SanityCheck() )
		return false;
	size_t cnt = m_packet ? m_packet->m_ranges.size() : 0;
	if( valuecount != cnt )
		return false;
	for( size_t i = 0; i < cnt; ++i )
	{
		const UIntSetRange &r = m_packet->m_ranges[i];
		if( r.m_min != values[0] || r.m_max != values[1] )
			return false;
		values += 2;
	}
	return true;
}

bool UIntSet::SanityCheck() const
{
	unsigned int prevmax = 0;
	size_t cnt = m_packet ? m_packet->m_ranges.size() : 0;
	for( size_t i = 0; i < cnt; ++i )
	{
		const UIntSetRange &r = m_packet->m_ranges[i];
		if( r.m_max < r.m_min )
			return false;
		if( i > 0 )
		{
			if( r.m_min <= prevmax )
				return false;
		}
		prevmax = r.m_max;
	}
	return true;
}
