// Copyright (c) 2008-2011 Mighty Heave Software LLC.
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
#include "nfa.h"
#include "rx.h"
#include <ctype.h>

bool NfaTransition::operator==(const NfaTransition &rhs) const
{
    return m_from == rhs.m_from && m_to == rhs.m_to;
}

bool NfaTransition::operator<(const NfaTransition &rhs) const
{
    if( m_from != rhs.m_from )
    {
        if( m_from < rhs.m_from )
            return true;
        return false;
    }
    return m_to < rhs.m_to;
}

int NfaTransition::NfaTransition_cmp(const NfaTransition *lhs, const NfaTransition *rhs)
{
    if( lhs->m_from != rhs->m_from )
		return lhs->m_from - rhs->m_from;
    return lhs->m_to - rhs->m_to;
}

bool NfaETransition::operator==(const NfaETransition &rhs) const
{
    return m_from == rhs.m_from && m_to == rhs.m_to;
}

bool NfaETransition::operator<(const NfaETransition &rhs) const
{
    if( m_from != rhs.m_from )
    {
        if( m_from < rhs.m_from )
            return true;
        return false;
    }
    return m_to < rhs.m_to;
}

int NfaETransition::NfaETransition_cmp(const NfaETransition *lhs, const NfaETransition *rhs)
{
    if( lhs->m_from != rhs->m_from )
		return lhs->m_from - rhs->m_from;
    return lhs->m_to - rhs->m_to;
}

Nfa::Nfa()
: m_nextstate(0), m_TransitionsNeedSort(false), m_eTransitionsNeedSort(false)
{
}

void Nfa::SortTransitions()
{
    if( m_TransitionsNeedSort )
    {
        m_transitions.sort(NfaTransition::NfaTransition_cmp);
        m_TransitionsNeedSort = false;
    }
    if( m_eTransitionsNeedSort )
    {
        m_etransitions.sort(NfaETransition::NfaETransition_cmp);
        m_eTransitionsNeedSort = false;
    }
}

void Nfa::addstartstate(int state)
{
    markstate(state, STATE_MARKER_STARTSTATE);
}

void Nfa::addendstate(int state)
{
    markstate(state, STATE_MARKER_ENDSTATE);
}

int Nfa::statecount()
{
    return m_nextstate;
}

void Nfa::AddLiteral(const char *str, bool case_insensitive)
{
    int i = 0;
	int startstate = addstate();
	int endstate = startstate;
    while( *str )
    {
        NfaTransition transition;
        transition.m_from = endstate;
		endstate = addstate();
        transition.m_to = endstate;
        int s = *str++;
        if( case_insensitive )
        {
            int upper = toupper(s);
            int lower = tolower(s);
            transition.m_on += upper;
            if( upper != lower )
                transition.m_on += lower;
        }
        else
            transition.m_on += s;
        m_transitions.append(transition);
    }
	addstartstate(startstate);
	addendstate(endstate);
}

int Nfa::AddRxRec(Rx *rx, int startstate)
{
    int finalstate = rx->AddToNfa(*this, startstate);
    return finalstate;
}

void Nfa::AddRx(Rx *rx)
{
    int initstate = addstate();
    addstartstate(initstate);
    int finalstate = AddRxRec(rx, initstate);
    addendstate(finalstate);
}

void Nfa::AddNfa(Nfa &src)
{
    int stateadjust = statecount();
    for( size_t i = 0, n = src.statecount(); i < n; ++i )
        addstate();
    for( size_t i = 0, n = src.m_startstates.size(); i < n; ++i )
        addstartstate(src.m_startstates[i]+stateadjust);
    for( size_t i = 0, n = src.m_endstates.size(); i < n; ++i )
        addendstate(src.m_endstates[i]+stateadjust);
    size_t oldtrnssize = m_transitions.size();
    m_transitions.insert(m_transitions.end(), src.m_transitions.begin(), src.m_transitions.end());
    for( vector<NfaTransition>::iterator trns = m_transitions.begin()+oldtrnssize; trns != m_transitions.end(); ++trns )
    {
        NfaTransition &t = *trns;
        t.m_from += stateadjust;
        t.m_to += stateadjust;
    }
    size_t oldetrnssize = m_etransitions.size();
    m_etransitions.insert(m_etransitions.end(), src.m_etransitions.begin(), src.m_etransitions.end());
    for( vector<NfaETransition>::iterator etrns = m_etransitions.begin()+oldetrnssize; etrns != m_etransitions.end(); ++etrns )
    {
        NfaETransition &et = *etrns;
        et.m_from += stateadjust;
        et.m_to += stateadjust;
    }
    m_TransitionsNeedSort = true;
    m_eTransitionsNeedSort = true;
}

void Nfa::AsDfa(Nfa &dfa, vector< vector<int> > &substates)
{
    substates.clear();
    SortTransitions();

	vector<int> nextset;
    vector<EdgeSet> edges;
    nextset = m_startstates;
    int startstate = dfa.addstate();
    closure(nextset);
    substates.resize(startstate+1);
    substates[startstate] = nextset;
    
	// For each state, we need to figure out which states they lead to.
    for( int cursetindex = startstate; cursetindex != dfa.statecount(); ++cursetindex )
    {
        // consolidate markers, including starting states, ending states, etc.
        int statemarker = 0;
        for( size_t i = 0, n = substates[cursetindex].size(); i < n; ++i )
            statemarker |= m_markers[substates[cursetindex][i]];
        dfa.markstate(cursetindex, statemarker);
        // discover edges, aka generate new states
        discover_edges(substates[cursetindex], edges);
		if( edges.size() == 0 )
			continue;
		for( vector<EdgeSet>::iterator curedge = edges.begin(), endedge = edges.end(); curedge != endedge; ++curedge )
        {
			UIntSet &isnextset = curedge->m_nextset;
			nextset.clear();
			for( UIntSet::iterator ns = isnextset.begin(); ns != isnextset.end(); ++ns )
			{
				for( unsigned int i = ns->m_min; i <= ns->m_max; ++i )
					nextset.append(i);
			}
            closure(nextset);
            int nextstateindex = -1;
            vector< vector<int> >::iterator nextstatesubstates = substates.find(nextset);
            if( nextstatesubstates == substates.end() )
            {
                int nextstateindex = dfa.addstate();
                substates.resize(nextstateindex+1);
                substates[nextstateindex] = nextset;
                dfa.addtransition_charset(cursetindex, nextstateindex, curedge->m_inputs);
            }
            else
            {
                int nextstateindex = (int)(nextstatesubstates - substates.begin());
                dfa.addtransition_charset(cursetindex, nextstateindex, curedge->m_inputs);
            }
        }
    }
    dfa.SortTransitions();
}

int Nfa::addstate()
{
    m_markers.push_back(0);
    return m_nextstate++;
}

void Nfa::markstate(int state, int marker)
{
    while( (int)m_markers.size() <= state )
        m_markers.push_back(0);
    marker &= ~m_markers[state];
    if( ! marker )
        return;
    m_markers[state] |= marker;
    if( marker & STATE_MARKER_STARTSTATE )
        m_startstates.push_back(state);
    if( marker & STATE_MARKER_ENDSTATE )
        m_endstates.push_back(state);
}

void Nfa::relocatedst(int olddst, int newdst)
{
    if( olddst == newdst )
        return;
    for( vector<NfaTransition>::iterator cur = m_transitions.begin(); cur != m_transitions.end(); ++cur )
    {
        if( cur->m_to == olddst )
        {
            cur->m_to = newdst;
            m_TransitionsNeedSort = true;
        }
    }
    for( vector<NfaETransition>::iterator cur = m_etransitions.begin(); cur != m_etransitions.end(); ++cur )
    {
        if( cur->m_to == olddst )
        {
            cur->m_to = newdst;
            m_eTransitionsNeedSort = true;
        }
    }
}

void Nfa::addtransition_empty(int from, int to)
{
    m_etransitions.append(NfaETransition(from, to));
    m_eTransitionsNeedSort = true;
}

void Nfa::addtransition_char(int from, int to, int on)
{
    NfaTransition t(from, to);
    vector<NfaTransition>::iterator cur = m_transitions.find(t);
    if( cur == m_transitions.end() )
    {
        cur = m_transitions.append(t);
        m_TransitionsNeedSort = true;
    }
    cur->m_on += on;
}

void Nfa::addtransition_charset(int from, int to, const UIntSet &on)
{
    NfaTransition t(from, to);
    vector<NfaTransition>::iterator cur = m_transitions.find(t);
    if( cur == m_transitions.end() )
    {
        cur = m_transitions.append(t);
        m_TransitionsNeedSort = true;
    }
    cur->m_on += on;
}

int cmp_int(const int *lhs, const int *rhs)
{
	return *lhs - *rhs;
}

// m_etransitions must be sorted by m_from member.
void Nfa::closure(vector<int> &seed)
{
    seed.unique(cmp_int);
    vector<int> prevset;
    while( seed != prevset )
    {
        prevset = seed;
        vector<int>::iterator curstate = prevset.begin(), endstate = prevset.end();
        vector<NfaETransition>::iterator curtransition = m_etransitions.begin(), endtransition = m_etransitions.end();
        while(curstate != endstate && curtransition != endtransition )
        {
            if( *curstate < curtransition->m_from )
            {
                int from = curtransition->m_from;
                ++curstate;
                while( curstate != endstate && *curstate < from )
                    ++curstate;
                continue;
            }
            else if( *curstate > curtransition->m_from )
            {
                int state = *curstate;
                ++curtransition;
                while( curtransition != endtransition && state > curtransition->m_from )
                    ++curtransition;
                continue;
            }
            else // *curstate == curtransition->m_from
            {
                int state = *curstate;
                while( curstate != endstate && *curstate == state )
                    ++curstate;
                while( curtransition != endtransition && state == curtransition->m_from )
                {
                    seed.append(curtransition->m_to);
                    ++curtransition;
                }
                continue;
            }
        }
        seed.unique(cmp_int);
    }
}

// Produce the intersections of all transitions from a set of states.
//
// m_transitions must be sorted by m_from member
// edges must be clear
// from must be sorted in ascending order
void Nfa::discover_edges(const vector<int> &from, vector<EdgeSet> &edges)
{
    SortTransitions();
	edges.clear();
    vector<int>::const_iterator curstate = from.begin(), endstate = from.end();
    vector<NfaTransition>::iterator curtransition = m_transitions.begin(), endtransition = m_transitions.end();
    while( curstate != endstate && curtransition != endtransition )
    {
        if( *curstate < curtransition->m_from )
        {
            int from = curtransition->m_from;
            ++curstate;
            while( curstate != endstate && *curstate < from )
                ++curstate;
            continue;
        }
        else if( *curstate > curtransition->m_from )
        {
            int state = *curstate;
            ++curtransition;
            while( curtransition != endtransition && state > curtransition->m_from )
                ++curtransition;
            continue;
        }
        else // *curstate == curtransition->m_from
        {
            int state = *curstate;
            while( curstate != endstate && *curstate == state )
                ++curstate;
            while( curtransition != endtransition && state == curtransition->m_from )
            {
				// This chunk handles multiple transitions to the same state.
				// because of this don't use curtransition after this chunk within this while loop.
				// use state, nextstate, _trnson instead of curtransition->m_from, curtransition->m_to, curtransition->m_on.
				int nextstate = curtransition->m_to;
				UIntSet _trnson = curtransition->m_on;
				++curtransition;
				while( curtransition != endtransition && state == curtransition->m_from && curtransition->m_to == nextstate )
				{
					_trnson += curtransition->m_on;
					++curtransition;
				}
				// } multiple transitions
				UIntSet leftovers = _trnson;
				for( vector<EdgeSet>::iterator curedge = edges.begin(); curedge != edges.end(); ++curedge )
				{
					UIntSet intersection;
					intersection = curedge->m_inputs & _trnson;
					leftovers -= curedge->m_inputs;
					if( intersection.count() == curedge->m_inputs.count() )
					{
						// all inputs follow the new transition also.
						curedge->m_nextset += nextstate;
					}
					else if( intersection.count() > 0 )
					{
						// some inputs follow only the transition to the old state.
						curedge->m_inputs -= _trnson;
						// some inputs also follow the transition to the new state.
						EdgeSet es;
						es.m_inputs = intersection;
						es.m_nextset = curedge->m_nextset;
						es.m_nextset += nextstate;
						curedge = edges.insert(curedge+1, es);
					}
					// else { nothing follows the transition to the new state, nothing to do }
				}
				// Anything left follows only the current transition.
				if( leftovers.count() )
				{
					EdgeSet es;
					es.m_inputs = leftovers;
					es.m_nextset += nextstate;
					edges.append(es);
				}
			}
            continue;
        }
    }
}

// m_transitions must be in m_from order.
void Nfa::transitionfnct(int input, int from, vector<int> &to)
{
    SortTransitions();
    vector<NfaTransition>::iterator cur = m_transitions.begin(), end = m_transitions.end();
    while( cur != end && cur->m_from < from )
        ++cur;
    while( cur != end && cur->m_from == from )
    {
        if( cur->m_on & input )
            to.append(cur->m_to);
        ++cur;
    }
}

class SimpleNfaDiagramHelper : public NfaDiagramHelper
{
	static SimpleNfaDiagramHelper m_Instance;
public:
	static NfaDiagramHelper *Instance()
	{
		return &m_Instance;
	}

	virtual String NodeLabel(Nfa &nfa, int node)
	{
        char buf[64];
        sprintf(buf, "%d", node);
        return String(buf);
	}

	virtual String TransitionLabel(Nfa &nfa, NfaTransition &trns)
	{
		String ret;
		String range;
		bool bAddComma = false;
		ret = "{";
		for( UIntSet::iterator cur = trns.m_on.begin(); cur != trns.m_on.end(); ++cur )
		{
			if( bAddComma )
				ret += ",";
			else
				bAddComma = true;
            char buf[128];
            sprintf(buf, "%d->%d", cur->m_min, cur->m_max);
			ret += buf;
		}
		ret += "}";
		return ret;
	}

	virtual String ETransitionLabel(Nfa &nfa, NfaETransition &etrns)
	{
		String ret;
		ret = "{}";
		return ret;
	}
};

SimpleNfaDiagramHelper SimpleNfaDiagramHelper::m_Instance;

void Nfa::Print(outbuf *s, NfaDiagramHelper *pHelper, String name)
{
	if( ! s )
		return;
	if( ! pHelper )
		pHelper = SimpleNfaDiagramHelper::Instance();
	UIntSet startstates;
	UIntSet endstates;
	for( vector<int>::iterator cur = m_startstates.begin(); cur != m_startstates.end(); ++cur )
		startstates += *cur;
	for( vector<int>::iterator cur = m_endstates.begin(); cur != m_endstates.end(); ++cur )
		endstates += *cur;
	ob_printf(s, "digraph %s ", name.c_str());
    ob_puts(s, "{\n    node [style=filled shape=circle width=0.5]\n");
	for( int i = 0, n = statecount(); i < n; ++i )
	{
		String nodelabel = pHelper->NodeLabel(*this, i);
		ob_printf(s, "    \"%d\" [ label=\"%s\" color=black", i, nodelabel.c_str());
		if( startstates & i )
			ob_puts(s, " fontcolor=black fillcolor=blue");
		else
			ob_puts(s, " fontcolor=black fillcolor=cadetblue1");
		if( endstates & i )
			ob_puts(s, " peripheries=2");
		 ob_puts(s, "];\n");
	}
	for( vector<NfaTransition>::iterator cur = m_transitions.begin(); cur != m_transitions.end(); ++cur )
	{
		String label = pHelper->TransitionLabel(*this, *cur);
		ob_printf(s, "    \"%d\" -> \"%d\" [arrowtail=none color=blue label=\"%s\"]\n", cur->m_from, cur->m_to, label.c_str());
	}
	for( vector<NfaETransition>::iterator cur = m_etransitions.begin(); cur != m_etransitions.end(); ++cur )
	{
		String label = pHelper->ETransitionLabel(*this, *cur);
		ob_printf(s, "    \"%d\" -> \"%d\" [arrowtail=none color=lightblue label=\"%s\"]\n", cur->m_from, cur->m_to, label.c_str());
	}
	ob_puts(s, "}\n");
}

FrozenDfaTransition::FrozenDfaTransition() : m_start(-1), m_stop(-1), m_to(-1)
{
}

FrozenDfaTransition::FrozenDfaTransition(int start, int stop, int to) : m_start(start), m_stop(stop), m_to(to)
{
}

FrozenDfa::FrozenDfa() : m_statecount(0)
{
}

FrozenDfa::~FrozenDfa()
{
}

int FrozenDfa::transition(int state, int c)
{
    if( state < 0 || state >= (int)m_transitions.size() )
        return -1;
    vector<FrozenDfaTransition> &transitions = m_transitions[state];
    int min = 0, max = (int)transitions.size()-1, mid = 0;
    while( min <= max )
    {
        mid = (min + max) / 2;
        FrozenDfaTransition &t = transitions[mid];
        if( c < t.m_start )
        {
            max = mid-1;
            continue;
        }
        if( c > t.m_stop )
        {
            min = mid+1;
            continue;
        }
        return t.m_to;
    }
    return -1;
}

static int sortbyfrom(const FrozenDfaTransition *lhs, const FrozenDfaTransition *rhs)
{
    return lhs->m_start - rhs->m_start;
}

// Freeze an NFA.  Results on a DFA are undefined.
void Nfa::Freeze(FrozenDfa &dfa)
{
    dfa.m_statecount = statecount();
    dfa.m_transitions.clear();
    dfa.m_transitions.resize(dfa.m_statecount);
    for(vector<NfaTransition>::iterator cur = m_transitions.begin(), end = m_transitions.end(); cur != end; ++cur )
    {
        vector<FrozenDfaTransition> &trns = dfa.m_transitions[cur->m_from];
        for( UIntSet::iterator currange = cur->m_on.begin(), endrange = cur->m_on.end(); currange != endrange; ++currange )
            trns.push_back( FrozenDfaTransition(currange->m_min, currange->m_max, cur->m_to) );
    }
    for(vector< vector<FrozenDfaTransition> >::iterator cur = dfa.m_transitions.begin(), end = dfa.m_transitions.end(); cur != end; ++cur )
        cur->sort(sortbyfrom);
}
