#ifdef __cplusplus
extern "C" {
#endif

GLR_API nodevisitor *nv_steal(nodevisitor *self)
{
    --self->ref;
    return self;
}

GLR_API void nv_addref(nodevisitor *self)
{
    ++self->ref;
}

GLR_API void nv_decref(nodevisitor *self)
{
    if( --self->ref == 0 )
    {
        if( self->destroy )
            self->destroy(self);
        vp_realloc(self, 0);
    }
}

GLR_API nodevisitor *new_nv(size_t size, nvf_destroy destroy, nvf_isvisitmatch isvisitmatch, nvf_visit visit)
{
    nodevisitor *self = (nodevisitor*)vp_realloc(0, size);
    vp_memzero(self, size);
    self->ref = 1;
    self->destroy = destroy;
    self->isvisitmatch = isvisitmatch;
    self->visit = visit;
    return self;
}

GLR_API int nv_isvisitmatch(nodevisitor *self, toknode *node)
{
    if( self->isvisitmatch )
        return self->isvisitmatch(self, node);
    return 0;
}

GLR_API int nv_visit(nodevisitor *self, toknode *node, toknode **newnode)
{
    if( self->visit )
        return self->visit(self, node, newnode);
    *newnode = node;
    return 1;
}

typedef struct tag_aggregate_nv
{
    nodevisitor base;
    int count, capacity;
    nodevisitor **visitors;
} aggregate_nv;

GLR_API void ag_destroy(aggregate_nv *self)
{
    int i = 0;
    for( ; i < self->count; ++i )
        nv_decref(self->visitors[i]);
    vp_realloc(self->visitors, 0);
    self->visitors = 0;
    self->count = self->capacity = 0;
}

GLR_API int ag_isvisitmatch(aggregate_nv *self, toknode *node)
{
    int i = 0;
    for( ; i < self->count; ++i )
    {
        if( ! nv_isvisitmatch(self->visitors[i], node) )
            return 0;
    }
    return 1;
}

GLR_API int ag_visit(aggregate_nv *self, toknode *node, toknode **newnode)
{
    int i = 0;
    *newnode = 0;
    for( ; i < self->count; ++i )
    {
        if( *newnode )
        {
            node = *newnode;
            *newnode = 0;
        }
        if( ! nv_visit(self->visitors[i], node, newnode) )
            return 0;
    }
    return 1;
}

GLR_API void ag_add_visitor(nodevisitor *self, nodevisitor *nv)
{
    aggregate_nv *ag_self = (aggregate_nv*)self;
    if( ag_self->count == ag_self->capacity )
    {
        if( ag_self->capacity == 0 )
            ag_self->capacity = 8;
        else
            ag_self->capacity *= 2;
        ag_self->visitors = (nodevisitor**)vp_realloc(ag_self->visitors, sizeof(nodevisitor*)*ag_self->capacity);
    }
    nv_addref(nv);
    ag_self->visitors[ag_self->count++] = nv;
}

GLR_API nodevisitor *new_agg_nv()
{
    return new_nv(sizeof(aggregate_nv), (nvf_destroy)ag_destroy, (nvf_isvisitmatch)ag_isvisitmatch, (nvf_visit)ag_visit);
}

typedef struct tag_match_sym_nv
{
    nodevisitor base;
    int sym;
} match_sym_nv;

GLR_API int match_sym_isvisitmatch(match_sym_nv *self, toknode *node)
{
    if( node->sym == self->sym )
        return 1;
    return 0;
}

GLR_API nodevisitor *new_match_sym(int sym)
{
    match_sym_nv *self = (match_sym_nv*)new_nv(sizeof(match_sym_nv), 0, (nvf_isvisitmatch)match_sym_isvisitmatch, 0);
    self->sym = sym;
    return (nodevisitor*)self;
}

typedef struct tag_with_subnode_nv
{
    nodevisitor base;
    int subnode;
    nodevisitor *subvisitor;
} with_subnode_nv;

GLR_API void with_subnode_destroy(with_subnode_nv *self)
{
    nv_decref(self->subvisitor);
    self->subvisitor = 0;
}

GLR_API int with_subnode_isvisitmatch(with_subnode_nv *self, toknode *node)
{
    if( self->subnode < 0 || self->subnode >= (int)node->sncount || node->subnodes[self->subnode] == 0 )
        return 0;
    return nv_isvisitmatch(self->subvisitor, node->subnodes[self->subnode]);
}

GLR_API void toknode_remove_subnode(toknode *node, int subnode)
{
    size_t i;
    toknode_decref(node->subnodes[subnode]);
    for( i = subnode+1; i < node->sncount; ++i )
        node->subnodes[i-1] = node->subnodes[i];
    node->sncount--;
    node->subnodes = (toknode**)vp_realloc(node->subnodes, node->sncount*sizeof(toknode*));
}

GLR_API int with_subnode_visit(with_subnode_nv *self, toknode *node, toknode **newnode)
{
    toknode *subnode = node->subnodes[self->subnode];
    toknode *newsubnode = 0;
    size_t i = 0;
    if( ! nv_visit(self->subvisitor, subnode, &newsubnode) )
        return 0;
    if( subnode != newsubnode )
    {
        if( newsubnode )
            node->subnodes[self->subnode] = newsubnode;
        else
            toknode_remove_subnode(node, self->subnode);
    }
    *newnode = node;
    return 1;
}

GLR_API nodevisitor *new_with_subnode(int subnode, nodevisitor *visitor)
{
    with_subnode_nv *self = (with_subnode_nv*)new_nv(sizeof(with_subnode_nv), (nvf_destroy)with_subnode_destroy, (nvf_isvisitmatch)with_subnode_isvisitmatch, (nvf_visit)with_subnode_visit);
    self->subnode = subnode;
    self->subvisitor = visitor;
    nv_addref(visitor);
    return (nodevisitor*)self;
}

typedef struct tag_modify_string
{
    nodevisitor base;
    char *str;
} modify_string_nv;

GLR_API void modify_string_destroy(modify_string_nv *self)
{
    vp_realloc(self->str, 0);
    self->str = 0;
}

GLR_API int modify_string_isvisitmatch(modify_string_nv *self, toknode *node)
{
    if( node->sncount == 0 && node->tokstr )
        return 1;
    return 0;
}

GLR_API int modify_string_visit(modify_string_nv *self, toknode *node, toknode **newnode)
{
    if( node->tokstr )
        vp_realloc(node->tokstr, 0);
    node->tokstr = nv_strdup(self->str, &node->tokslen);
    *newnode = node;
    return 1;
}

GLR_API nodevisitor *new_modify_string(const char *str)
{
    modify_string_nv *self = (modify_string_nv*)new_nv(sizeof(modify_string_nv), (nvf_destroy)modify_string_destroy, (nvf_isvisitmatch)modify_string_isvisitmatch, (nvf_visit)modify_string_visit);
    self->str = nv_strdup(str, 0);
    return (nodevisitor*)self;
}

typedef struct tag_swap_subnodes
{
    nodevisitor base;
    int a, b, swapws;
} swap_subnodes_nv;

GLR_API int swap_subnodes_isvisitmatch(swap_subnodes_nv *self, toknode *node)
{
    if( self->a < 0 || self->a >= (int)node->sncount || self->b < 0 || self->b >= (int)node->sncount )
        return 0;
    return 1;
}

GLR_API int swap_subnodes_visit(swap_subnodes_nv *self, toknode *node, toknode **newnode)
{
    toknode *tmp;
    tmp = node->subnodes[self->a];
    node->subnodes[self->a] = node->subnodes[self->b];
    node->subnodes[self->b] = tmp;
    if( self->swapws )
    {
        tmp = node->subnodes[self->a]->wstop;
        node->subnodes[self->a]->wstop = node->subnodes[self->b]->wstop;
        node->subnodes[self->b]->wstop = tmp;
        tmp = node->subnodes[self->a]->wsbottom;
        node->subnodes[self->a]->wsbottom = node->subnodes[self->b]->wsbottom;
        node->subnodes[self->b]->wsbottom = tmp;
    }
    *newnode = node;
    return 1;
}

GLR_API nodevisitor *new_swap_subnodes(int a, int b, int swapws)
{
    swap_subnodes_nv *self = (swap_subnodes_nv*)new_nv(sizeof(swap_subnodes_nv), 0, (nvf_isvisitmatch)swap_subnodes_isvisitmatch, (nvf_visit)swap_subnodes_visit);
    self->a = a;
    self->b = b;
    self->swapws = swapws;
    return (nodevisitor*)self;
}

GLR_API int topdown_visit(toknode *node, toknode **newnode, nodevisitor *visitor)
{
    size_t i;
    toknode *subnode, *newsubnode;
    if( ! node )
    {
        *newnode = 0;
        return 1;
    }
    if( nv_isvisitmatch(visitor, node) )
        return nv_visit(visitor, node, newnode);
    for( i = 0; i < node->sncount; ++i )
    {
        subnode = node->subnodes[i];
        if( ! subnode )
            continue;
        newsubnode = 0;
        if( ! topdown_visit(subnode, &newsubnode, visitor) )
            return 0;
        if( subnode != newsubnode )
        {
            if( newsubnode )
                node->subnodes[i] = newsubnode;
            else
            {
                toknode_remove_subnode(node, i);
                --i;
            }
        }
    }
    *newnode = node;
    return 1;
}

GLR_API void reconstitute(outbuf *buf, toknode *node)
{
    size_t i = 0;
    if( ! node )
        return;
    if( node->wstop )
        reconstitute(buf, node->wstop);
    if( node->sncount == 0 && node->tokstr )
        ob_write(buf, node->tokstr, node->tokslen);
    else
    {
        for( i = 0; i < node->sncount; ++i )
            reconstitute(buf, node->subnodes[i]);
    }
    if( node->wsbottom )
        reconstitute(buf, node->wsbottom);
}

#ifdef __cplusplus
} /* extern "C" */
#endif
