// Copyright (c) 2008-2011 Mighty Heave Software LLC.
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
#ifdef WIN32
#include <windows.h>
#endif
#include "str.h"
#include <math.h>
#include <stdlib.h>
#include "chtest.h"
#include <string.h>
#ifdef _CRTDBG_MAP_ALLOC
#include <crtdbg.h>
#endif
#include <stdio.h> // vsprintf

unsigned int mbget(const char *s, size_t bufcount, size_t *count, int *err)
{
	if( bufcount == 0 )
	{
        *count = 0;
		*err = MBGET_ZERO_LENGTH_BUFFER;
		return '?';
	}
    unsigned int c;
    if( *s <= 0x7F )
    {
        c = *s;
        *count = 1;
    }
    else
    {
        unsigned int b0 = (unsigned int)*s++;
        unsigned int encodedbytes = 0;
        while( b0 & 0x0080 )
        {
            b0 <<= 1;
            encodedbytes++;
        }
		if( bufcount < encodedbytes  )
		{
			*err = encodedbytes; // this works, yields two to six.
			*count = bufcount;
			return '?';
		}
        b0 &= 0x007F;
        b0 >>= encodedbytes;
        c = b0;
        *count = encodedbytes;
        --encodedbytes;
		int i = 0;
        while( encodedbytes-- )
        {
            unsigned int bN = (unsigned int)*s++;
			if( (bN & 0xC0000000) != 0x80000000 )
				*err |= (MBGET_BAD_MULTIBYTE_ENCODING_BYTE2<<i);
			if( i == 0 && (bN & 0x3FFFFFFF) == 0 )
				*err |= MBGET_OVERLONG_ENCODING;
            c = (c<<6)|(bN&0x7F);
			++i;
        }
    }
    return c;
}

size_t mbput(char *dst, unsigned int c)
{
	if( c == UNICODE_REPLACEMENT_CHARACTER )
	{
        if( dst )
            *dst = '?';
        return 1;
	}
    else if( c <= 0x7F )
    {
        if( dst )
            *dst = c;
        return 1;
    }
    // Shift until the bits fall off.
    size_t bits = 0;
    unsigned int tmp = c;
    while(tmp)
    {
        tmp >>= 1;
        ++bits;
    }
    // Figure out how many bytes we're going to output.
    // Then setup byte 0 with that many high bits.
    size_t bytes = (bits / 6) + 1;
    // If we're not writing, we're done.
    if( ! dst )
        return bytes;
    unsigned int b0 = 0;
    for( size_t i = 0; i < bytes; ++i )
    {
        b0 >>= 1;
        b0 |= 0x80;
    }
    // Great, now chop off the bits for the first byte, and add them to byte 0.
    size_t bitoff = bits % 6;
    b0 |= ((c >> (bits-bitoff))&0x3F);
    *dst++ = (char)(unsigned char)b0;
    // Now stuff the remaining bytes, 6 bits at a time.
    for( size_t i = 1; i < bytes; ++i )
    {
        unsigned int bN = ((c >> (bits-bitoff+6))&0x3F);
        *dst++ = (char)(unsigned char)bN;
        bitoff += 6;
    }
    return bytes;
}

long LIBGLR_EXPORT glr_atoi(const char *s, const char *ends, const char **ps)
{
    size_t rlength = ends ? ends-s : strlen(s);
    size_t read_count = 0;
    int err = 0;
    codepoint cp;
    while(1)
    {
        cp = mbget(s, rlength, &read_count, &err);
        if( read_count == 0 || ! glr_isspace(cp) )
            break;
        s += read_count;
        rlength -= read_count;
    }
    int64_t ret = 0;
    int i;
    bool bNeg = false;
    cp = mbget(s, rlength, &read_count, &err);
    if( read_count && (cp == '+' || cp == '-') )
    {
        s += read_count;
        rlength -= read_count;
        if( cp == '-' )
            bNeg = true;
    }
    while(1)
    {
        cp = mbget(s, rlength, &read_count, &err);
        if( read_count == 0 || (i = glr_digitvalue(cp)) == -1 )
            break;
        s += read_count;
        rlength -= read_count;
        ret *= 10;
        ret += i;        
    }
    if( bNeg )
        ret = -ret;
    if( ps )
        *ps = s;
    return (long)ret;
}

void StringPacket::bump_capacity(size_t newmin)
{
    if( m_capacity >= newmin )
        return;
    size_t newcapacity = m_capacity;
    while( newcapacity < newmin )
    {
        if( newcapacity == 0 )
            newcapacity = 64;
        else
            newcapacity *= 2;
    }
    m_str = (char*)realloc(m_str, newcapacity);
    m_capacity = newcapacity;
}

StringPacket::StringPacket() : m_ref(1), m_count(0), m_capacity(0), m_str(0)
{
    bump_capacity(1);
    m_count = 1;
    *m_str = 0;
}

StringPacket::~StringPacket()
{
    if( m_str )
        free(m_str);
}

void StringPacket::IncRef()
{
#ifdef WIN32
    InterlockedIncrement(&m_ref);
#else
    ++m_ref;
#endif
}

void StringPacket::DecRef()
{
#ifdef WIN32
    if( InterlockedDecrement(&m_ref) == 0 )
        delete this;
#else
    if( --m_ref == 0 )
        delete this;
#endif
}

void StringPacket::set(const char *s, int len)
{
    if( len == -1 )
        len = strlen(s);
    if(s == m_str)
    {
        // handle self assignment without crashing.
        if( (size_t)(len+1) > m_count )
            len = m_count-1;
        m_str[len] = 0;
        m_count = len+1;
        return;
    }
    bump_capacity(len+1);
    for( int i = 0; i < len; ++i )
        m_str[i] = s[i];
    m_count = len+1;
    m_str[len] = 0;
}

void StringPacket::append(const char *s, int len)
{
    if( len == -1 )
        len = strlen(s);
    bump_capacity(m_count+len);
    char *dst = m_str+(m_count-1);
    for( int i = 0; i < len; ++i )
        dst[i] = s[i];
    dst[len] = 0;
    m_count += len;
}

void String::prep_packet()
{
    if( m_packet )
    {
        if( m_packet->m_ref > 1 )
        {
            StringPacket *newpacket = new StringPacket();
            newpacket->set(m_packet->m_str, m_packet->m_count-1);
            m_packet->DecRef();
            m_packet = newpacket;
        }
    }
    else
        m_packet = new StringPacket();
}

String::String() : m_packet(0)
{
}

String::String(const char *rhs) : m_packet(0)
{
    if( *rhs )
    {
        m_packet = new StringPacket();
        m_packet->set(rhs, -1);
    }
}

String::String(const String &rhs) : m_packet(0)
{
    if( rhs.m_packet )
    {
        m_packet = rhs.m_packet;
        m_packet->IncRef();
    }
}

String::~String()
{
    if( m_packet )
        m_packet->DecRef();
}

String &String::operator=(const String &rhs)
{
    if( m_packet )
        m_packet->DecRef();
    m_packet = rhs.m_packet;
    if( m_packet )
        m_packet->IncRef();
    return *this;
}

String &String::operator=(const char *rhs)
{
    if(!*rhs)
    {
        if( m_packet )
        {
            m_packet->DecRef();
            m_packet = 0;
        }
    }
    else
    {
        prep_packet();
        if( m_packet )
            m_packet->set(rhs, -1);
    }
    return *this;
}

void String::operator+=(int ch)
{
    prep_packet();
    char sfull[7] = {0, 0, 0, 0, 0, 0, 0};
    size_t len = mbput(sfull, ch);
    m_packet->append(sfull, len);
}

void String::operator+=(const char *rhs)
{
    if( ! *rhs )
        return;
    prep_packet();
    m_packet->append(rhs, -1);
}

void String::operator+=(const String &rhs)
{
    if( ! rhs.length() )
        return;
    prep_packet();
    m_packet->append(rhs.c_str(), rhs.length());
}

bool String::operator==(const char *rhs) const
{
    return strcmp(c_str(),rhs) == 0;
}

bool String::operator==(const String &rhs) const
{
    if( m_packet == rhs.m_packet )
        return true;
    return strcmp(c_str(),rhs.c_str()) == 0;
}

bool String::operator!=(const char *rhs) const
{
    return ! operator==(rhs);
}

bool String::operator!=(const String &rhs) const
{
    return ! operator==(rhs);
}

bool String::operator!() const
{
    return length() == 0;
}

String String::operator+(const String &rhs)
{
    String newstr = *this;
    newstr += rhs;
    return newstr;
}

String String::operator+(const char *rhs)
{
    String newstr = *this;
    newstr += rhs;
    return newstr;
}

const char *String::c_str() const
{
    if( m_packet )
        return m_packet->m_str;
    return "";
}

size_t String::length() const
{
    if( m_packet && m_packet->m_count )
        return m_packet->m_count-1;
    return 0;
}

String String::substr(int start, int end)
{
    if( ! m_packet )
        return String();
    int len = length();
    if( start < 0 ) start += len;
    if( start < 0 ) start = 0;
    else if( start >= len ) start = len;
    if( end < 0 ) end += len;
    if( end < 0 ) end = 0;
    else if( end >= len ) end = len;
    String newstr;
    newstr.prep_packet();
    newstr.m_packet->set(c_str()+start, end-start);
    return newstr;
}

String String::replace(const char *sold, const char *snew)
{
    if( ! length() )
        return String();
    String ret;
    ret.prep_packet();
    size_t lenold = strlen(sold);
    size_t lennew = strlen(snew);
    const char *s = c_str();
    char c = *sold;
    const char *ends = s;
    while( *ends )
    {
        if( *ends != c )
            ++ends;
        else if( strncmp(ends, sold, lenold) == 0 )
        {
            ret.m_packet->append(s, ends-s);
            ret.m_packet->append(snew, lennew);
            s = ends + lenold;
            ends = s;
        }
        else
            ++ends;
    }
    if( ends > s )
        ret.m_packet->append(s, ends-s);
    return ret;
}

String String::toupper()
{
    if( ! length() )
        return String();
    codepoint c = 0;
    size_t offset = 0;
    String ret;
    while( (c = get_codepoint(offset)) != 0xffffffff )
    {
        c = glr_toupper(c);
        ret += c;
    }
    return ret;
}

codepoint String::get_codepoint(size_t &offset)
{
    if( ! m_packet || offset >= m_packet->m_count ) return 0xffffffff;
    size_t read_count = 0;
    int err = 0;
    codepoint c = mbget(m_packet->m_str+offset, m_packet->m_count-offset-1, &read_count, &err);
    if( err )
        c = 0xffffffff;
    offset += read_count;
    return c;
}

void String::printf(const char *fmt, ...)
{
    // Until we can write a better version.
    char buf[1024];
    va_list arg;
    va_start(arg, fmt);
    vsprintf(buf, fmt, arg);
    va_end(arg);
    *this += buf;
}
