// Copyright (c) 2008-2011 Mighty Heave Software LLC.
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
#ifndef __str_h
#define __str_h

#include <stdarg.h>
#include "glrexport.h"
#include <stddef.h>

class LIBGLR_EXPORT StringPacket
{
public:
    StringPacket();
    ~StringPacket();
    void IncRef();
    void DecRef();
    volatile long m_ref;
    size_t m_count, m_capacity;
    char *m_str;
    void set(const char *s, int len);
    void append(const char *s, int len);
    void bump_capacity(size_t newmin);
};

// holds a utf-8 string.
class LIBGLR_EXPORT String
{
protected:
    StringPacket *m_packet;
    void prep_packet();

public:
    String();
    String(const char *rhs);    // utf-8
    String(const String &rhs);
    ~String();
    String &operator=(const String &rhs);
    String &operator=(const char *rhs);
    void operator+=(int ch);
    void operator+=(const char *rhs);
    void operator+=(const String &rhs);
    bool operator==(const char *rhs) const;
    bool operator==(const String &rhs) const;
    bool operator!=(const char *rhs) const;
    bool operator!=(const String &rhs) const;
	bool operator!() const;
    String operator+(const String &rhs);
    String operator+(const char *rhs);

    const char *c_str() const;
    size_t length() const;

    String substr(int start, int end);
	String replace(const char *sold, const char *snew);

    String toupper();
    void printf(const char *fmt, ...);

    unsigned int get_codepoint(size_t &offset);
};

// Handles any decimal digit known to unicode 5.0.1
long LIBGLR_EXPORT glr_atoi(const char *s, const char *ends, const char **ps);
// Read a character.
unsigned int LIBGLR_EXPORT mbget(const char *mbstr, size_t src_length, size_t *read_count, int *err);

// 1 through 6 are mutually exclusive, and will not appear with any other errors.
#define MBGET_ZERO_LENGTH_BUFFER  (1)  // src_length was 0, returned UNICODE_REPLACEMENT_CHARACTER
#define MBGET_2_CHAR_DECODING_REQ (2)  // src_length was 1, but need 2 to decode, returned UNICODE_REPLACEMENT_CHARACTER, read_count += src_length
#define MBGET_3_CHAR_DECODING_REQ (3)  // src_length was <3, but need 3 to decode, returned UNICODE_REPLACEMENT_CHARACTER, read_count += src_length
#define MBGET_4_CHAR_DECODING_REQ (4)  // src_length was <4, but need 4 to decode, returned UNICODE_REPLACEMENT_CHARACTER, read_count += src_length
#define MBGET_5_CHAR_DECODING_REQ (5)  // src_length was <5, but need 5 to decode, returned UNICODE_REPLACEMENT_CHARACTER, read_count += src_length
#define MBGET_6_CHAR_DECODING_REQ (6)  // src_length was <6, but need 6 to decode, returned UNICODE_REPLACEMENT_CHARACTER, read_count += src_length
// bad encoding byte 2 through byte 6 are flags, but will not appear mixed with 1 through 7
#define MBGET_BAD_MULTIBYTE_ENCODING_BYTE2	   (0x0008) // byte 2 did not start with 10b (decoding and read_count unaffected)
#define MBGET_BAD_MULTIBYTE_ENCODING_BYTE3	   (0x0010) // byte 3 did not start with 10b (decoding and read_count unaffected)
#define MBGET_BAD_MULTIBYTE_ENCODING_BYTE4	   (0x0020) // byte 4 did not start with 10b (decoding and read_count unaffected)
#define MBGET_BAD_MULTIBYTE_ENCODING_BYTE5	   (0x0040) // byte 5 did not start with 10b (decoding and read_count unaffected)
#define MBGET_BAD_MULTIBYTE_ENCODING_BYTE6	   (0x0080)	// byte 6 did not start with 10b (decoding and read_count unaffected)
// a mask to test for any bad multibyte encoding.
#define MBGET_BAD_MULTIBYTE_ENCODING		   (MBGET_BAD_MULTIBYTE_ENCODING_BYTE2|MBGET_BAD_MULTIBYTE_ENCODING_BYTE3|MBGET_BAD_MULTIBYTE_ENCODING_BYTE4|MBGET_BAD_MULTIBYTE_ENCODING_BYTE5|MBGET_BAD_MULTIBYTE_ENCODING_BYTE6)
// overlong encoding encountered, may be mixed with bad multibyte encoding flags.
#define MBGET_OVERLONG_ENCODING				   (0x0100) // an overlong encoded was encountered (decodeing and read_count are unaffected)
#define UNICODE_REPLACEMENT_CHARACTER (0x0000FFFD)

// mbput - put a unicode character into a utf-8 buffer.
// dst - the place we put the utf-8 characters.
// c - the unicode character.
//
// Puts up to 6 multibyte characters.  Just make sure there's enough room.
// To check, you may call with dst of null, it will still return
// the number of char's that would have been encoded.
// whenever UNICODE_REPLACEMENT_CHARACTER is encountered, it is encoded as '?'.
size_t LIBGLR_EXPORT mbput(char *dst, unsigned int c);

#endif
