// Copyright (c) 2008-2011 Mighty Heave Software LLC.
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
#include "templatehelp.h"
#include <string.h>
#include <malloc.h>
#include <math.h>

extern uint32_t hashbig( const void *key, size_t length, uint32_t initval);

uint32_t hash_djb2(const char *str)
{
    uint32_t hash = 5381;
    int c;
    while (c = *str++)
        hash = ((hash << 5) + hash) + c; /* hash * 33 + c */
    return hash;
}

uint32_t hash_jenkins(uint32_t a)
{
    a = (a+0x7ed55d16) + (a<<12);
    a = (a^0xc761c23c) ^ (a>>19);
    a = (a+0x165667b1) + (a<<5);
    a = (a+0xd3a2646c) ^ (a<<9);
    a = (a+0xfd7046c5) + (a<<3);
    a = (a^0xb55a4f09) ^ (a>>16);
    return a;
}



uint32_t hash_uint32s(const uint32_t *values, size_t count, uint32_t initval)
{
    return hashbig(values, count*sizeof(uint32_t), initval);
}

int cmpuint32(const uint32_t *lhs, const uint32_t *rhs)
{
    if( *lhs > *rhs )
        return 1;
    else if( *lhs < *rhs )
        return -1;
    return 0;
}

// Do a binary search, return not just the insertion point,
// but the last places know to definitely be lower/higher than the test.
static bool base_binsearch(void *base, size_t num, size_t size, cmpfnct cmp, void *test, int *index, int *plow_bound, int *phigh_bound)
{
    int low = 0;
    int high = (int)num-1;
    int low_bound = -1;      // -1 is definetly lower that the test.
    int high_bound = (int)num;    // high is definetly higher than the test.
    int mid = low;
    while(low <= high)
    {
        mid = (low+high)/2;
        void *pmid = (char*)base+size*mid;
        int result = cmp(pmid, test);
        if( result > 0 )
        {
            high_bound = mid; // mid is now definitely higher than the test
            high = mid-1;
        }
        else if( result < 0 )
        {
            low_bound = mid; // mid is now definitely lower than the test
            low = mid+1;
        }
        else
        {
            if( plow_bound )
                *plow_bound = low_bound;
            if( phigh_bound )
                *phigh_bound = high_bound;
            *index = mid;
            return true;
        }
    }
    if( plow_bound )
        *plow_bound = low_bound;
    if( phigh_bound )
        *phigh_bound = high_bound;
    *index = high_bound;
    return false;
}

bool find_binsearch(void *base, size_t num, size_t size, cmpfnct cmp, void *test, int *index)
{
    return base_binsearch(base, num, size, cmp, test, index, 0, 0);
}

static int cmp_normalize(int r)
{
    if( r < 0 )
        return -1;
    else if( r > 0 )
        return 1;
    return 0;
}

static int findedge_binsearch(void *base, size_t num, size_t size, cmpfnct cmp, void *test, int low_edge, int low_edge_cmp, int high_edge, int high_edge_cmp)
{
    while(high_edge - low_edge > 1)
    {
        int mid = (low_edge+high_edge)/2;
        void *pmid = (char*)base+size*mid;
        int result = cmp_normalize(cmp(pmid, test));
        if( result == low_edge_cmp )
        {
            low_edge = mid;
            continue;
        }
        if( result = high_edge_cmp )
        {
            high_edge = mid;
            continue;
        }
        // unsorted list or other bad input.
        return -1;
    }
    return low_edge;
}

bool findrange_binsearch(void *base, size_t num, size_t size, cmpfnct cmp, void *test, int *begin, int *end)
{
    int index, low_bound, high_bound;
    if( ! base_binsearch(base, num, size, cmp, test, &index, &low_bound, &high_bound) )
    {
        *end = *begin = index;
        return false;
    }
    // here we run into a dilema on allowing duplicates.  When there are few duplicates, we will waste
    // lots of time searching for the lower boundaries, that a simple check on a values before and after
    // the binary searched index could have found.  When there are lots of duplicates, a binary search
    // of the area of possibly duplictes value will be more efficient.
    int low_index = index;
    int high_index = index;
    if( low_index > 0 )
        low_index = findedge_binsearch(base, num, size, cmp, test, low_bound, -1, low_index, 0) + 1;
    if( high_index < ((int)num-1) )
        high_index = findedge_binsearch(base, num, size, cmp, test, high_index, 0, high_bound, 1);
    *begin = low_index;
    *end = high_index + 1;
    return true;
}

static void quicksort_inner(void *data, size_t datasize, int lo, int hi, cmpfnct cmp, void *tmp)
{
	int i = lo, j = hi;
	unsigned char *x = ((unsigned char*)data)+datasize*((lo+hi)/2);
	unsigned char *pi = ((unsigned char*)data)+datasize*i;
	unsigned char *pj = ((unsigned char*)data)+datasize*j;
	do
	{
		while( cmp(pi, x) < 0 ) { pi+=datasize; i++; }
		while( cmp(pj, x) > 0 ) { pj-=datasize; j--; }
		if(i<=j)
		{
			memcpy(tmp, pi, datasize);
			memcpy(pi, pj, datasize);
			memcpy(pj, tmp, datasize);
			if( x == pi )
				x = pj;
			else if( x == pj )
				x = pi;
			pi+=datasize; i++;
			pj-=datasize; j--;
		}
	} while(i<=j);
	if(lo<j) quicksort_inner(data, datasize, lo, j, cmp, tmp);
	if(i<hi) quicksort_inner(data, datasize, i, hi, cmp, tmp);
}

void glr_quicksort(void *data, size_t datasize, size_t count, cmpfnct cmp)
{
    if( count < 2 )
        return;
    void *tmp = malloc(datasize);
    quicksort_inner(data, datasize, 0, (int)count-1, cmp, tmp);
    free(tmp);
}

void glr_reverse(void *data, int datasize, int count)
{
    if( count < 2 )
        return;
	void *tmp = malloc(datasize);
	for( int i = 0, m = count, n = count / 2; i < n; ++i )
	{
		register void *lhs = ((unsigned char*)data)+datasize*i;
		register void *rhs = ((unsigned char*)data)+datasize*(m-i-1);
		memcpy(tmp, lhs, datasize);
		memcpy(lhs, rhs, datasize);
		memcpy(rhs, tmp, datasize);
	}
  free(tmp);
}

double round(double d)
{
	double whole = floor(d);
	d -= whole;
	if( d < 0.5 )
		return whole;
	return whole + 1.0;
}
