// Copyright (c) 2008-2011 Mighty Heave Software LLC.
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.
#ifndef __vec_h
#define __vec_h

#pragma warning( disable : 4251 )
#pragma warning( disable : 4345 )

#include <new>
#include "templatehelp.h"

#ifdef DO_HEAPCHECK
#define HEAPCHECK _heapchk();
#else
#define HEAPCHECK
#endif

#include <stdlib.h>
#ifdef _CRTDBG_MAP_ALLOC
#include <crtdbg.h>
#endif
#include <string.h>
#include <stddef.h>

template<class T>
class vector
{
protected:
	T *m_data;
	size_t m_count;
	size_t m_capacity;

private:
	static void move(T *dst, T *src, size_t sz)
	{
		if( sz == 0 )
			return;
		HEAPCHECK
        memmove(dst, src, sz*sizeof(T));
		HEAPCHECK
	}

	static void zero(T *dst, size_t sz)
	{
		if( sz == 0 )
			return;
		HEAPCHECK
        memset(dst, 0, sz*sizeof(T));
		HEAPCHECK
	}

public:
	typedef T *iterator;
	typedef const T *const_iterator;

	vector() : m_data(0), m_count(0), m_capacity(0)
	{
	}

	vector(const vector<T> &rhs) : m_data(0), m_count(0), m_capacity(0)
	{
		appendmany(rhs.m_data, rhs.m_count);
	}

	vector<T> &operator=(const vector<T> &rhs)
	{
		clear();
		appendmany(rhs.m_data, rhs.m_count);
		return *this;
	}

	~vector()
	{
		erase();
	}

	void reserve(size_t count)
	{
		if( m_capacity >= count )
            return;
		HEAPCHECK
		T *data =  (T*)realloc(m_data, sizeof(T)*count);
        if( ! data )
        {
            // This could mean out of memory.
            // However, it usually means you've screwed up and used
            // realloc/malloc/free before the CRT was initialized.
            data = (T*)malloc(sizeof(T)*count);
            if( count > m_count )
                memcpy(data, m_data, sizeof(T)*m_count);
        }
        m_data = data;
		HEAPCHECK
		m_capacity = count;
	}

	void resize(size_t count, const T &t)
    {
		if( count > m_count )
		{
            size_t reservecount = m_capacity;
            while( reservecount < count )
            {
                if( reservecount == 0 )
			        reservecount = 16;
                else
                    reservecount *= 2;
            }
            reserve(reservecount);
			zero(m_data+m_count, count-m_count);
			for( size_t i = m_count, n = count; i < n; ++i )
				new(m_data+i) T(t);
			m_count = count;
		}
		else if( count < m_count )
		{
			for( size_t i = count, n = m_count; i < n; ++i )
				m_data[i].~T();
			m_count = count;
		}
    }

	void resize(size_t count)
	{
		if( count > m_count )
		{
            size_t reservecount = m_capacity;
            while( reservecount < count )
            {
                if( reservecount == 0 )
			        reservecount = 16;
                else
                    reservecount *= 2;
            }
            reserve(reservecount);
			zero(m_data+m_count, count-m_count);
			for( size_t i = m_count, n = count; i < n; ++i )
				new(m_data+i) T();
			m_count = count;
		}
		else if( count < m_count )
		{
			for( size_t i = count, n = m_count; i < n; ++i )
				m_data[i].~T();
			m_count = count;
		}
	}

	size_t size() const
	{
		return m_count;
	}

    bool empty() const
    {
        return m_count == 0;
    }

	size_t capacity() const
	{
		return m_capacity;
	}

	bool full() const
	{
		return m_count == m_capacity;
	}

	void clear()
	{
		for( size_t i = 0, n = m_count; i < n; ++i )
			m_data[i].~T();
		m_count = 0;
	}

	bool operator==(const vector<T> &rhs) const
	{
		if( m_count != rhs.m_count )
			return false;
		for( size_t i = 0, n = m_count; i < n; ++i )
			if( ! (m_data[i] == rhs.m_data[i]) )
				return false;
		return true;
	}

	bool operator!=(const vector<T> &rhs) const
	{
		return ! operator==(rhs);
	}

    bool operator<(const vector<T> &rhs) const
    {
        size_t cnt = m_count;
        if( rhs.m_count < m_count )
            cnt = rhs.m_count;
		for( size_t i = 0, n = cnt; i < n; ++i )
        {
            if( m_data[i] < rhs.m_data[i] )
                return true;
            if( rhs.m_data[i] < m_data[i] )
                return false;
        }
        if( cnt < rhs.m_count )
            return true;
        return false;
    }

	iterator begin()
	{
		return m_data;
	}

	iterator end()
	{
		return m_data+m_count;
	}

	const_iterator begin() const
	{
		return m_data;
	}

	const_iterator end() const
	{
		return m_data+m_count;
	}

	void push_back(const T &t)
	{
		resize(m_count+1);
		m_data[m_count-1] = t;
	}

	void pop_back()
	{
		m_data[m_count-1].~T();
		--m_count;
	}

	T &front()
	{
		return m_data[0];
	}

	const T &front() const
	{
		return m_data[0];
	}

	T &back()
	{
		return m_data[m_count-1];
	}

	const T &back() const
	{
		return m_data[m_count-1];
	}

	iterator insert(iterator at, const T &t)
    {
        return insertat(at-begin(), t);
    }

    template<class InputIterator>
	void insert(iterator at, InputIterator b, InputIterator e)
	{
		size_t count = e-b;
		size_t pos = at-m_data;
		reserve(m_count+count);
		at = m_data+pos;
		move(at+count, at, m_count-pos);
		m_count += count;
		zero(at, count);
		for( size_t i = 0; i < count; ++i )
		{
			new(at+i) T();
			at[i] = *b;
            ++b;
		}
	}

	iterator newitem()
	{
		resize(m_count+1);
		return m_data+m_count-1;
	}

	iterator newitems(size_t count)
	{
		resize(m_count+count);
		return m_data+m_count-count;
	}

	iterator insertat(size_t i, const T &t)
	{
		reserve(m_count+1);
		move(m_data+i+1, m_data+i, m_count-i);
		m_count++;
		zero(m_data+i, 1);
		new(m_data+i) T();
		m_data[i] = t;
		return m_data+i;
	}

	iterator append(const T &t)
	{
		push_back(t);
		return m_data+m_count-1;
	}

	void append(const vector<T> &rhs)
	{
		size_t origcount = m_count;
		resize(m_count+rhs.m_count);
		for( size_t i = 0, n = rhs.m_count; i < n; ++i )
			m_data[origcount+i] = rhs.m_data[i];
	}

	void appendmany(const T *rhs, size_t count)
	{
		size_t origcount = m_count;
		resize(m_count+count);
		for( size_t i = 0, n = count; i < n; ++i )
			m_data[origcount+i] = rhs[i];
	}

	void erase(iterator b, iterator e)
	{
		size_t removecount = e-b;
		for( T *f = b; f < e; ++f )
			f->~T();
		move(b, e, m_data+m_count-e);
		m_count -= removecount;
	}

    iterator erase(iterator b)
    {
        if( b < m_data || b >= m_data+m_count )
            return m_data+m_count;
        size_t pos = b - m_data;
        erase(b, b+1);
        return m_data+pos;
    }

	void erase()
	{
		clear();
		if( m_data )
		{
			free(m_data);
			m_data = 0;
			m_capacity = 0;
		}
	}

	void removeat(size_t index, size_t count)
	{
		if( index >= m_count || count == 0)
			return;
		if( count > (m_count-index) )
			count = m_count-index;
		for( T *f = m_data+index, *e=m_data+index+count; f < e; ++f )
			f->~T();
		move(m_data+index, m_data+index+count, m_count-index-count);
		m_count -= count;
	}

	iterator find(const T &t, size_t startat = 0)
	{
		if( startat > m_count )
			startat = m_count;
		for( T *i = m_data+startat, *n = m_data+m_count; i < n; ++i )
        {
			if( *i == t )
				return i;
        }
		return m_data+m_count;
	}

	bool remove(const T &t)
	{
		iterator iter = find(t);
		if( iter == end() )
			return false;
		erase(iter, iter+1);
        return true;
	}

public:
	void sort(int (*cmp)(const T *lhs, const T *rhs))
	{
        glr_quicksort(m_data, sizeof(T), m_count, (cmpfnct)cmp);
	}

	void unique(int (*cmp)(const T *lhs, const T *rhs))
	{
		if( m_count <= 1 )
			return;
		sort(cmp);
		for( int i = 0; i < (int)m_count-1; ++i )
		{
			T *lhs = m_data+i;
			T *rhs = lhs+1;
			if( cmp(lhs,rhs) == 0 )
			{
				int numtomove = (int)(m_count-i)-2;
				if( numtomove > 0 )
					move(rhs, rhs+1, numtomove);
				--m_count;
			}
		}
	}

	// Note, both 'this' and 'rhs' are assumed to be sorted the same way.
	bool IntersectsWith(const vector<T> &rhs) const
	{
		const_iterator _First1 = begin();
		const_iterator _Last1 = end();
		const_iterator _First2 = rhs.begin();
		const_iterator _Last2 = rhs.end();
		for (; _First1 != _Last1 && _First2 != _Last2; )
			if( *_First1 < *_First2 )
				++_First1;
			else if(*_First2 < *_First1)
				++_First2;
			else
				return true;
		return false;
	}

    bool binsearch(const T &t, iterator &first, iterator &last, int (*cmp)(const T *lhs, const T *rhs)) const
    {
        T *base = m_data;
        int num = m_count;
        first = last = base;
        int low = 0;
        int high = (int)m_count-1;
        int highmatch = -1;
        while(low <= high)
        {
            int mid = (low+high)/2;
            T *pmid = base+mid;
            int result = cmp(pmid, &t);
            if( result > 0 )
                high = mid-1;
            else if( result < 0 )
            {
                first = pmid+1;
                low = mid+1;
            }
            else
            {
                if( mid > highmatch )
                    highmatch = mid;
                if( pmid == base )
                {
                    first = pmid;
                    break;
                }
                T *pprev = pmid-1;
                if( cmp(pprev, &t) != 0 )
                {
                    first = pmid;
                    break;
                }
                high = mid-1;
            }
        }
        if( highmatch == -1 )
        {
            last = first;
            return false;
        }
        T *baselast = base+(num-1);
        low = highmatch;
        high = (int)num-1;
        while(low <= high)
        {
            int mid = (low+high)/2;
            T *pmid = base+mid;
            int result = cmp(pmid, &t);
            if( result > 0 )
                high = mid-1;
            else if( result < 0 )
                low = mid+1;
            else
            {
                if( pmid == baselast )
                {
                    last = pmid+1;
                    return true;
                }
                T *pnext = pmid+1;
                if( cmp(pnext, &t) != 0 )
                {
                    last = pnext;
                    return true;
                }
                low = mid+1;
            }
        }
        return false;
    }

    bool isvalidindex(int index) const
    {
		if( index < 0 )
			index += m_count;
        if( index < 0 || index >= (int)m_count )
            return false;
        return true;
    }

	const T &operator[](int index) const
	{
		if( index < 0 )
			index += m_count;
		return m_data[index];
	}

	T &operator[](int index)
	{
		if( index < 0 )
			index += (int)m_count;
		return m_data[index];
	}
#if 0
	const T &operator[](size_t index) const
	{
		if( index < 0 )
			index += m_count;
		return m_data[index];
	}

	T &operator[](size_t index)
	{
		if( index < 0 )
			index += (int)m_count;
		return m_data[index];
	}
#endif
	void reverse()
	{
        glr_reverse(m_data, sizeof(T), m_count);
	}
};

#endif
